.section .iwram,"ax",%progbits
.align 4
.arm

@ keypad interrupt in asm.


.extern thisKeys
.extern palBuf

#if defined(ELITE_HACK)
#define MENU_KEYS  (4)
#else
#define MENU_KEYS  (4|8)
#endif 

#define curkey        r3
#define nowpressed    r4
#define keyid    r5
#define row    r5
#define col    r6
#define tk            r7
#define lk            r8

.global irqarm
.type irqarm,function

irqarm:
    @ flow reaches here when an interrupt occurs
    @ the only interrupt I enable is the vblank one
    @load reg_ie
    stmfd    sp!,{r0-r8,lr} 
    
    @ copy palette buffer to palette mem    
     
    ldr r3,copyColours
    cmp r3,#0
    beq skip_colour
#if defined(ELITE_HACK)
@ check the character line, copy low half colour to palette 2
    ldr     r5,=palBuf
    ldrh    r1,[r5,#18]  @ 19th colour
		ldr     r5,Lwhite
    cmp     r1,r5
    moveq   r5,#0      @ we ate the carry flag subtracting, overflow
    movne   r5,#16*2   @ 2nd palette area offset
    @beq     skip_colour
@    cmp     r1,#0
@    bmi     skip_colour  @ the plan here was to skip "off screen" copies
    @cmp     r1,#0x1a
    @bpl     skip_colour
   
#endif
colour_copy:
    subs r3,r3,#1  
    ldrne r0,Lobjp      @ OBJPaletteMem
    addne r0,r0,r3,lsl#1      @ r2 is address which colour we change
    ldreq r0,Lbgp       @ BGPalette mem if changing colour 0 (otherwise we see no change)
#if defined(ELITE_HACK)
    addne  r0,r0,r5
#endif
    ldr   r2,=palBuf
    addne r2,r2,r3,lsl#1
    ldrh  r2,[r2]
    strh  r2,[r0]   
    bne colour_copy
skip_colour:
     
    @ check the keys
     @ read the current value from key reg 
     ldr   tk,REG_KEYS
     ldr   tk,[tk]           @ tk == KEY VAL
     ldr   r1,=thisKeys   @ check against the old value
     ldr   lk,[r1]        @ lk == lastKeys val
     str   tk,[r1]        @ store the last register in the buffer
     cmp   lk,tk
     beq endKeyChecks     @ skip everything if they are the same
     @ first check if the menu keys are pressed...
     @ xxxxxxTLxx -> & 1100 -> 000000TL00
     and    r1,tk,#MENU_KEYS
     @ 000000TL00 -> ^ 1100 -> 000000~T~L00  (recall OFF is really on)
     eors   r1,r1,#MENU_KEYS
     cmp    r1,#MENU_KEYS
     beq    showMenuPressed
     @ now for each key pressed  set the corresponding Beeb Key
     @ for each key let-go this time, unset the Beeb Key
          
     @ there are 10 keys, loop over all of them
@ order is A,B,SEL,ST,Ri,Le,Up,Dw,R,L; but that doesn't matter
convertKeys:
     mov    curkey,#0
     eor    lk,tk,lk         @ lk = lk ^ tk  (ie sets those that are different)
nextKeyLoop:
     add curkey,curkey,#1
     cmp curkey,#11
     beq endKeyChecks
     movs   tk,tk,lsr #1
     movcc  nowpressed,#1   
     movcs  nowpressed,#0   
     movs   lk,lk,lsr #1
      
@ case 1 : nowpressed  ---> Set PRESSED
@ case 2 : !nowpressed ---> UNSet PRESSED
     @cmp nowpressed,waspressed
     bcc   nextKeyLoop
     @ so which is it?
     bl  SetPressedStatus 
     b   nextKeyLoop
endKeyChecks:    
   @ REG_IF |= INT_VBLANK;
    ldr   r0,REG_IE
    ldrh  r1,[r0,#2]  
    orr   r1,r1,#1
    strh  r1,[r0,#2]  
   @ (*(volatile u32*)0x03fffff8) |= INT_VBLANK;
    ldr   r0,BIOS_IR
    ldr   r1,[r0]  
    orr   r1,r1,#1
    strh  r1,[r0]  
@ and end..
    ldmfd    sp!,{r0-r8,lr} 
    mov      pc,lr

showMenuPressed:
    ldr      r0,=needKeyInterrupt
    mov      r1,#4
    str      r1,[r0]
    adr      lr,endKeyChecks
    ldr      r1,=releaseAllKeys
    bx       r1


Lwhite: .word ( 31|(31<<5)|(31<<10)  )
Lobjp: .word 0x5000200
Lbgp: .word 0x5000000
REG_KEYS: .word 0x04000130
REG_IE : .word 0x4000200
@REG_IF : .word 0x4000202
BIOS_IR : .word 0x03fffff8

.global needKeyInterrupt
needKeyInterrupt : .space 4
.global copyColours
copyColours : .space 4

.type SetPressedStatus,function
.global SetPressedStatus
SetPressedStatus:
    @check  nowpressed
@ if nowpressed == 1, then set the key pressed
@else if nowpressed == 0, set the key as released
    @ get the address of the key pressed in the mapped table
    ldr    r0,=mappedKeys
    sub    curkey,curkey,#1
    ldrb   keyid,[r0,curkey]  @ e.g. 0xd, ' ', 'Z', etc 
    cmp    keyid,#0
    moveq  pc,lr   @ key is 0, i.e. "off" 
    cmp    keyid,#255  @ key is "show vk"
    beq    showVirtualKeyBoard
    add    curkey,curkey,#1
    
    @ now from key id, get ROW, COL pressed. This is a pain
    @#define GET_TABLE_VAL(key, rowCol) ( (key < 0xe) ? specialsTable[key][rowCol] : \
    @                         ((key >= ' ' && key <= 0x6a)?   \
    @                             asciiTable[key-0x20][rowCol]   \
    @                            : 0 \
    @                            )\
    @                          )
    cmp      keyid,#0xe 
    ldrmi    r0,=specialsTable
    bmi      gotTable
    subs     keyid,keyid,#0x20
    ldrpl    r0,=asciiTable
    movmi    col,#0 
    movmi    row,#0 
    bmi      gotRowCol
gotTable:
    @ once we have table, get row and col for this keyid    
    add       r0,r0,keyid,lsl#1     @ byte table, 2 bytes per entry
    ldrb      row,[r0,#0]
    ldrb      col,[r0,#1]
gotRowCol:
    @ have row, col, so set/unset value in     
    ldr      r0,=SysViaKbdState @ the table 
    add      keyid,row,col,lsl #3            @ keyid = Row + 8*KBDCol
    strb     nowpressed,[r0,keyid]     @ if now pressed this will be 1, else is 0
    cmp      nowpressed,#0
    @   KeysDown++;
    ldr      r0,=KeysDown
    ldr      r1,[r0]
    addne    r1,r1,#1
    subeqs   r1,r1,#1
    movmi    r1,#0       @ perhaps we messed up with menus
    str      r1,[r0]
    moveq    pc,lr
    
    @ set the interrupt value
    ldr      r0,=needKeyInterrupt
    str      nowpressed,[r0]
.global ski
.type ski,function
ski:
    mov  pc,lr


.global loadingIR
.type loadingIR,function
loadingIR:
  @ this is called when you say "load something" via disk
  stmfd    sp!,{r0-r8,lr} 
  ldr      r1,=shiftDownFrames
  ldr      r0,[r1]
  subs     r0,r0,#1
  str    r0,[r1] 
  bpl endKeyChecks
@ mi means that we have finished, unpress shift and set the interrupt 
@ routine to the usual one
  ldr      r1,=SysViaKbdState
  mov      r0,#0
  strb     r0,[r1]      @ unpress shift
  ldr      r1,=KeysDown
  str      r0,[r1]      @ and nothing pressed
  ldr      r1,REG_IRQ
  ldr      r0,=irqarm
  str      r0,[r1]
  @ldmfd    sp!,{r0-r8,lr} 
  b        endKeyChecks
   
.type showVirtualKeyBoard, function
showVirtualKeyBoard:
  @ show vk, set interrupt
  ldr      r1,REG_IRQ
  ldr      r0,=VKInterruptHandler
  str      r0,[r1]
  ldr      r1,=apply_vk_press
  mov      r0,#1
  str      r0,[r1]

  ldr      r0,=show_vkb
  adr      lr,endKeyChecks
  bx       r0
  
REG_IRQ: .word 0x3007FFC
.global shiftDownFrames
shiftDownFrames: .space 4
.pool

.end
