//#include "agbprint.h"
#include "gba.h"
#include "gfx.h"
// DFS emulation

extern int initial_byte;    // in asm module
extern int final_byte;
int status = 0;
int result = 0;
int command = 0;
int datareg = 0;

typedef struct
{
  int command;
  int params;
  int mask;
  void (*func)();
  void (*intfunc)();
} fdc_comm_t;

// internal values
int n_params = 0;  // number of parameters
int current_param = 0; // current parameter
const fdc_comm_t * current_command = 0; // current command
int command_val = 0; // current command value
int disc_trigger = 0; // interrupt timer
#define DISC_CYCLES  160

int current_track = 0;  // only 1 drive
int current_sector = 0;
int sectors_left = 0;
int sector_len = 0; 
int bytes_in_sector = 0;

// hold the parameters
#define MAX_COMM_PARAMS  4
int command_params[MAX_COMM_PARAMS] = {0,0,0,0};
int NMIintStatus = 0;
#define updateNMI()  if (status&8) NMIintStatus |= 8; else NMIintStatus &= ~(16|8);

//#define DISC _binary_Elite_A_ssd_start
//extern unsigned char DISC[];
#define DISC 0

unsigned char * discs  = DISC;

void reset_disc_drive(void) {
  status = 0;
  datareg = 0;
  result = 0;
  updateNMI();

  n_params = 0;
  current_param = 0;
  initial_byte = 0;
  final_byte = 0;
}


void read_drive_status(void) {
  //gbaprintf("%s\n",__PRETTY_FUNCTION__);
  result = 0x88; // write protect, ready
  
  if (current_track == 0)
    result |= 2;
  
  result |= 0x4; // drive 0...
  // bug fix for elite - read drive status causes an interrupt
  status |= 0x18;
  // nmi
  updateNMI();  // no interrrupt

}
void write_special(void) {
  /*gbaprintf("%s switch %x %x \n",__PRETTY_FUNCTION__, command_params[0],
      command_params[1]);*/
  switch (command_params[0]) {
      case 6 : // internal sector number
        break;
      case 0x14 : // scan count
        break;
      case 0x13:// scan count
        break;
      case 0x12 : // current track
        current_track = command_params[1];
        break;
      case 0x1a : // current track, disc 2
        break;
      default : break;
  }
}
void read_special(void) {
//  gbaprintf("%s\n",__PRETTY_FUNCTION__);
//  while (1) {}
}
void specify(void) {
//  gbaprintf("%s - does nothing!\n",__PRETTY_FUNCTION__);
  status |= 0x00;
  // nmi
  updateNMI();  // no interrrupt
}
void seek(void) {
  //gbaprintf("%s %x %x\n",__PRETTY_FUNCTION__, command_val,  command_params[0]);
  current_track = command_params[0];
  status = 0x80;
  updateNMI();
  disc_trigger = 160;
}
void seekint(void) {
  //gbaprintf("%s\n",__PRETTY_FUNCTION__);
  status = 0x18; /* Result with interrupt */
  updateNMI();
  result = 0;
}

void read_varlen(void) {
  //gbaprintf("%s",__PRETTY_FUNCTION__);
  // read_varlen has 3 parameters
  // select drive (always 0?)
  // load head (always ignored?)
  //
  if (!(command_val & 0x40)) {
    status |= 0x80;
    //gbaprintf("\n");
    return;
  }
  current_track = command_params[0];
  current_sector = command_params[1];
  sectors_left = command_params[2]&0x1f;
  sector_len =1<<(7+((command_params[2]>>5)&7)); 
  bytes_in_sector = 0;
  
  // this makes life easier when reading bytes from disc..
  initial_byte = (current_sector*256) + (current_track*10*256);
  final_byte   = initial_byte+sector_len;
  
  status = 0x80;
  //nmi
  updateNMI();
  // set poll time
  disc_trigger = DISC_CYCLES;
  //gbaprintf(" tr %x, sec %x, for %x secs, ln %x\n",current_track, current_sector,sectors_left,sector_len);
  printTxtPal("loading...", 20,19,1);
}
// the read interrupt...
// this is written in assembler
void readint(void);

void write_varlen(void) {
  //gbaprintf("%s\n",__PRETTY_FUNCTION__);
  //while (1) {}
}
void writeint(void) {
//  gbaprintf("%s\n",__PRETTY_FUNCTION__);
//  while (1) {}
}
void verify_varlen(void) {
 // gbaprintf("%s\n",__PRETTY_FUNCTION__);
//  while (1) {}
}
void verifyint(void) {
//  gbaprintf("%s\n",__PRETTY_FUNCTION__);
//  while (1) {}
}

#define COMMANDS 8
const fdc_comm_t disc_commands[COMMANDS+1] =
{
        {0x0b,3,0x3F,write_varlen,writeint},
        {0x13,3,0x3F,read_varlen,readint},
        {0x1f,3,0x3F,verify_varlen,verifyint},
        {0x29,1,0x3F,seek,seekint},
        {0x2C,0,0x3F,read_drive_status,0},
        {0x35,4,0xFF,specify,0},
        {0x3A,2,0x3F,write_special,0},
        {0x3D,1,0x3F,read_special,0},
        {0x00,0,0x00,0,0}
};


const fdc_comm_t * get_command(int val) {
  const fdc_comm_t * p = disc_commands;
  for (; p->command != (val & p->mask); p++);
  return p;
}

void command_write(int val) {
  current_command = get_command(val);
  n_params = current_command->params;
  command_val = val;
  current_param = 0;

  status |= 0x90;
  updateNMI();

  if (n_params == 0) {
    status &= 0x7e;
    updateNMI();
    current_command->func();
  };
}

void breakhandler(void) {}

// address = 0xf8n&0xf
int do8271read(int address) {
  int val = 0;
  switch ( address ) {
    case 0 :
      val = status;
      break;
    case 1 :
      status &= ~0x18;
      updateNMI();
      val = result;
      result = 0;
      //gbaprintf("result read %x\n",val);
      break;
    case 4:
      status &= ~0xc;
      updateNMI();
      val = datareg;
      break;
    default : 
      //gbaprintf("unknown disc read %x\n",address);
      break;
  }
  return val;
}

// address is 0xfe8n & 0xf
void do8271write(int val, int address) {
  switch (address) {
    case 0 : 
       // command reg write
       //gbaprintf("comm write %x\n",val);
       command_write(val);
       break;
    case 1 :
       // parameter write
       command_params[current_param++] = val;
       status &= 0xfe;
       if (current_param == n_params) {
         status &= 0x7e;
         updateNMI();
         current_command->func();
       } else {
         updateNMI();
       }
       break;
    case 2:
       // disk reset
       //gbaprintf("disc reset \n",val);
       reset_disc_drive();
       break;
    case 4:
       status &= ~0xc;
       // update nmi
       updateNMI();
       datareg = val;
       //gbaprintf("set datareg %x \n",val);
       break;

    default :
       break;
  }
}

