#include "gba.h"
#include "bbcdisc.h"
#include "agbprint.h"
#include "gfx.h"


int disc_trigger = 0;
int NMIintStatus = 0;
u32 status1770 = 0;
u32 track      = 0; // the actual track
u32 trackReg   = 0; // the track we see externally
u32 sector     = 0;
u32 data1770   = 0;
u32 headDirection   = 0;
u32 headPosition   = 0;

#define BYTES_PER_SECTOR 256
#define SECTORS_PER_TRACK 10
#define BYTES_PER_TRACK (BYTES_PER_SECTOR*SECTORS_PER_TRACK)

u32 current_byte = 0; // absolute offset into disc image

#define NOCMD          0 
#define RESTORE        1 
#define SEEK           2 
#define STEP           3  
#define STEP_IN        4
#define STEP_OUT       5
#define READING        6
#define WRITING        7
#define READ_SECTOR    8
#define WRITE_SECTOR   9
#define SPIN_UP_DELAY  11
#define SPIN_DOWN_DELAY  12
#define CONFUSED       13
#define READ_ADDRESS   14
#define HEAD_MOVED     15

u32 command1770 = NOCMD;
u32 nextCommand = NOCMD;
u32 updateTrack = 0;
u32 multiSector = 0;
u32 byteCount = 0;

//#define DISC _binary_Elite_A_ssd_start
//extern unsigned char DISC[];
#define DISC 0

unsigned char * discs  = DISC;

#define SetStatus(n) status1770 |= (1<<n)

#define ResetStatus(n) status1770 &= ~(1<<n)

#define ClearNMI() NMIintStatus &= ~(8|16)
#define SetNMI()   NMIintStatus |= (8)

int do1770read(u32 reg) {
  int val = 0;
  switch (reg) {
    case 0:
      ClearNMI();
      val = status1770;
      break;
    case 1:
      val = trackReg;
      break;
    case 2:
      val = sector;
      break;
    case 3:
      {
	if (command1770 > STEP_OUT) {
	  ResetStatus(1);
	  ClearNMI();
	}
      }
      val = data1770;
      break;
    default:
      break;

  }
  return val;
}

void do1770write(u32 value, u32 reg) {
  if (reg == 0) {
    // the biggy
    ClearNMI();
    int com  = value & 0xf0;
    int hcom = value & 0xe0;
    //gbaprintf("com %x hcom %x \n",com, hcom);
    int bit4set = value & 0x10;
    if (hcom < 0x80) {
      // seek command
      SetStatus(0);
      ResetStatus(3);
      ResetStatus(4);
      switch (hcom) {
	case 0x60: { command1770 = STEP_OUT; updateTrack = bit4set; } break;
	case 0x40: { command1770 = STEP_IN; updateTrack = bit4set; } break;
	case 0x20: { command1770 = STEP; updateTrack = bit4set; } break; 
	default: break;
      }
      if (com & 0x10) command1770 = SEEK;
      if (!com) command1770 = RESTORE;

      if (command1770 < READING) {
	ResetStatus(5);
	SetStatus(0);
	// timing, add spin up delay
	//if (!(status1770 & 0x80)) {
	//  SetStatus(7);
	//}
	// do not density check
	disc_trigger = 160;
      }

    } else if (hcom == 0x80) {
      // read a sector
      SetStatus(0);
      ResetStatus(1);
      command1770 = READ_SECTOR;
      multiSector = bit4set;
    } else if (hcom == 0xa0) {
      SetStatus(0);
      ResetStatus(1);
      command1770 = WRITE_SECTOR;
      multiSector = bit4set;
    }
    if (com == 0xd0) {
      if (command1770) {
	ResetStatus(0);
      } else {
	ResetStatus(0);
	ResetStatus(1);
	ResetStatus(3);
	ResetStatus(4);
	ResetStatus(5);
      }
      command1770 = SPIN_DOWN_DELAY;
      disc_trigger = 160;
      data1770 = 0;
      if ( (value & 0xf) ) {
	//gbaprintf("Setting NMI\n");
	SetNMI();
      }
    } else if (com == 0xc0) {
      // read address
      command1770 = READ_ADDRESS;
      SetStatus(0);
      byteCount = 6;
      disc_trigger = 160;
    }
    if (command1770 == READ_SECTOR || command1770 == WRITE_SECTOR) {
      disc_trigger = 160;
      printTxtPal("loading...", 20,19,1);

    }

  } else if (reg == 1) {
    track = value;
    trackReg = value;
  } else if (reg == 2) {
    sector = value;
  } else if (reg == 3) {
    data1770 = value;
    if (command1770 > 5) {
      // clear interrupts
      ResetStatus(1);
      ClearNMI();
    }
  }
}
// in arm asm...
void poll1770(void);

void reset_disc_drive(void) {
  status1770 = 0;
  NMIintStatus = 0;
}
