#include "bbcmemory.h"
#include "sysvia.h"
#include "uservia.h"
#include "fastvideo.h"

#include "savestate.h"
#include "6502core.h"
#include "gfx.h"

#include <stdlib.h>
#include <string.h>

unsigned char * WholeRam = (unsigned char *)0x2000000;

int PagedRomReg = 0x0;
#define OS_ROM _binary_os12_rom_start
extern unsigned char OS_ROM[];

#define BASIC_ROM _binary_basic_rom_start
extern unsigned char BASIC_ROM[];

#if defined(USE_1770_DISC)
#define DISK_ROM _binary_ddfs_1_54t_rom_start
#else
#define DISK_ROM _binary_DFS_0_9_rom_start
#endif
extern unsigned char DISK_ROM[];

const char * const ROM_BANKS[16] =  {
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  DISK_ROM,
  BASIC_ROM,
};

void changeRomBank(int newBank, int address) {
  PagedRomReg = newBank&0xf;
  if (ROM_BANKS[PagedRomReg] != 0) {
    memcpy(WholeRam+0x8000, ROM_BANKS[PagedRomReg], 0x4000);
  }
}


// memory map:
// 0-8000 = main ram (32k)
// 8000-c000 = Basic rom (16k?)
// c000-fc00 = OS rom 
// fc00-ff00 = IO space (ULA, VIA, CRTC)
void BeebReadRoms(void) {
  
  PagedRomReg=0xf;  // basic rom
  // load the os rom into main mem...
  memcpy(WholeRam+0xc000, OS_ROM, 0x4000);
  // load basic rom into mem too
  memcpy(WholeRam+0x8000, ROM_BANKS[PagedRomReg], 0x4000);
  // load other roms too, if needed...
}

typedef void (*LineRoutinePtr)(void);

extern LineRoutinePtr scanlineFunction;
extern unsigned long cyclesPerScanline;
void initCore(void) {
    T6502State *state;
    state = (T6502State*)&Asm6502State;

    state->programCounter = WholeRam[0xfffc]|(WholeRam[0xfffd]<<8); 
    state->accumulator=0;
    state->XReg=0;
    state->YReg=0;
    stack=(0xff)|0x100;
    state->PSR=4;
    intStatus=0;
    
    cyclesPerScanline=128;
    scanlineFunction = (VideoULA_ControlReg & 0x10)?lowLevelDoScanLineNarrow:
                                                  lowLevelDoScanLineWide;

    initColour();
}

void BeebMemInit(void) {
  memset(WholeRam ,0,65536);
  BeebReadRoms();
}


void LoadMainMemUEF(const unsigned char * const SUEF) {
  int i;
  for (i = 0; i < 32768; i++) {
    WholeRam[i] = SUEF[i];
  }
}

