#define rstat     r9
#define rcommand  r10

#define NOCMD          0 
#define RESTORE        1 
#define SEEK           2 
#define STEP           3  
#define STEP_IN        4
#define STEP_OUT       5
#define READING        6
#define WRITING        7
#define READ_SECTOR    8
#define WRITE_SECTOR   9
#define SPIN_UP_DELAY  11
#define SPIN_DOWN_DELAY  12
#define CONFUSED       13
#define READ_ADDRESS   14
#define HEAD_MOVED     15

#define BIT(n)  (1<<n)

#define BYTES_PER_SECTOR 256
#define SECTORS_PER_TRACK 10
#define BYTES_PER_TRACK (BYTES_PER_SECTOR*SECTORS_PER_TRACK)

@ 1770 disc emulation in arm
.section .iwram,"ax",%progbits
.arm

.macro setDiscTrigger cycs
  ldr r4,=disc_trigger
  mov r5,#\cycs
  str r5,[r4]
.endm

.macro  SetStatus bits
@    ldr    r5,[r4]
    orr    rstat,rstat,#\bits
    ldr    r4,=status1770
    str    rstat,[r4]
.endm

.macro  ResetStatus bits
@    ldr    r5,[r4]
    bic    rstat,rstat,#\bits
    ldr    r4,=status1770
    str    rstat,[r4]
.endm

.macro  clearNMI
    ldr    r4,=NMIintStatus
    ldr    r5,[r4]
    bic    r5,r5,#(8|16)
    str    r5,[r4]
.endm
.macro  setNMI
    ldr    r4,=NMIintStatus
    ldr    r5,[r4]
    orr    r5,r5,#(8)
    str    r5,[r4]
.endm

#if 1
.global poll1770
.type poll1770, function

poll1770:
  ldr    r0,=status1770
  ldr    rstat,[r0]
  ldr    r0,=command1770
  ldr    rcommand,[r0]
  tst    rstat,#1
  beq    spinDown
  ldr    r0,=NMIintStatus
  ldr    r1,[r0]
  tst    r1,#16
  bne    spinDown
  @ if ( (status1770 & 1) && !(NMIintStatus & 16)) {
  cmp  rcommand,#READING 
  bmi  ltReading
gteReading:
  @ command >= READING
  bne  notReading
  @ so rcommand == READING
  @ skip cycles check
  setDiscTrigger 160
  tst  rstat,#2
  movne pc,lr
@    if ( (status1770 & 2) == 0) {
@      ResetStatus(5); ResetStatus(4); ResetStatus(3); ResetStatus(2);
  ResetStatus ( BIT(5) | BIT(4) | BIT(3) | BIT(2)  )
  ldr  r2, =data1770 
  ldr  r3, =discs
  ldr  r3, [r3]
  ldr  r8, =current_byte
  ldr  r5, [r8]
  ldrb r3, [r3,r5]
  add  r5,r5,#1
  str  r5, [r8]        @ store current_byte
  str  r3, [r2]        @ store data1770
  @data1770 = discs[current_byte++];
  SetStatus (BIT(1))
  SetNMI
  ldr  r3, =byteCount
  ldr  r2,[r3]
  subs r2,r2,#1
  str  r2, [r3]       @ store byteCount--
  @ byteCount--;
  @    if (byteCount == 0 && (!multiSector) ) {
  movne pc,lr             @ byteCount != 0
  ldr  r3, =multiSector
  ldr  r2,[r3]
  cmp  r2,#0
  movne pc,lr             @ multiSector != 0
  ResetStatus (BIT(0)) 
@	ResetStatus(0);
@	current_byte = headPosition;
  ldr    r3, =headPosition
  ldr    r2,[r3]
  str    r2,[r8]      @ current_byte = headPosition
  ldr    r0,=command1770
@  command1770 = 10; 
  mov    rcommand, #10
  str    rcommand, [r0]
  mov pc,lr
  
notReading:
  cmp   rcommand,#HEAD_MOVED
  bne   notHeadMove
@ is a head move operation
@ if (command1770 == HEAD_MOVED) 
  setDiscTrigger 160
@ disc_trigger = 160;
 @   command1770 = SPIN_DOWN_DELAY;
  mov   rcommand,#SPIN_DOWN_DELAY
  ldr   r0,=command1770
  str   rcommand,[r0]
  SetNMI
  @ assume always watford, which inverts track 0
  ldr      r3,=track
  ldr      r3,[r3] 
  cmp      r3,#0
@  if (track != 0)
@    ResetStatus(2);
@  else
@    SetStatus(2);
  bicne    rstat,rstat,#BIT(2)
  orreq    rstat,rstat,#BIT(2)
  @    ResetStatus(5); ResetStatus(0);
  ResetStatus (BIT(5) | BIT(0))  
  mov pc,lr  @ end head move

notHeadMove:
  cmp  rcommand,#WRITING
  beq  writing
  @ not writing, what about read/write sector?
  cmp rcommand,#READ_SECTOR
  beq sectorReadWrite 
  cmp rcommand,#WRITE_SECTOR
  movne pc,lr     @ goodness knows what it is, bail out
sectorReadWrite:
   @if (command1770 == READ_SECTOR || command1770 == WRITE_SECTOR ) {
   @   disc_trigger = 160;
   setDiscTrigger 160
   @    SetStatus(0);
   SetStatus (BIT(0))
   ResetStatus (BIT(5) | BIT(6) | BIT(2) )
   ldr  r3, =byteCount
   mov  r2,#0x100   @ 256
   add  r2,r2,#1    @ 257
   @   byteCount = 257;
   str  r2, [r3]       @ store byteCount

   @    headPosition = current_byte;
   ldr    r3, =headPosition
   ldr    r8, =current_byte 
   ldr    r2,[r8]
   str    r2,[r3]      

   @    current_byte = BYTES_PER_TRACK * track + BYTES_PER_SECTOR * sector;
   ldr    r2, =track
   ldr    r3, =sector
   ldr    r2,[r2]    @ r2 = track
   ldr    r3,[r3]    @ r3 = sector
   mov    r0,#BYTES_PER_TRACK 
   mul    r1,r2,r0   @ r1 = track * BYTES_PER_TRACK
   add    r1,r1,r3,lsl #8  @ r1 = r1 + sector * 256
   str    r1,[r8]    @ current_byte = r1 = t*BPT + s*BPS

   
@      if (command1770 == READ_SECTOR) {
@	command1770 = READING;
   ldr    r0,=command1770
   cmp    rcommand, #READ_SECTOR
   moveq  rcommand, #READING
   streq  rcommand,[r0]
   moveq  pc,lr
   @ else must be write... check just in case 
   cmp    rcommand, #WRITE_SECTOR
   movne  pc,lr
   mov  rcommand, #WRITING
   str  rcommand,[r0]
   SetStatus (BIT(1))
   SetNMI 
@      } else if (command1770 == WRITE_SECTOR) {
@	command1770 = WRITING;
@	SetStatus(1);
@	SetNMI();
@      }
  mov pc,lr @ end sector read write
 
writing: 
  @ if (command1770 == WRITING) {
    @ none of that!
    
    ResetStatus (BIT(0))
    SetStatus (BIT(6))
    mov  rcommand,#0
    ldr r0,=command1770
    str rcommand,[r0]
    @ command1770 = NOCMD;
    SetNMI
    mov pc,lr @ end write
  
ltReading:
   @ if (command1770 < READING) {
      @ seek etc
      ResetStatus (BIT(4)| BIT(3) )
      @ branch table
#define  rupdateAddr  r1
#define  rupdate      r2
#define  rcurrbAddr  r3
#define  rcurrb      r4
#define  rtrackAddr  r5
#define  rtrack      r6

      ldr   rupdateAddr,=updateTrack
      ldr   rupdate,[rupdateAddr]
      ldr   rcurrbAddr,=current_byte
      ldr   rcurrb,[rcurrbAddr]
      ldr   rtrackAddr,=track
      ldr   rtrack,[rtrackAddr]

      adr     r0,LseekSwitch
      ldr     pc,[r0,rcommand,lsl#2]     @ 4 bytes per symbol
caseRESTORE : 
          
	  @updateTrack = 1;
          mov   rupdate,#1
	  @current_byte = 0;
          mov   rcurrb,#0
	  @track = 0;
          mov   rtrack,#0
	  @break;
          b 1f

caseSEEK : 
	  @ seek to 10 * sector length * data register
	  @ updateTrack = 1;
          mov   rupdate,#1
	  
          @ current_byte = BYTES_PER_TRACK * data1770;
	  mov    r0,#BYTES_PER_TRACK 
          ldr    r7,=data1770
          ldr    r7,[r7]
	  mul    rcurrb,r7,r0   @ r1 = track * BYTES_PER_TRACK
	  @ track = data1770;
          mov    rtrack,r7
	  @break;
          b 1f

caseSTEP : 
	  @ keep moving in the same direction
	  @ updateTrack = 1;
          mov rupdate,#1
	  @current_byte = current_byte + ( (headDirection)?BYTES_PER_TRACK:-BYTES_PER_TRACK);
          ldr r7,=headDirection
          ldr r7,[r7]
          cmp r7,#0
          addne rcurrb,rcurrb,#BYTES_PER_TRACK 
          subeq rcurrb,rcurrb,#BYTES_PER_TRACK 
	  @track = track + ( (headDirection)?1:-1);
          addne rtrack,rtrack,#1
          subeq rtrack,rtrack,#1
           
	  @break;
          b 1f

caseSTEP_IN : 
	  @ headDirection = 1;
          ldr r7,=headDirection
          mov rcommand,#1
          str rcommand,[r7]
          
	  @ current_byte += BYTES_PER_TRACK;
          add rcurrb,rcurrb,#BYTES_PER_TRACK
	  @ track++;
          add rtrack,rtrack,#1
	  @ break;
          b 1f

caseSTEP_OUT : 
	  @ headDirection = 0;
          ldr r7,=headDirection
          mov rcommand,#0
          str rcommand,[r7]
	  @current_byte -= BYTES_PER_TRACK;
          sub rcurrb,rcurrb,#BYTES_PER_TRACK
	  @track--;
          sub rtrack,rtrack,#1
	  @break;
	  b 1f
defaultCase:
          mov pc,lr

1:
      @ if (updateTrack) {
      @   trackReg = track;
      @}
      cmp  rupdate,#0
      ldrne  r7,=trackReg
      strne    rtrack, [r7]  
      str    rtrack, [rtrackAddr]  
      str    rcurrb, [rcurrbAddr]  
      
      @ updateTrack = 0;
      mov  rupdate,#0
      str  rupdate,[rupdateAddr]
      mov rcommand,#HEAD_MOVED
      @ command1770 = HEAD_MOVED;
      ldr   r0,=command1770
      str  rcommand,[r0]
      
      @ timing? who wants that...
      @disc_trigger = 160;
      setDiscTrigger 160

      mov pc,lr


LseekSwitch:
.word defaultCase
.word caseRESTORE
.word caseSEEK
.word caseSTEP
.word caseSTEP_IN
.word caseSTEP_OUT


@  } else {
spinDown:
@ at this point rcommand may be either SPIN change or an address read
  cmp rcommand,#READ_ADDRESS
  beq readAddress
    @if (command1770 == SPIN_UP_DELAY || command1770 == SPIN_DOWN_DELAY) {
      @ command1770 = NOCMD;
      mov rcommand,#NOCMD
      ldr r0,=command1770
      str rcommand,[r0]
     
   @   printTxtPal("          ", 20,19,1);
@   stmfd sp!,{r1-r11,lr}
   mov   r0,#19
@   adr   lr,0f
   ldr   r1,=clearTextRow
   bx    r1
@0:
@   ldmfd sp!,{r1-r11,lr}
@   mov pc,lr
    @}
readAddress:

@    if (command1770 == READ_ADDRESS) {
     ldr r0,=byteCount
     ldr r1,[r0]
     sub r2,r1,#1
     str r2,[r0]
     adr r0,LbyteCountSwitch
     cmp r0,#7
     movpl pc,lr
     ldr r2,=data1770
     ldr     pc,[r0,r1,lsl#2]     @ 4 bytes per symbol
  
      @switch (byteCount) {
caseCount0:
	  ResetStatus (BIT(0))
	  @ command1770 = NOCMD;
	  mov rcommand,#NOCMD
	  @ break;
          ldr  r0,=command1770
          str  rcommand,[r0] 
	  SetNMI
	  mov pc,lr
          
caseCount1:
	  @ data1770 = 0;
          mov   r0,#0
          str   r0,[r2]
	  SetNMI
	  mov pc,lr
caseCount2:
	  @ data1770 = 0;
          mov   r0,#0
          str   r0,[r2]
	  SetNMI
	  mov pc,lr
caseCount3:
	  @ data1770 = 1;
          mov   r0,#1
          str   r0,[r2]
	  SetNMI
	  mov pc,lr
caseCount4:
	  @ data1770 = 1;
          mov   r0,#1
          str   r0,[r2]
	  SetNMI
	  mov pc,lr
caseCount5:
	  @ data1770 = headPosition;
          ldr   r3,=headPosition
          ldr   r0,[r3]
          str   r0,[r2]
	  SetNMI
	  mov pc,lr
caseCount6:
	  @ data1770 = track;
          ldr   r3,=track
          ldr   r0,[r3]
          mov   r0,#0
          str   r0,[r2]
	  SetNMI
	  mov pc,lr
@	default: break;
@      byteCount--;

LbyteCountSwitch:
.word caseCount0
.word caseCount1
.word caseCount2
.word caseCount3
.word caseCount4
.word caseCount5
.word caseCount6

#endif
.pool

.end

