#include "fastvideo.h"
#include "bbcmemory.h"
#include "gfx.h"
#include "screenmode.h"

typedef struct {
  int Addr;
  int PixmapLine;
  int PreviousFinalPixmapLine;
  int IsTeletext;
  char *DataPtr;

  int CharLine;
  int InCharLine;
  int InCharLineUp;
  int VSyncState;
} VideoStateT;

VideoStateT VideoState = {};

u8 FastTable[256] = {};   // 256 * 1 bytes = 256
TwoUChars FastTableDWidth[256] = {};   // 256 * 2 bytes = 512
FourUChars FastTable1bpp[256] = {};   // 256 * 4 bytes = 1024

// VideoULA register controls "stuff"
/*
  BIT   7 6 5 4 3 2 1 0
        | | | |     | +-  if set, compliment (reverse) palette index
        | | | |     +---  if set, is in Teletext mode
        | | | +---------  set = fast clock pixels... also determines cycles taken per scan line
        | | +-----------  \__ kind of "colours used"...
        | +-------------  /  
        +---------------  Not set == Mode 7 (?)
*/
unsigned char VideoULA_ControlReg=0x9c;


unsigned char VideoULA_Palette[16] = {};

CRTC_t CRTCControl = 
{
	0,63,40,51,36,  30,2,25,28,147,  7,114,19,40,0,41,35,0,0
};

typedef void (*LineRoutinePtr)(void);
extern LineRoutinePtr scanlineFunction;

void makeFastTable(void) {
  unsigned int beebpixvl,beebpixvr;
  unsigned int bplvopen,bplvtotal;
  unsigned char tmp;
  for(beebpixvl=0;beebpixvl<16;beebpixvl++) {
    bplvopen=((beebpixvl & 8)?128:0) |
             ((beebpixvl & 4)?32:0) |
             ((beebpixvl & 2)?8:0) |
             ((beebpixvl & 1)?2:0);
    for(beebpixvr=0;beebpixvr<16;beebpixvr++) {
      bplvtotal=bplvopen |
             ((beebpixvr & 8)?64:0) |
             ((beebpixvr & 4)?16:0) |
             ((beebpixvr & 2)?4:0) |
             ((beebpixvr & 1)?1:0);
      tmp=VideoULA_Palette[beebpixvl];
      tmp=beebpixvl;
      FastTable[bplvtotal] = tmp;
      
      tmp= VideoULA_Palette[beebpixvr];
      tmp= beebpixvr;
      FastTable[bplvtotal] |= (tmp<<4);
      
    }
  }
}

void DoFastTable4XStep4(void) {
  unsigned char tmp;
  unsigned long beebpixv,pentry;

  for(beebpixv=0;beebpixv<256;beebpixv++) {
    pentry=((beebpixv & 128)?8:0)
           | ((beebpixv & 32)?4:0)
           | ((beebpixv & 8)?2:0)
           | ((beebpixv & 2)?1:0);
    tmp= pentry;//VideoULA_Palette[pentry];
    // this is really 2 pixels when we double width our x
    FastTableDWidth[beebpixv].data[0] = tmp; //
    
    
    pentry=((beebpixv & 64)?8:0)
           | ((beebpixv & 16)?4:0)
           | ((beebpixv & 4)?2:0)
           | ((beebpixv & 1)?1:0);
    tmp= pentry ;//VideoULA_Palette[pentry];
    // we are really up to 4 pixels double widthing our x
    FastTableDWidth[beebpixv].data[0] |= (tmp<<4);

    pentry=((beebpixv & 32)?8:0)
           | ((beebpixv & 8)?4:0)
           | ((beebpixv & 2)?2:0)
           | 1;
    tmp= pentry; //VideoULA_Palette[pentry];
    // 6 pixels!
    FastTableDWidth[beebpixv].data[1] = tmp;//|(tmp<<4);
    
    pentry=((beebpixv & 16)?8:0)
           | ((beebpixv & 4)?4:0)
           | ((beebpixv & 1)?2:0)
           | 1;
    tmp= pentry;//VideoULA_Palette[pentry];
    // 8 pixels!
    FastTableDWidth[beebpixv].data[1] |= (tmp<<4);
    
  }
}


void DoFastTable2(void) {
  unsigned char tmp;
  unsigned long beebpixv,beebpixvt,pentry;
  int pix;

  for(beebpixv=0;beebpixv<256;beebpixv++) {
    // temp
    beebpixvt=beebpixv;

    // each loop fills up 8 bits (2 pixels)
    // 2 loops needed for a 16 bit VRAM entry (4 pixels) 
    // 4 loops needed for a 32 bit value (2 VRAM entries, 8 pixels)
    for(pix=0;pix<4;pix++) {
      pentry=((beebpixvt & 128)?8:0)
             | ((beebpixvt & 32)?4:0)
             | ((beebpixvt & 8)?2:0)
             | ((beebpixvt & 2)?1:0);
      beebpixvt *= 2;
      beebpixvt |=1;
      
      tmp=pentry;
      
      // need to squash 2 pixels per entry
      FastTable1bpp[beebpixv].data[pix] = tmp; 
      pentry=((beebpixvt & 128)?8:0)
             | ((beebpixvt & 32)?4:0)
             | ((beebpixvt & 8)?2:0)
             | ((beebpixvt & 2)?1:0);
      beebpixvt *= 2;
      beebpixvt |=1;
      
      tmp=pentry;
      
      // need to squash 2 pixels per entry
      FastTable1bpp[beebpixv].data[pix] |= (tmp<<4); 
    }
  }
}

extern unsigned long cyclesPerScanline;
void LoadVideoUEF(const unsigned char * StateData) {
  int i = 0;
  CRTCControl.CRTC_HorizontalTotal=StateData[i++];
  CRTCControl.CRTC_HorizontalDisplayed=StateData[i++];
  CRTCControl.CRTC_HorizontalSyncPos=StateData[i++];
  CRTCControl.CRTC_SyncWidth=StateData[i++];
  CRTCControl.CRTC_VerticalTotal=StateData[i++];
  CRTCControl.CRTC_VerticalTotalAdjust=StateData[i++];
  CRTCControl.CRTC_VerticalDisplayed=StateData[i++];
  CRTCControl.CRTC_VerticalSyncPos=StateData[i++];
  CRTCControl.CRTC_InterlaceAndDelay=StateData[i++];
  CRTCControl.CRTC_ScanLinesPerChar=StateData[i++];
  CRTCControl.CRTC_CursorStart=StateData[i++];
  CRTCControl.CRTC_CursorEnd=StateData[i++];
  CRTCControl.CRTC_ScreenStartHigh=StateData[i++];
  CRTCControl.CRTC_ScreenStartLow=StateData[i++];
  CRTCControl.CRTC_CursorPosHigh=StateData[i++];
  CRTCControl.CRTC_CursorPosLow=StateData[i++];
  CRTCControl.CRTC_LightPenHigh=StateData[i++];
  CRTCControl.CRTC_LightPenLow=StateData[i++];
  VideoULA_ControlReg=StateData[i++];
  
  scanlineFunction = (VideoULA_ControlReg & 0x10)?lowLevelDoScanLineNarrow:
                                                  lowLevelDoScanLineWide;

  cyclesPerScanline = (CRTCControl.CRTC_HorizontalTotal+1)*((VideoULA_ControlReg & 16)?1:2);

  int col;
  for (col = 0; col < 16; ++col) {
    VideoULA_Palette[col]=StateData[i++]^7;
  }
  // update the video with these values
  updateColours();
  
  i+=2;
  i+=4;
  CRTCControl.CRTCControlReg=StateData[i++];
  i++;
	extern int haveClearedTT;
  if (VideoULA_ControlReg & 2) 
		haveClearedTT=0; 
  VideoInit();
}

void clearTeletext() {
  
  int x,y;
  for (y = 0; y < 32 && y <CRTCControl.CRTC_VerticalDisplayed; y++) {
    for (x = 0; x < 32 && x <CRTCControl.CRTC_HorizontalDisplayed; x++) {
      // print a row of text
      bg_teletextfg.mapData[x+y*32] = 0;
      bg_teletextbg.mapData[x+y*32] = 0;
    }
  }
  REG_DISPCNT |= OBJ_ENABLE;
  cyclesPerScanline = (CRTCControl.CRTC_HorizontalTotal+1)*((VideoULA_ControlReg & 16)?1:2);
}

void VideoInit(void) {
  VideoState.DataPtr= WholeRam+0x3000;

  makeFastTable();
  DoFastTable4XStep4();
  DoFastTable2();

  VideoState.PixmapLine=0;
  VideoState.PreviousFinalPixmapLine=255;
  
  VideoState.IsTeletext = ((VideoULA_ControlReg&2)!=0);

  // a reasonable default value...
  cyclesPerScanline = 128;
  
  copyColours = COPY_COLOURS;
  
  if (!VideoState.IsTeletext)
    clearTeletext();
}

