#include <string.h>
#include "gba.h"
#include "sprite.h"
#include "dma.h"
#include "screenmode.h"
//#include "qmath.h"
#include "bios_div.h"

#include "virtualkb.h"
#include "inter.h"
#include "fastvideo.h"
#include "gfx.h"
#include "stdlib.h"

#define FONT_COL_INDEX 5

OAMEntry sprites[128] = {};
//OAMEntry * sprites = (OAMEntry*)0x7000000;

#define NULL_BG {}

Bg bg_teletextfg = NULL_BG;
Bg bg_teletextbg = NULL_BG;
Bg bg_menu = NULL_BG;


#define get_oam64(x, y)  (x*64 + y * 32 * 64)
#define get_oam64_half(x, y)  (x*32 + y * 32 * 32)

void init_sprite64(int x, int y, int priority) {
  sprites[x+y*4].attribute0 = ROTATION_FLAG|SQUARE|COLOR_16|SIZE_DOUBLE; 
  sprites[x+y*4].attribute1 = SIZE_64|ROTDATA(0); 
  sprites[x+y*4].attribute2 = PRIORITY(priority) | (get_oam64(x,y)/8);
}

void init_sprite64_half(int x, int y, int priority) {
  sprites[x+y*4].attribute0 = ROTATION_FLAG|SQUARE|COLOR_16; 
  sprites[x+y*4].attribute1 = SIZE_64|ROTDATA(0); 
  sprites[x+y*4].attribute2 = PRIORITY(priority) | (get_oam64(x,y)/8);
}


unsigned char fadedup = 0;
void load_font(u8 * fontGFX);
void load_text_gfx();

#if 0
// clears out the video memory, which may have been set already by the boot menu
void clear_vram(void) {
  u16 i;
  u32 zero = 0;
  // fade down
  /*
  REG_BLDMOD = BLDMOD_FIRST_BG1 | BLDMOD_FIRST_BG0 | BLDMOD_FIRST_BG2 |
               BLDMOD_FIRST_BG3 |
               BLDMOD_FIRST_OBJ | BLDMOD_SECOND_BD | BLDMOD_MODE(3);
  REG_COLEY = 16;
  */
  for (i = 0; i < 256; i++)
    BGPaletteMem[i] = 0;

  REG_DMA3SAD = (u32)&zero;
  REG_DMA3DAD = (u32)&(VideoBuffer[0]); // VRAM
  REG_DMA3CNT = 0x10000 |DMA_SOURCE_FIXED|DMA_16NOW; 
}
#endif


#define COLVAL 31 
const u32 colours[8] = {
RGB16(0,0,0),
RGB16(COLVAL,0,0),
RGB16(0,COLVAL,0),
RGB16(COLVAL,COLVAL,0),
RGB16(0,0,COLVAL),
RGB16(COLVAL,0,COLVAL),
RGB16(0,COLVAL,COLVAL),
RGB16(COLVAL,COLVAL,COLVAL),

};

#define OS_ROM _binary_os12_rom_start
//#define OS_ROM _binary_font8pt_raw_start
extern unsigned char OS_ROM[];
extern unsigned short palBuf[];
void updateColours() {
  u32 zero = 0;
  REG_DMA3SAD = (u32)&zero;
  REG_DMA3DAD = (u32)&(OAMData[0]); // tile data
  REG_DMA3CNT = (0x4000) |DMA_SOURCE_FIXED|DMA_16NOW; // 16 bit transfers from gfx array via DMA
  int i;
  for (i = 0; i < 16; i++) {
    OBJPaletteMem[i] = colours[VideoULA_Palette[i]];
    palBuf[i] = colours[VideoULA_Palette[i]];
  }
  BGPaletteMem[0] = colours[VideoULA_Palette[0]];
  
}

void get_pal_indices(unsigned char * p) {

  int i, col;
  for (i = 0; i < 16; i++) {
    for (col = 0; col < 8; col++) {
      if (palBuf[i] == colours[col]) {
	p[i] = col^7;
	break;
      }
    }
  }
  
}

void initColour(void) {
  int i;
  for (i = 0; i < 16; i++) {
    OBJPaletteMem[i] = colours[i];
  }
  BGPaletteMem[0] = colours[0];
}

#if 0
void initBackground(Bg * bg, int number,
                    int charBlock, int screenBlock, 
                    int priority) 
{
  bg->number = number;
  bg->charBaseBlock = charBlock;
  bg->screenBaseBlock = screenBlock;
  bg->colorMode = BG_COLOR_16;
  bg->size = TEXTBG_SIZE_256x256; 
  // not really mosaic, but priority setting. This could be called 
  // "flags" really, as it's all just ORed together with the size and colour.
  bg->mosiac = priority;
 // bg->y_scroll = (24) &0xff;
  EnableBackground(bg);
  UpdateBackground(bg);
}
#endif
void initBG(void) {
  
  initBackground(&bg_menu      ,  0, 0, 30, 0);
  initBackground(&bg_teletextbg,  2, 1, 22, 3);
  initBackground(&bg_teletextfg,  1, 1, 25, 1);
  
  int i,x,y;
  for (i = 0; i < 16; i++) {
    OBJPaletteMem[i]= colours[i&7];
    BGPaletteMem[i*16+FONT_COL_INDEX] = colours[i];
  }
  
  setSpriteScreen(128,256);
  autoScale(); 
  for (y=0; y < 20; y++) {
    for (x=0; x < 30; x++) {
      bg_teletextfg.mapData[x+y*32] = 0;
      bg_teletextbg.mapData[x+y*32] = 0;
      bg_menu.mapData[x+y*32] = 0;
    }
  }
  
  load_font(OS_ROM);
  load_text_gfx();
  load_vk_tiles();
}



void fade_up(void) {
  u8 loop;
  
  REG_BLDMOD = BLDMOD_FIRST_BG1 | BLDMOD_FIRST_BG0 | BLDMOD_FIRST_BG2 |
               BLDMOD_FIRST_BG3 |
               BLDMOD_FIRST_OBJ | BLDMOD_SECOND_BD | BLDMOD_MODE(3);
  
  for (loop = 0; loop < 17; loop+=2) {
    wait_vsync_int();
    REG_COLEY = (16-loop);
  }
  wait_vsync_int();
  REG_COLEY = 0;
  fadedup = 1;
}     

void fade_down(void) {
  u8 loop;
  
  REG_BLDMOD = BLDMOD_FIRST_BG1 | BLDMOD_FIRST_BG0 | BLDMOD_FIRST_BG2 |
               BLDMOD_FIRST_BG3 |
               BLDMOD_FIRST_OBJ | BLDMOD_SECOND_BD | BLDMOD_MODE(3);
  
  for (loop = 0; loop < 17; loop+=2) {
    wait_vsync_int();
    REG_COLEY = loop;
  }
  wait_vsync_int();
  REG_COLEY = 16;
  fadedup = 0;
}

#if 0
void CopyOAM(void)
{
  unsigned short int loop;
  unsigned short int* temp;
  temp = (unsigned short int*)sprites;
  for(loop = 0; loop < 32*4; loop++)
  {
  	OAM[loop] = temp[loop];
  }

}

void MoveSprite(OAMEntry* sp, int x, int y)
{
  if(x < 0) //if it is off the left corect
    x = 512 + x;
  if(y < 0) //if off the top corect
    y = 256 + y;

  /* because x is 9 bits, we need to clear out bits 10-15,
     so we dont mess up any rotation data */
  x &= 0x01FF;
  sp->attribute1 = sp->attribute1 & 0xFE00;  //clear the old x value
  sp->attribute1 = sp->attribute1 | x;
  y &= 0x00FF;
  sp->attribute0 = sp->attribute0 & 0xFF00;  //clear the old y value
  sp->attribute0 = sp->attribute0 | y;

}


void move_sprite(int id, int x, int y) {
  if (id < 128) {
    MoveSprite(&sprites[id], x, y);
    CopyOAM();
  }
}

void remove_sprite(int id) {
  if (id < 32) {
    sprites[id].attribute0 = SIZE_DOUBLE;
  }
  
  OAM[id*4] = SIZE_DOUBLE;
  
}
#endif

const unsigned short paperMasks[] = {
  (FONT_COL_INDEX),     //0x000F,   
  (FONT_COL_INDEX<<4),  //0x00F0,   
  (FONT_COL_INDEX<<8),  //0x0F00,   
  (FONT_COL_INDEX<<12)  //0xF000    
};

#define FONT_TILE 3
#define DOUBLE_HEIGHT_0  ((FONT_TILE*16) + (96*2*16))
#define DOUBLE_HEIGHT_1  ((FONT_TILE*16) + (96*3*16))
#if 1
void load_font(u8 * fontGFX) {
  // the font is stored in 1 bit per pixel format
  // expand it to 4bpp
  // each letter is 8*8, 4bpp = 8*4 pixels = 32 u8s per char
  
  int i, j, k, m;
  
  // amount to shift by
  int shift; 
  // four pixels, 4 bits each
  u16 pixelQuad; 
  // start index in tile ram
  u32 address = FONT_TILE*16;
  // offset in fontGfX array of the character
  int offset = 0;
  u8 theByte;
  u8 bitMask;
  for (j = 0; j < 96; j++) {
    //offset = j*8;
    // loop over characters
    for (i = 0; i < 8; i++) {
      // loop over rows
      theByte = fontGFX[offset++];
      if (theByte != 0) {
        // has data
        // 66
        for (k = 0; k < 2; k++) {
          // each row requires 2 writes to vram
          
          pixelQuad = (u16)0x0000;
          // create the u16 of 4 pixels:
          for (m = 0; m < 4; m++) {
            bitMask = 1;
            // pixel k*4+m on this row
            shift = k*4 + m;
            
            if (shift != 7)
              bitMask <<= (7-shift);
            
            if ( (theByte & bitMask) != 0) {
//              if (m == 0)
//                pixelQuad = FONT_COL_INDEX;
//              else
              pixelQuad |= paperMasks[m];//(FONT_COL_INDEX<<(m*4));
            }
          }
          if (i < 4) {
            bg_teletextfg.tileData[j*16 + i*4  + k + DOUBLE_HEIGHT_0 ] = pixelQuad;
            bg_teletextfg.tileData[j*16 + i*4+2+ k + DOUBLE_HEIGHT_0 ] = pixelQuad;
          }
          else {
            bg_teletextfg.tileData[j*16 + (i-4)*4  + k + DOUBLE_HEIGHT_1 ] = pixelQuad;
            bg_teletextfg.tileData[j*16 + (i-4)*4+2+ k + DOUBLE_HEIGHT_1 ] = pixelQuad;
          }
          bg_menu.tileData[address] = pixelQuad;
          bg_teletextfg.tileData[address++] = pixelQuad;
          
        }
      } else {
        // the byte is empty, draw a blank row
        for (m = 0; m < 2; m++) {
          bg_menu.tileData[address] = 0x0000;
          bg_teletextfg.tileData[address++] = 0x0000; // pixel quad of 00 00
        }
      }
    }
  }  
}
#else
void load_font(u8 * fontGFX) {
    REG_DMA3SAD = (u32)(fontGFX);
    REG_DMA3DAD = (u32)(&bg_teletextfg.tileData[FONT_TILE*16]);
    REG_DMA3CNT = (96*16) | DMA_16NOW; // 16 bit transfers from gfx array via DMA
    REG_DMA3SAD = (u32)(fontGFX);
    REG_DMA3DAD = (u32)(&bg_menu.tileData[FONT_TILE*16]);
    REG_DMA3CNT = (96*16) | DMA_16NOW; // 16 bit transfers from gfx array via DMA
}
#endif

// create the teletext extra font set...
void load_text_gfx(void) {
  // 96 glyphs
  int m7cc;

  u32 address = FONT_TILE*16 + 96*16 ; //+ 32*16;
  u32 * vidmem = (u32*)&bg_teletextfg.tileData[address];
  for (m7cc=0;m7cc<96;m7cc++) {
    // here's how it works: top two blocks: 1 & 2
    // middle two blocks: 4 & 8
    // bottom two blocks: 16 & 64
    // its only a grpahics character if bit 5 (32) is clear.
    if (!(m7cc & 32)) {
      int row1=0; int row2=0; int row3=0;
      // left block has a value of 4032, right 63 and both 4095
      // or for 4 bpp, 8 pixels wide = 8*4 = 32 bits 
      // left 4 pixels = 16 bits = ffff0000 = 
      // right 4 pixs  = 16 bits = 0000ffff = 
      if (m7cc & 1) row1|=0x00005555; // 111111000000
      if (m7cc & 2) row1|=0x55550000;   // 000000111111
      if (m7cc & 4) row2|=0x00005555; // 111111000000
      if (m7cc & 8) row2|=0x55550000;   // 000000111111
      if (m7cc & 16) row3|=0x00005555;// 111111000000
      if (m7cc & 64) row3|=0x55550000;  // 000000111111
      // now input these values into the array
      // top row of blocks
//      [m7cc][0]=[m7cc][1]=[m7cc][2]=row1;
//      [m7cc][3]=[m7cc][4]=[m7cc][5]=row1;
       *vidmem++ = row1;
       *vidmem++ = row1;
       *vidmem++ = row1;
       *vidmem++ = row2;
       *vidmem++ = row2;
       *vidmem++ = row3;
       *vidmem++ = row3;
       *vidmem++ = row3;
      // middle row of blocks
//      [m7cc][6]=[m7cc][7]=[m7cc][8]=row2;
//      [m7cc][9]=[m7cc][10]=[m7cc][11]=row2;
//      [m7cc][12]=[m7cc][13]=row2;
      // Bottom row
//      [m7cc][14]=[m7cc][15]=[m7cc][16]=row3;
//      [m7cc][17]=[m7cc][18]=[m7cc][19]=row3;
    } // check for valid char to modify
    else {
      vidmem+=8;
    }
  } // character loop.
}



// print character at row, col in text bg
// use palette pal
void print_text(char text, int x, int y, int pal) {
  if (text != 0x0) {
    bg_menu.mapData[x+y*32] = (FONT_TILE + (text)-0x20)|((pal&0xf)<<12);
  }
}
void print_teletext(int x, int y, int pal, char text) {
  if (text != 0x0) {
    bg_teletextfg.mapData[x+y*32] = (FONT_TILE +  (text)-0x20)|((pal&0xf)<<12);
  }
}
void print_teletextbg(int x, int y, int pal) {
  bg_teletextbg.mapData[x+y*32] = (FONT_TILE + ('~'-0x20+1) )|((pal&0xf)<<12);
}

void printTxt(const char * const text, int row, int col) {
  int i;
  int len = strlen(text);
  for (i = 0; i < len && i < 30; i++) {
    print_text(text[i],row+i,col,0);
  }
}
void printTxtPal(char * text, int row, int col,int pal) {
  int i;
  for (i = 0; i < strlen(text); i++) {
    print_text(text[i],row+i,col,pal);
  }
}

void setRowPalette(int row, int newPalette) {
  int x;
  for (x= 0; x < 32; x++) {
    int index = x+row*32;
    bg_menu.mapData[index] = (bg_menu.mapData[index] & ~0xF000) | ((newPalette<<12)&0xF000);
  }
}

void clearText(void) {
  int x,y; 
  for (y=0; y < 20; y++) {
    for (x=0; x < 30; x++) {
      bg_menu.mapData[x+y*32] = 0;
    }
  }
  bg_menu.x_scroll = 0;
  bg_menu.y_scroll = 0;
  UpdateBackground(&bg_menu);
}

void clearTextRow(int row) {
  // first clear the row
  int x;
  for (x=0; x < 32; x++) {
    bg_menu.mapData[x+row*32] = 0;
  }
}

//void autoScale(void) {
void autoScaleDoubleWidth(void) {
  // auto scale based on pixel type plus
  // CRTC_HorizontalDisplayed
  // CRTC_VerticalDisplayed

  const static int xhave = 240; // 240/2...

  // this is the real size in pixels of the screen
  int xwanted = CRTCControl.CRTC_HorizontalDisplayed*2;
  if (0 == (VideoULA_ControlReg & 0x10)) {
    // uses fat pixels
    xwanted *= 2;
  }

  int xscale, yscale, remainder;
  FastIntDivide(xwanted*256, xhave , &xscale,&remainder);
  FastIntDivide(CRTCControl.CRTC_VerticalDisplayed*256*8, 160 , &yscale,&remainder);
  
  setSpriteScreen(xscale,yscale);
}


void setSpriteScreenHalfWidth(int scale_x,int scale_y) {
  int x, y;
  int realWidth, realHeight;
  if (scale_x == 0) {
    scale_x = rotData[0].pa;
  }
  if (scale_y == 0) {
    scale_y = rotData[0].pd;
  }
  if (scale_y < 256) {
    // bigger than should be
    return;
  }
  int remainder; 
  FastIntDivide(0x4000, scale_x , &realWidth ,&remainder);
  realWidth = 0x40;
  FastIntDivide(0x4000, scale_y , &realHeight ,&remainder);
  int xoff = (64 - realWidth)/2;  // xoff = 3 or 3
  int yoff = (64 - realHeight)/2;
  
  for (x = 0 ; x < 4; x++) {
    for (y = 0 ; y < 4; y++) {
      init_sprite64_half(x,y,1);
      MoveSprite(&sprites[x+y*4], (x*realWidth)-xoff, y*realHeight-yoff);
    }
  }
  
  rotData[0].pa = (u16)0x100;
  rotData[0].pb = rotData[0].pc =0;
  rotData[0].pd = (u16)scale_y;
  
  CopyOAM();
}

#if defined(ELITE_HACK)
void setEliteScanner(void) {
  int remainder, realWidth, realHeight; 
  int scale_x = 128;
  int scale_y = rotData[0].pd;
  FastIntDivide(0x4000, scale_x , &realWidth ,&remainder);
  FastIntDivide(0x4000, scale_y , &realHeight ,&remainder);
  int xoff = (128 - realWidth)/2;  // xoff = 3 or 3
  int yoff = (128 - realHeight)/2;
  int x, y;
  for (x = 0 ; x < 4; x++) {
    for (y = 3 ; y < 4; y++) {
      init_sprite64(x,y,1);
      sprites[x+y*4].attribute1 = SIZE_64|ROTDATA(1); 
      sprites[x+y*4].attribute2 |= PALETTE(1); 
      MoveSprite(&sprites[x+y*4], 
          (x*realWidth)-xoff, 
          y*realHeight-yoff
          );
    }
  }
  rotData[1].pa = (u16)128;
  rotData[1].pb = rotData[1].pc =0;
  rotData[1].pd = (u16)scale_y;
  CopyOAM();
}
#endif

//void autoScaleHalfWidth(void) {
void autoScale(void) {
  // for fat pixels mode, it should be possible to auto scale the height
  // and make the width half, while keeping the sprites not SIZE_DOUBLE
  const static int xhave = 240; // 240/2...

  // this is the real size in pixels of the screen
  int xwanted = CRTCControl.CRTC_HorizontalDisplayed*2;
  if (0 == (VideoULA_ControlReg & 0x10)) {
    // uses fat pixels
    xwanted *= 2;
  }
  // but set the pixels to half of this...
  xwanted /= 2;

  int xscale, yscale, remainder;
  FastIntDivide(xwanted*256, xhave , &xscale,&remainder);
  
  FastIntDivide(CRTCControl.CRTC_VerticalDisplayed*256*8, 160 , &yscale,&remainder);
 
  // now set the sprites so they do not have the DOUBLE flag set.. 
  setSpriteScreenHalfWidth(xwanted,yscale);

#if defined(ELITE_HACK)
  setEliteScanner();
#endif
  
}

// this works fine in emulators... but on the real thing
// the graphics chip can't draw so much fancy sprite and
// the result is ugly black gaps...
void setSpriteScreen(int scale_x, int scale_y) {
  int x,y;
  int realWidth, realHeight;
  if (scale_x == 0) {
    scale_x = rotData[0].pa;
  }
  if (scale_y == 0) {
    scale_y = rotData[0].pd;
  }
  if (scale_y < 256 || scale_x < 128) {
    // bigger than should be
    return;
  }

  int remainder; 
  FastIntDivide(0x4000, scale_x , &realWidth ,&remainder);
  FastIntDivide(0x4000, scale_y , &realHeight ,&remainder);
  int xoff = (128 - realWidth)/2;  // xoff = 3 or 3
  int yoff = (128 - realHeight)/2;

  for (x = 0 ; x < 4; x++) {
    for (y = 0 ; y < 4; y++) {
      init_sprite64(x,y,1);
      MoveSprite(&sprites[x+y*4], 
          (x*realWidth)-xoff, 
          y*realHeight-yoff
          );
    }
  }
  
  rotData[0].pa = (u16)scale_x;
  rotData[0].pb = rotData[0].pc =0;
  rotData[0].pd = (u16)scale_y;
  
  CopyOAM();
  
}

void adjustOffset(int dx, int dy) {
  
  if (dx == 0 && dy == 0)
    return;
  
  int curr_x, curr_y, x, y;  
  for (x = 0 ; x < 4; x++) {
    for (y = 0 ; y < 4; y++) {
      curr_x = sprites[x+y*4].attribute1 & 0x1ff;
      curr_y = sprites[x+y*4].attribute0 & 0xff;
      MoveSprite(&sprites[x+y*4], curr_x+dx, (curr_y+dy));
    }
  }
  CopyOAM();
}

