/* The format of this file is arm assembler
 * it is pre processed by the gnu pre processor, so C/C++isms are allowed
 * comments and defines, ifdefs etc allowed - C code is not! 
 */

/*
The way this works is like this:

PC contains the program counter. It's initialised to the start of the prgram
and starts emulating from there.

As an instruction is read, the PC is increased to fetch each byte of the
intruction, then at the end the doCycles macro updates the cycles count. If the
cycles run out (ie a scanline event occurs) then the scanline routine is
called, othewise the bbc_pc is used to get the next opcode, the address of the
opcode is read from the opcodes table and the gba pc is set to the opcode
address. The process repeats...
*/


#define MACRO_LABEL myLabel\\@
#define MACRO_LABEL_NAME myLabel\\@:

#define PSR_N 0x80000000
#define PSR_Z 0x40000000
#define PSR_C 0x20000000
#define PSR_V 0x10000000


#define bbc_cycles  r3
#define bbc_pc      r4
#define bbc_a       r5
#define bbc_x       r6
#define bbc_y       r7
#define bbc_c       r8    // PSR_C
#define bbc_nz      r9    // PSR_N = N , bits 0-7 == 0 PSR_Z       also contains next 8bit value
#define bbc_opcode  r10   // opcode address to jump to
#define bbc_optbl   r10   // op code table

#define bbc_mem     r11   // memory
#define memptr      r12   // memory pointer into bbc_mem

@r13 SP
@r14 LR
@r15 PC

#define bbc_stack  [bbc_optbl,#(256*4 )]
#define bbc_di  [bbc_optbl,#(256*4 + 4)]
#define bbc_v  [bbc_optbl,#(256*4 + 8)]
#define bbc_event [bbc_optbl,#(256*4 + 12)]
#define bbc_intstat   [bbc_optbl,#(256*4 + 16)]
//#define readmem_offset   (20)
//#define writemem_offset  (readmem_offset + 4*16)
//#define Asm6502State_offset (writemem_offset + 4*16)

#define BRANCH   1

#define IMMEDIATE   1
#define ZEROPAGE    2
#define ABSOLUTE    3

//.equ CYCLE, 16       // 128 * CYCLES per scaline
.equ C, 0b00000001  
.equ Z, 0b00000010
.equ I, 0b00000100
.equ D, 0b00001000
.equ B, 0b00010000
.equ R, 0b00100000
.equ V, 0b01000000
.equ N, 0b10000000
.equ IRQ_VECTOR, 0xfffe
.equ NMI_VECTOR, 0xfffa
.equ BRK_VECTOR, 0xfffe

// decrease cycle count, check events.
.macro doCycles cycs  
#ifdef DEBUG
  mov r2,#\cycs
  bl  debugOutput
#endif
  subs bbc_cycles,bbc_cycles,#\cycs          // *3*CYCLE
  ldrplb r0,[bbc_pc],#1                      // if positive or 0, load r0 with the next opcode
  ldrpl pc,[bbc_optbl,r0,lsl#2]              // load the GBA pc with the next opcode address..
  ldr pc,bbc_event                             // otherwise load with the next event (eg. scanline)
.endm

// this possibly hacky interrupt check s for after CLI is called...
.macro  doInterruptCheck
  ldr     r1,=intStatus
  ldr     r1,[r1]
  cmp     r1, #0      @ Z is set if statis == 0
  @messing about to try and get this working...
  ldrne   pc,bbc_event 
@  beq     1f
 @ ldrb r0,[bbc_pc,#-1]
 @ ldrb r1,[bbc_pc,#-2]
 @ cmp    r0,#0x58 
 @ cmpne  r1,#0x58 
  @andne   r0,r0,#I
  @eornes  r0,r0,#I    @ NE if status != 0, 
@  ldreq   pc,bbc_event @ NE if status != 0 && oldI^I == I
@1:
.endm

// translate bbc_pc from 6502 PC to rom offset
// on the BBC it is easier than nes, no bank swapping..
.macro encodePC
  add bbc_pc,bbc_mem,bbc_pc         // PC = PC+bank
.endm

// push and pull
.macro push8 reg
  ldrb r2,bbc_stack               // load the stack address (which is main ram adress + 0x1ff)
  orr  r2,r2,#0x100
  strb \reg,[bbc_mem,r2]
  sub  r2,r2,#1                   // store the register in the stack space
  strb r2,bbc_stack               // and update stack pointer
.endm
  
.macro  push16                          //save PC, r0
  mov r1,r0,lsr#8                       // r1 = r0>>8
  ldrb r2,bbc_stack
  orr r2,r2,#0x100     
  strb r1,[bbc_mem,r2]
  sub  r2,r2,#1 
@  strb r1,[r2],#-1     
  orr r2,r2,#0x100     
  strb r0,[bbc_mem,r2]
  sub  r2,r2,#1  
@  strb r0,[r2],#-1                  // r0&ff
  strb r2,bbc_stack
.endm

.macro pop16                     
  ldrb r2,bbc_stack              // stack == f8
  add r2,r2,#1                  
  orr r2,r2,#0x100              // 1f8
  ldrb bbc_pc,[bbc_mem,r2]      
  add r2,r2,#1     
  strb r2,bbc_stack             // store stack             
  @ldr r2,bbc_stack              
  @ldrb r0,[r2]                  
  ldrb r0,[bbc_mem,r2]      
  orr bbc_pc,bbc_pc,r0,lsl#8    
.endm

.macro pop8 reg
  ldrb r2,bbc_stack         @ value of stack, eg 0xfa       
  add r2,r2,#1              @ r2 = 0xfb
  strb r2,bbc_stack         @ store new stack value
  orr r2,r2,#0x100          @ r2 = 0x1fb
  ldrsb \reg,[bbc_mem,r2]   @ reg = bbc_mem[r2]
.endm

//unpack 6502 flags from r0
.macro decodeP
  and r1,r0,#D+I+B
  str r1,bbc_di     @DI
  mov r1,r0,lsl#22              
  str r1,bbc_v      @V
  mov bbc_c,r0,lsl#29   @C
  mov bbc_nz,r0,lsl#24    @N
  tst r0,#Z
  orreq bbc_nz,bbc_nz,#1    @Z
  
  
.endm

.macro encodeP extra            
  and r1,bbc_nz,#0x80000000     
  ldr r0,bbc_di                      //@DI
  orr r0,r0,r1,lsr#24                //@N
  tst bbc_nz,#0xff              
  orreq r0,r0,#Z                
  tst bbc_c,#PSR_C              
  orrne r0,r0,#C                
  ldr r1,bbc_v                  
  tst r1,#PSR_V                 
  orrne r0,r0,#V                
  orr r0,r0,#\extra               
.endm

/*
 Start of memory writing defines
*/
//typedef void (*funcPtr)(int val);
//
//extern funcPtr writemem_tbl[];
/*
  r0 = value, r12 = address to these routines
*/


.macro BBC_writemem
/* less than 0x8000 can write direct to memory
@@ if it's >= 0x8000 && < 0xfc00 return (writing to rom)
@@ otherwise write to hardware registers:
@@ 0xfe00 crtc Write
@@ 0xfe20 video ula write
@@ 0xfe40, 0xfe50 = Sys VIA write (cycles++)
@@ 0xfe60, 0xfe70 = User VIA write (cycles++)
@ r0 = byte to write, memptr = address to write to
 */
  @ if < 0x8000 can do a direct write:
  cmp     memptr,#0x8000
  strccb  r0,[bbc_mem,memptr]
  bcc     0f
  @ else >= 0x8000
  cmp     memptr,#0xfe00
  bcc     0f
  @ else >= 0xfe00 
  cmp     memptr,#0xff00
  strcsb  r0,[bbc_mem,memptr]       @ eg interrupt vector
  bcs     0f
  @ else >= 0xfe00 && < 0xff00, need to check io space
  and     r1,memptr,#0xf0             
  @ rubbish way commented out
  @mov     r2,#(256*4)           
  @add     r2,r2,#writemem_offset          @r2 = address of write mem table
  @add     r2,bbc_optbl,r2
  ldr     r2,=writemem_tbl
  adr     lr,0f                    @return vector == endRead
  ldr     pc,[r2,r1,lsr#2]         @shift down 4, then times by 4 , i.e. >>4 << 2 = >>2
                                   @call the routine for reading mem at this address
0:
  
.endm
    
.macro writememzp
  strb r0,[bbc_mem,memptr]
.endm

/*
 Start of memory reading defines
*/

// immediate.
// memptr = PC
.macro BBC_ImmRead 
  ldrsb bbc_nz,[bbc_pc],#1
.endm

// zero page 
// memptr = WholeRam[PC]
.macro BBC_ZP_Read 
  ldrsb bbc_nz,[bbc_mem,memptr]
.endm

// absolute
// < 0xfc00 can read direct from memory
// >= 0xff00 can read direct from memory
// otherwise IO space
// 0xfe00 crtc 
// 0xfe20 video ula
// 0xfe40, 0xfe50 = Sys VIA 
// 0xfe60, 0xfe70 = User VIA
// 0xfea0, return 0xfe
// else return 0
// bbc_nz holds return value
.macro BBC_AbsRead
  @ < fc00 or >= ff00 do a direct read
  cmp     memptr,#0xfc00
  ldrccsb bbc_nz,[bbc_mem,memptr]       @ sign extend needed for Load
  ldrccb  r0,[bbc_mem,memptr]           @ r0 load needed for ROR
  bcc     0f
  
  cmp     memptr,#0xff00
  ldrcsb  bbc_nz,[bbc_mem,memptr]
  bcs     0f
  
  @ otherwise use the table of mem read routines
  and     r0,memptr,#0xf0
  @ comments were the rubbish old way I did this
  @mov     r2,#(256*4)           
  @add     r2,r2,#readmem_offset    @r2 = address of read mem table
  @add     r2,bbc_optbl,r2
  ldr     r2,=readmem_tbl
  adr     lr,0f                    @return vector == endRead
  ldr     pc,[r2,r0,lsr#2]         @shift down 4, then times by 4 , i.e. >>4 << 2 = >>2
                                   @call the routine for reading mem at this address
  @ldrb  bbc_nz,[bbc_mem,memptr]
0:
.endm

.macro BBC_ReadMem
  .ifeq AddressMode-IMMEDIATE   
  @.print "IMMEDIATE!"
     BBC_ImmRead
  .endif
  .ifeq AddressMode-ZEROPAGE
    @.print "ZEROPAGE!"
    BBC_ZP_Read
  .endif
  .ifeq AddressMode-ABSOLUTE
  @.print "ABSOLUTE!"
    BBC_AbsRead
  .endif
.endm

/*********************************************
  Start addressing mode defines
**********************************************/

// absolute $nnnn mode
.macro doABS                                       
.set AddressMode, ABSOLUTE     
  ldrb memptr,[bbc_pc],#1     
  ldrb r0,[bbc_pc],#1         
  orr memptr,memptr,r0,lsl#8
.endm

// absolute indexed X     $nnnn,X
.macro doAIX                              
.set AddressMode, ABSOLUTE             
  ldrb memptr,[bbc_pc],#1             
  ldrb r0,[bbc_pc],#1                 
  orr memptr,memptr,r0,lsl#8          
  add memptr,memptr,bbc_x,lsr#24
.endm
//bic memptr,memptr,#0xff0000 @Base Wars needs this


// absolute indexed Y     $nnnn,Y
.macro doAIY                           
.set AddressMode, ABSOLUTE             
  ldrb memptr,[bbc_pc],#1             
  ldrb r0,[bbc_pc],#1                 
  orr memptr,memptr,r0,lsl#8          
  add memptr,memptr,bbc_y,lsr#24
  bic memptr,memptr,#0xff0000    @ Elite needs this
                                 @ STA $FFFD,Y  <- Y up to 6
.endm


// Indirect,X    ($44,X)
/*
memptr = $44[X]|$44[X+1]

*/
.macro doIIX 
.set AddressMode, ABSOLUTE   
  ldrb r0,[bbc_pc],#1               // r0 == zp val (bbc_pc points to rom mem)
  add r0,bbc_x,r0,lsl#24            // r0 == x + zp val<<24
  ldrb memptr,[bbc_mem,r0,lsr#24]    // memptr = bbc_mem[r0>>24]
  add r0,r0,#0x01000000           // r0 = r0 + 0x01000000
  ldrb r1,[bbc_mem,r0,lsr#24]   // r1 = bbc_mem[1 + r0>>24]
  orr memptr,memptr,r1,lsl#8        // memptr = memptr + r1<<8
.endm

//@indirect indexed Y     ($nn),Y
.macro doIIY                            
.set AddressMode, ABSOLUTE               
  ldrb r0,[bbc_pc],#1                   
  ldrb memptr,[r0,bbc_mem]!             
  ldrb r1,[r0,#1]                       
  orr memptr,memptr,r1,lsl#8            
  add memptr,memptr,bbc_y,lsr#24
.endm
  
//  bic addy,addy,#0xff0000 @Zelda2 needs this

// zero page memptr = *bbc_pc++
.macro doZ                
.set AddressMode, ZEROPAGE 
  ldrb memptr,[bbc_pc],#1
.endm

//@zero page indexed X    $nn,X
.macro  doZIX                  
.set AddressMode, ZEROPAGE  
  ldrb memptr,[bbc_pc],#1  
  add memptr,memptr,bbc_x,lsr#24 
  and memptr,memptr,#0xff 
.endm

//and addy,addy,#0xff @Rygar needs this

//@zero page indexed Y    $nn,Y
.macro doZIY                           
.set AddressMode, ZEROPAGE     
  ldrb memptr,[bbc_pc],#1     
  add memptr,memptr,bbc_y,lsr#24 
  and memptr,memptr,#0xff
.endm

.macro  doIMM
.set AddressMode, IMMEDIATE   
@.print "setting IMMEDIATE!"
.endm
  
/*********************************************
  Start opcode defines
**********************************************/
// A = A|M
// NZ = A


// e.g.  
/*
case 0x01:
  // e.g Indirect,X    ORA ($44,X)   $01  2   6
  ORAInstrHandler(IndXAddrModeHandler_Data());
  break;
  
  doIX
  doOra
  doCycles(3)
 
*/

// ASL  = A<<1
.macro doASL  
.ifeq AddressMode-ABSOLUTE          
    BBC_AbsRead
    mov bbc_c,bbc_nz,lsl#22         
    mov r0,bbc_c,lsr#21             
    orr bbc_nz,r0,r0,lsl#24         
    BBC_writemem
.else                               
    ldrb r0,[bbc_mem,memptr]        
    add r0,r0,r0                    
    orrs bbc_nz,r0,r0,lsl#24  @NZ   
    mrs bbc_c,cpsr                  
    strb r0,[bbc_mem,memptr]        
.endif                              
.endm


.macro doAND                       
  BBC_readmem                      
  and bbc_a,bbc_a,bbc_nz,lsl#24    
  mov bbc_nz,bbc_a,asr#24
.endm

.macro doADC
  BBC_readmem
  eor bbc_nz,bbc_nz,#0xff
  msr cpsr_f,bbc_c                     //@get C
  sbcs bbc_a,bbc_a,bbc_nz,lsl#24
  mov bbc_nz,bbc_a,asr#24                 //@NZ      
  mrs bbc_c,cpsr                          //@C       
  str bbc_c,bbc_v                         //@V       
  and bbc_a,bbc_a,#0xff000000
.endm

.macro doBCDADC
  BBC_readmem  
  @ low value is A&f + M&f + Carry
  and   r0,bbc_a,#0x0f000000
  and   r1,bbc_nz,#0x0f
  msr   cpsr_f,bbc_c                     @get C
  adc   r0,r1,r0,lsr#24       @ r0 = result (add together in the low byte)
  @ if this result is > 9, need to "carry" to the top nibble and round off ln
  cmp   r0,#0x0a
  addcs r0,r0,#0x06
  andcs r0,r0,#0x0f
  @ hi value is A&f0 + M&f0 + Low Nibble Carry
  and   r1,bbc_nz,#0xf0     @ hi nibble of M
  addcs r1,r1,#0x10         @ carry from low nibble
  and   r2,bbc_a,#0xf0000000  @ A&f0
  adds  r1,r2,r1,lsl#24       @ hi nibble = A&f0 + M&f0 + Low Nibble Carry
  and   r2,r1,#(PSR_Z|PSR_N)    @N|V setup
  @ now fix r1 if it's > 0x90
  cmpcc r1,#0xa0000000    @ careful, that's very negative.
  addcs r1,r1,#0x60000000
  andcs r1,r1,#0xf0000000
  movcs bbc_c,#PSR_C
  movcc bbc_c,#0
  orr   bbc_a,r1,r0,lsl#24
  ands  r1,r2,#(PSR_Z)    @V setup 
  orrne bbc_c,bbc_c,r1
  mov bbc_nz,bbc_a,lsr#24                 @ Z is incorrect, should be 2's compliment result
  orr bbc_nz,bbc_nz,r2                    @ N
  str bbc_c,bbc_v                         @ V       
  and bbc_a,bbc_a,#0xff000000  
.endm

.macro doCompleteADC cycs
  ldr   r0,bbc_di
  tst   r0,#D
  @ if D set then need to do bcd add
  @ else do 2s compliment add:
  bne   2f
  doADC
  doCycles \cycs
2:
  doBCDADC
  doCycles \cycs
.endm



.macro doBIT
  BBC_readmem
  and r0,bbc_nz,bbc_a,lsr#24          //@Z
  mov r1,bbc_nz,lsl#22            
  str r1,bbc_v                        //@V
  orr bbc_nz,r0,bbc_nz,lsl#24         //@N
.endm 


.macro doCOMP x
  BBC_readmem
  subs bbc_nz,\x,bbc_nz,lsl#24
  mov bbc_nz,bbc_nz,asr#24      @NZ
  mrs bbc_c,cpsr                @C
.endm

.macro doDEC
  .ifeq AddressMode-ABSOLUTE
    BBC_AbsRead
    and r0,bbc_nz,#0xff
    sub r0,r0,#1
    orr bbc_nz,r0,r0,lsl#24 //@NZ
    BBC_writemem
  .else
    ldrb r0,[bbc_mem,memptr]
    sub r0,r0,#1
    orr bbc_nz,r0,r0,lsl#24 //@NZ
    strb r0,[bbc_mem,memptr]
  .endif
.endm

.macro doEOR
  BBC_readmem
  eor bbc_a,bbc_a,bbc_nz,lsl#24
  mov bbc_nz,bbc_a,asr#24           //@NZ
.endm

.macro doINC
  .ifeq AddressMode-ABSOLUTE
    BBC_AbsRead
    add r0,bbc_nz,#1
    orr bbc_nz,r0,r0,lsl#24 @NZ
    BBC_writemem
  .else
    ldrb r0,[bbc_mem,memptr]
    add r0,r0,#1
    orr bbc_nz,r0,r0,lsl#24 @NZ
    strb r0,[bbc_mem,memptr]
  .endif
.endm


.macro doLOAD x
  BBC_readmem
  mov \x,bbc_nz,lsl#24
.endm

.macro doLSR
  .ifeq AddressMode-ABSOLUTE
    BBC_AbsRead
    mov bbc_c,bbc_nz,lsl#29
    and bbc_nz,bbc_nz,#0xfe @(N=0)
    mov r0,bbc_nz,lsr#1
    BBC_writemem
  .else
    ldrb bbc_nz,[bbc_mem,memptr]
    movs bbc_nz,bbc_nz,lsr#1  @NZ
    mrs bbc_c,cpsr      @C
    strb bbc_nz,[bbc_mem,memptr]
  .endif
.endm

.macro doORA
  BBC_readmem
  orr bbc_a,bbc_a,bbc_nz,lsl#24
  mov bbc_nz,bbc_a,asr#24
.endm



.macro doROL
  .ifeq AddressMode-ABSOLUTE
     BBC_AbsRead
     msr cpsr_f,bbc_c @get C
     @adc r0,bbc_nz,bbc_nz
     adc r0,r0,r0
     orrs bbc_nz,r0,r0,lsl#24 @NZ
     mrs bbc_c,cpsr   @C
    BBC_writemem
  .else
     ldrb r0,[bbc_mem,memptr]
     msr cpsr_f,bbc_c @get C
     adc r0,r0,r0
     orrs bbc_nz,r0,r0,lsl#24 @NZ
     mrs bbc_c,cpsr   @C
    strb r0,[bbc_mem,memptr]
  .endif
.endm




.macro doROR
.ifeq AddressMode-ABSOLUTE
     BBC_AbsRead
     tst bbc_c,#PSR_C             
     orrne r0,bbc_nz,#0x100       @ if bbc_c set, r0 = 0x100|r0 BUT using nz takes care of sign
     movs r0,r0,lsr#1             @ r0 = r0/2
     mrs bbc_c,cpsr               
     orr bbc_nz,r0,r0,lsl#24      @ bbc_nz = r0 | (r0<<24) 

@     mov  r0,bbc_nz,lsl#24                 @r0 = value
@     orr  r0,r0,bbc_c,lsr#29               @r0 = r0 | (C>>29)   (if carry set, r0 = 2?)
@     movs r0,r0,ror#25                     @r0 = r0>>25       (if carry set, r0 = 0x100)
@     mrs  bbc_c,cpsr                       @bbc_c = hardware carry flag
@     orr  bbc_nz,r0,r0,lsl#24              @bbc_nz = r0 | (r0<<24)

    BBC_writemem
  .else
     ldrb r0,[bbc_mem,memptr]     @ r0 = value
     tst bbc_c,#PSR_C             
     orrne r0,r0,#0x100           @ if bbc_c set, r0 = 0x100|r0
     movs r0,r0,lsr#1             @ r0 = r0/2
     mrs bbc_c,cpsr               
     orr bbc_nz,r0,r0,lsl#24      @ bbc_nz = r0 | (r0<<24)
     strb r0,[bbc_mem,memptr]
  .endif
.endm

@ this is fine, but the Z,C,N anv V flgs are most likely WRONG
.macro doBCDSBC
  BBC_readmem                      @ bbc_nz has value
  @ low value is A&f - M&f - (1 - Carry)
  and   r0,bbc_a,#0x0f000000
  and   r1,bbc_nz,#0x0f
  @ Z result is 2s compliment result...
  msr   cpsr_f,bbc_c
  sbc   r12,bbc_a,bbc_nz,lsl#24

  @ now for the bcd version of SBC ...
  msr   cpsr_f,bbc_c                     @get Carry
  sbcs  r0,r0,r1,lsl#24       @ r0 = result (subtract M from A in the low byte)

  
  @ if ln < 0 need to subtract "carry" from the top nibble and round off ln
  submi r0,r0,#0x06000000     @ ln -= 6
  andmi r0,r0,#0x0f000000     @ ln &= f
  @ hi value is A&f0 - M&f0 - Low Nibble Carry
  and   r1,bbc_nz,#0xf0     @ hi nibble of M
  addmi r1,r1,#0x10         @ carry from low nibble
  and     r2,bbc_a,#0xf0000000  @ A&f0 
  subs    r1,r2,r1,lsl#24       @r1 is hi nibble = A&f0 - M&f0 - Low Nibble Carry
  
  @ the N flag is set before adjusting the hi nib result 
  and   r2,r1,#(PSR_N)    @N setup

  @ now fix r1 if  we wrapped...
  subcc r1,r1,#0x60000000
  andcc r1,r1,#0xf0000000
  movcc bbc_c,#0
  movcs bbc_c,#PSR_C          @ C flag
  orr   bbc_a,r1,r0           @ ACC 
  @ands  r1,r2,#(PSR_Z)       @V setup
  @orrne bbc_c,bbc_c,r1
  
  mov bbc_nz,r12,asr#24            @ Z = 2s compliment sbc result
  orr bbc_nz,bbc_nz,r2             @ N = &0x80 on hn before adjust
  str bbc_c,bbc_v                  @ V
  and bbc_a,bbc_a,#0xff000000
.endm

.macro doSBC
  BBC_readmem
  msr cpsr_f,bbc_c        @get C
  sbcs bbc_a,bbc_a,bbc_nz,lsl#24
  mov bbc_nz,bbc_a,asr#24 @NZ
  mrs bbc_c,cpsr          @C
  str bbc_c,bbc_v         @V
  and bbc_a,bbc_a,#0xff000000
.endm

.macro doCompleteSBC cycs
  ldr   r0,bbc_di
  tst   r0,#D
  @ if D set then need to do bcd subtract
  @ else do 2s compliment add:
  bne   2f
  doSBC
  doCycles \cycs
2:
  doBCDSBC
  doCycles \cycs
.endm

.macro doSTORE x
  mov r0,\x,lsr#24
  .ifeq AddressMode-ABSOLUTE
    BBC_writemem
  .else
    strb r0,[bbc_mem,memptr]
  .endif
.endm


