
#include "gba.h"
#include "bbcsound.h"

// this is 4 ints of tone,
//         4 ints of colume
//         1 int  of lasttone
extern u32 AsmSoundState[4+4+1];

//SGn0
#define   SOUND_LENGTH(n)               n
#define   ENVELOPE_START_VALUE(n)       (n<<12)

//SGCNT1
#define SOUND_MASTER_ENABLE			BIT07
#define SWEEP_CHANGE_FREQ(n)		(n<<4)

#define DUTY_CYCLE(n)		(n<<6)

#define   SOUND_OUTPUT_FULL				BIT01


#define SOUND3BANK32		0x0000	// Use two banks of 32 steps each
#define SOUND3BANK64		0x0020	// Use one bank of 64 steps
#define SOUND3PLAY			0x0080	// Output sound
#define SOUND3OUTPUT1		0x2000	// Output unmodified
#define SOUND3OUTPUT14		0x6000	// Output 1/4 
#define SOUND3SETBANK1		0x0040
#define SOUND3SETBANK0		0x0000

#define   ENVELOPE_VOL_S1    5
#define   ENVELOPE_VOL_S2    5
#define   ENVELOPE_VOL_S4    3

void initSound1(void) {
  REG_SG10_L = 0x8;    //;square0 sweep off
  // length is 8 bits on this one
  REG_SG10_H = SOUND_LENGTH(0x10) | DUTY_CYCLE(1) | ENVELOPE_START_VALUE(ENVELOPE_VOL_S1);
  
}

void initSound2(void) {
  // length is just 6 bits
  REG_SG20 = SOUND_LENGTH(0x10) | DUTY_CYCLE(3) | ENVELOPE_START_VALUE(ENVELOPE_VOL_S2);
}

void initSound3(void) {
  REG_SG30_L = SOUND3BANK64 | SOUND3SETBANK0;
  REG_SG30_L |= SOUND3PLAY;
  REG_SG30_H =  SOUND3OUTPUT1|0x11;
  
  // load the wave ram bank 0
  // square wave 32 bits
  /* ..................................
     7 ................
     6 |              |
     5 |              |        
     4 |              |           
     3 |              |            
     2 |              |              
     1 |              |             
     0                |
     f                 |              |
     e                 |              | 
     d                 |              | 
     c                 |              | 
     b                 |              | 
     A                 |              |
     9                 |              |
     8                 ................
   */

#if 0
  REG_SGWR0=0x77887788;
  REG_SGWR1=0x77887788;
  REG_SGWR2=0x77887788;
  REG_SGWR3=0x77887788;
#else
  REG_SGWR0=0x77777777;
  REG_SGWR1=0x88888888;
  REG_SGWR2=0x77777777;
  REG_SGWR3=0x88888888;
#endif
  
}

void initSound4(void) {
  // 6 bit length
  REG_SG40 = SOUND_LENGTH(0x10) | ENVELOPE_START_VALUE(ENVELOPE_VOL_S4);
  
}

void initSoundRegs(void) {
  // initialise the sound 1-4 registers
  
  REG_SGCNT1 = SOUND_MASTER_ENABLE; // only allowed to write this bit
  REG_SGCNT0_H = SOUND_OUTPUT_FULL;
  REG_SGCNT0_L = 0x0077;  // full sound
  
  initSound1(); 
  initSound2();
  initSound3(); // variable
  initSound4(); // noise
}

void dumpSoundReg(int ch, int freq, int vol, int SGn1) {
#if 0
	if (vol)
		gbaprintf("ch %02x Vol %02x F %04x REG %08x\n", ch, vol, freq, SGn1);
#endif
} 

void LoadSoundUEF(const unsigned char * SUEF) {

	// untested, probably doesn't work...
	int Chan;
	int Data;
	int RegVal;
	for (Chan=0;Chan<3;Chan++) {
		Data=(*SUEF)|((*(SUEF+1))<<8);
		SUEF+=2;
		RegVal=(((Chan)*2)<<4)|128;
		RegVal|=(Data&15);
//		writeSound(RegVal);
		RegVal=(Data&1008)>>4;
//		writeSound(RegVal);
	}
	for (Chan=0;Chan<3;Chan++) {
		Data=*SUEF++;
		RegVal=((((Chan)*2)+1)<<4)|128;
		RegVal|=Data&15;
	}
}
