
#include "gba.h"
#include "screenmode.h"
#include "interrupts.h"
#include "sprite.h"
#include "keypad.h"

#include "gfx.h"
#include "inter.h"
#include "savestate.h"
#include "input.h"
#include "sysvia.h"
#include "uservia.h"
#include "fastvideo.h"
#include "6502core.h"
#include "visoly.h"
#include "gamemenu.h"
#include "bbcmemory.h"
#include "bbcsound.h"
#include "bbcdisc.h"
#include "settings.h"
#include "pogo.h"

// set to 1 and run a single game test (STATE_FILE)
// set to 0 and have to add uefs via menu
#define STAND_ALONE 0

#define STATE_FILE _binary_elite_uef_start
#define STATE_FILE_END _binary_elite_uef_end

#if 0
#define STATE_FILE _binary_dodgems_uef_start
#define STATE_FILE_END _binary_dodgems_uef_end

#define STATE_FILE _binary_chuckie_uef_start
#define STATE_FILE_END _binary_chuckie_uef_end

#define STATE_FILE _binary_monsters_uef_start
#define STATE_FILE_END _binary_monsters_uef_end

#define STATE_FILE _binary_repton_uef_start
#define STATE_FILE_END _binary_repton_uef_end


#define STATE_FILE _binary_boxer_uef_start
#define STATE_FILE_END _binary_boxer_uef_end


#define STATE_FILE _binary_elite_uef_start
#define STATE_FILE_END _binary_elite_uef_end
#define STATE_FILE _binary_felix_uef_start
#define STATE_FILE_END _binary_felix_uef_end

#define STATE_FILE _binary_dodgems_uef_start
#define STATE_FILE_END _binary_dodgems_uef_end

#define STATE_FILE _binary_pengo_uef_start
#define STATE_FILE_END _binary_pengo_uef_end

#define STATE_FILE _binary_mushtext_uef_start
#define STATE_FILE_END _binary_mushtext_uef_end
#define STATE_FILE _binary_snapper_uef_start
#define STATE_FILE_END _binary_snapper_uef_end
#endif

extern unsigned char STATE_FILE[];
extern unsigned char STATE_FILE_END;


//char * IntrTable;

#if 0
// uses simple interupts...
void init_interrupts(void) {
  
  REG_IME = 0x00;
  REG_IE = INT_VBLANK;
  REG_IF = INT_VBLANK;
  // Enable Display V-Blank
  REG_DISPSTAT = BIT03;
  // Set interrupt proc address (IRQHandler) 
  REG_INTERUPT = (u32)irqarm;  
  // turn on master interrupt
  REG_IME = 0x01;
  
}
#endif

void Setup(void) {
  
  // set all the sprites to "not for processing" state
  u16 loop;
  for(loop = 0; loop < 128; loop++)  {
    remove_sprite(loop);
  }
  SetMode(MODE_0 | OBJ_ENABLE | OBJ_MAP_2D);
  
  initBG();
  initSoundRegs();

}
void resetBeeb(int soft) {
  if (!soft) {
    BeebMemInit();
  }
  initCore();
  
  if (!soft) {
    SysVIAReset();
    UserVIAReset();
    VideoInit();
  }
  reset_disc_drive();
}


void showSplash(void);
#ifndef DevkitARM
int AgbMain(void) {
#else
int main(void) {  
#endif

  g_isPogoshell = 0;
  // check for pogoshell start
  // ...
  checkPogoShell();
  // not Pogo
  UpdateKeys();
  releaseAllKeys();

  init_interrupts(INT_VBLANK,irqarm);
  clear_vram();

  lastKeys = thisKeys;
  showSplash();
  Setup();
  fade_up();
  resetBeeb(0);

  if (g_isPogoshell) {
    loadFromPogo();
  } else {

#if STAND_ALONE
    g_isPogoshell = 1; //  a good approximation
    LoadUEFState(STATE_FILE, &STATE_FILE_END);
#else
    int startPointer, endPointer;
    getGameToLoad(&startPointer,&endPointer,0,IS_UEF);
    if (endPointer != 0 && startPointer != 0) 
      LoadUEFState((u8*)startPointer,(u8*)(endPointer));
#endif  

#ifdef DEBUG    
    keyPressed(4,9);
#endif
  } // end if pogo or not

  CallAsm(run_emu);
  
  return 0;

}
