#include <stdio.h>
#include <string.h>
#include "gba.h"
#include "dma.h"
#include "sprite.h"
#include "visoly.h"
#include "screenmode.h"
#include "keypad.h"
#include "savestate.h"
#include "bbcmemory.h"
#include "menu.h"
#include "gamemenu.h"
#include "menudata.h"
#include "inter.h"
#include "input.h"
#include "inputcodes.h"
#include "6502core.h"
#include "gfx.h"
#include "fastvideo.h"
#include "virtualkb.h"
#include "settings.h"
#include "pogo.h"
#include "reset.h"
#include "sysvia.h"

#define MENU_EXIT 0
#define MAIN_MENU 1
#define CHANGE_KEY_MENU 2
#define SAVE_MENU 3
#define LOAD_MENU 4
#define VIDEO_MENU 5
#define RESET_MENU 6
#define SRAM_MENU 7

#define MENU_KEY_DELAY 10




int menuStatus = 0;

int g_maxItem        = 1; // how many items there are
int g_hiliteMenuItem = 0; // which item is highlit
int save_reg_sgcnt0_l = 0;
int selectedGameIndex = 0;
int selectedGameType = IS_UEF;

FunctionPtr leftHandler = 0;
FunctionPtr rightHandler = 0;
FunctionPtr AHandler = 0;

extern unsigned char * discs;// = DISC;
extern int shiftDownFrames;
void resetBeeb(int soft);


void initMenu(void) {
  // first, switch off that annoying sound!!
  save_reg_sgcnt0_l = REG_SGCNT0_L;
  REG_SGCNT0_L=0x0000;  // disabled. (no channels, no volume)
  // convert the core so it gets restored correctly when we exit:
  T6502State *state;
  state = (T6502State*)&Asm6502State;
  state->programCounter = state->programCounter-(unsigned int)WholeRam;
  state->accumulator >>= 24;
  state->XReg >>= 24;
  state->YReg >>= 24;
  
  REG_INTERUPT = (u32)MenuInterruptHandler;
  REG_IME = 0x01;
  menuStatus = MAIN_MENU;
  shiftDownFrames = 0;
}


u32 returnIRQ = 0;
void endMenu(void) {
  REG_IME = 0; // switch off 
  // turn the sound to back how it was...
  REG_SGCNT0_L=0x0077; 
  clearText();
  REG_BLDMOD = BLDMOD_FIRST_OBJ | BLDMOD_SECOND_BD | BLDMOD_MODE(3);
  REG_COLEY = 0;
  
  REG_INTERUPT = returnIRQ;
  
  REG_IME = 1; // switch on interrupts
  
  // carry on from where we were...
  //__FarProcedure(run_emu);
  if (apply_vk_press) {
    show_vkb();
  }
  CallAsm(run_emu);
  
}

void initMenuGui(void) {
  // do the main menu
  
  REG_BLDMOD = BLDMOD_FIRST_BG2 | BLDMOD_FIRST_BG1 | BLDMOD_FIRST_OBJ | BLDMOD_SECOND_BD | BLDMOD_MODE(3);
  REG_COLEY = 6;
  clearText();
  int i;
  for (i = 0 ; i < MENU_CNT; i++) {
    // create the menu option
    printTxt(g_theMenus[i].cName, 2,MENU_Y_OFFSET+i);
  }
  
  g_hiliteMenuItem = 0;
  g_maxItem = MENU_CNT;
  menuStatus = MAIN_MENU;
}

void showToolTip(const char * tip) {
  static int lastMenu = MENU_EXIT;
  static int lastIndex = -1;
  static int lastPosition = 0;
  static int counter = 0;
  if (g_hiliteMenuItem != lastIndex || menuStatus != lastMenu) {
    clearTextRow(19);
    printTxt(tip, 0,19);
    setRowPalette(19,3);
    lastMenu = menuStatus;
    lastIndex = g_hiliteMenuItem;
    lastPosition = 0;
    counter = 180;
  } else if (--counter == 0) {
    int len = strlen(tip);
    if (len > 30) {
       lastPosition++;
       if (lastPosition == len) {
         lastPosition = 0;
       }  
       clearTextRow(19);
       printTxt(&tip[lastPosition], 0,19);
       setRowPalette(19,3);
       counter = 10;
    }
  }
}

const char * const cLoadTip = "Press A to load game. up/down to change highlighted game. left/right change page"; 
void updateGui(void) {
  // show the currently selected menu and stuff
  // highlight the current item
  int i;
  for (i = 0 ; i < g_maxItem ; i++) {
    setRowPalette(MENU_Y_OFFSET+i,7);
  }
  
  setRowPalette(MENU_Y_OFFSET+g_hiliteMenuItem,6);

  if (menuStatus == MAIN_MENU) {
    showToolTip(g_theMenus[g_hiliteMenuItem].cTip);
  } else if (menuStatus == VIDEO_MENU) {
    showToolTip(g_videoOptions[g_hiliteMenuItem].cTip);
  } else if (menuStatus == CHANGE_KEY_MENU) {
    showToolTip(g_keyOptions[g_hiliteMenuItem].cTip);
  } else if (menuStatus == RESET_MENU) {
    showToolTip(g_resetOptions[g_hiliteMenuItem].cTip);
  } else if (menuStatus == SRAM_MENU) {
    showToolTip(g_sramOptions[g_hiliteMenuItem].cTip);
  } else if (menuStatus == LOAD_MENU && g_maxItem) {
    int th = g_hiliteMenuItem;
    g_hiliteMenuItem = selectedGameIndex;
    showToolTip(cLoadTip);
    g_hiliteMenuItem = th;
  } 
}

void wait_for_letgo(void) {
  for (;;) {
    wait_vsync_int();
    UpdateKeys();
    if (!(KeyPressed(KEY_A) || 
          KeyPressed(KEY_B) || 
          KeyPressed(KEY_START) || 
          KeyPressed(KEY_SELECT) ||
          KeyPressed(KEY_RIGHT) || 
          KeyPressed(KEY_LEFT) || 
          KeyPressed(KEY_UP) || 
          KeyPressed(KEY_DOWN) || 
          KeyPressed(KEY_R) || 
          KeyPressed(KEY_L)  
        ) )
    {
      break;
    }
  }
}
  
// C entry for the menu. Careful, interrupts are off at the mo'
#if INCLUDE_DEBUG
int needDebug = 0;
#endif

extern void dis6502(unsigned long offset, char *dest, int flags);

void doMenu() {
  initMenu();
  
  initMenuGui();
#if INCLUDE_DEBUG
  if (needDebug) {
    T6502State *state;
    state = (T6502State*)&Asm6502State;
    char str[256];
    siprintf(str,"PC=0x%x A=0x%x X=0x%x Y=0x%x ",
              state->programCounter,state->accumulator,
              state->XReg,state->YReg);
    printTxt(str,0,0);
    siprintf(str," S=0x%x PSR=0x%x ",stack|0x100,state->PSR);
    printTxt(str,3,1);
    setRowPalette(0,5);
    setRowPalette(1,5);

    // disassembled routine:
    // flags = 2, print opcode byte
    
    dis6502(state->programCounter, str, 2);
    printTxt(str,0,2);
    setRowPalette(2,5);
    
    needDebug = 0;
    g_hiliteMenuItem = MENU_CNT -1;
      
  }
#endif
  returnIRQ = (u32)irqarm;
  while (menuStatus!=MENU_EXIT) {
    wait_vsync_int();
    updateGui();
    menuKeys();
    
  }
  // wait for all keys to be let go of...
  wait_for_letgo();
  
  // finish off the menu and return
  endMenu();
}


// called every v blank..
void menuKeys(void) {
  
  if (KeyPressedNoBounce(KEY_B,MENU_KEY_DELAY)) {
    if (menuStatus == MAIN_MENU) {
      menuStatus = MENU_EXIT;
    }
    else {
      initMenuGui();
    }
  }
  
  if (KeyPressedNoBounce(KEY_A,MENU_KEY_DELAY)) {
    
    if (menuStatus == MAIN_MENU) {
      // select item selected
      g_theMenus[g_hiliteMenuItem].createMenu();
    } else {
      AHandler();
    }
    
  }
  
  if (KeyPressedNoBounce(KEY_UP,MENU_KEY_DELAY)) {
    if (g_hiliteMenuItem)
      g_hiliteMenuItem--;
    else
      g_hiliteMenuItem = g_maxItem-1;
  }
  if (KeyPressedNoBounce(KEY_DOWN,MENU_KEY_DELAY)) {
    g_hiliteMenuItem++;
    if (g_hiliteMenuItem == g_maxItem)
      g_hiliteMenuItem = 0;
  }
  
  if (KeyPressedNoBounce(KEY_LEFT,MENU_KEY_DELAY)) {
    if (menuStatus != MAIN_MENU) {
      leftHandler();
    }
    
  }
  
  if (KeyPressedNoBounce(KEY_RIGHT,MENU_KEY_DELAY)) {
    if (menuStatus != MAIN_MENU) {
      rightHandler();
    }
  }
}


//void defaultLeftHandler(void) {
void emptyHandler(void) {
  //decrease the current value
  
  
}
/*
void defaultRightHandler(void) {
  //increase the current value
  
}
*/

void displayKey(int key) {
  
  clearTextRow(MENU_Y_OFFSET+key);
  char buffer[32];
  if (key == g_optionCount[0] - 1) {
    siprintf(buffer, "%s %s", g_keyOptions[key].cName, apply_vk_press?"[-ON-]":"[-OFF-]" );
  } else {
    // then do the key thing
    if (mappedKeys[key] <= MAX_SPECIAL_KEY) {
      siprintf(buffer, "%s %s", g_keyOptions[key].cName, specialsNames[mappedKeys[key]]);
    }
    else if (mappedKeys[key] > 0x20 && mappedKeys[key] < 0x61) {
      siprintf(buffer, "%s %c", g_keyOptions[key].cName, mappedKeys[key]);
    }
    else if (mappedKeys[key] == 0x20) {
      siprintf(buffer, "%s [SPACE]", g_keyOptions[key].cName);
    } else if (mappedKeys[key] >= 0x61 && mappedKeys[key] <= 0x6a) {
      siprintf(buffer, "%s F%d", g_keyOptions[key].cName, mappedKeys[key]-0x61);
    } else if (mappedKeys[key] == 0xff) {
      siprintf(buffer, "%s [VKB]", g_keyOptions[key].cName);
    }
  }
  printTxt(buffer, 2,MENU_Y_OFFSET+key);
}

void increaseKey(void) {
  
  if (mappedKeys[g_hiliteMenuItem] >= MAX_ASCII_KEY)
    return;
  
  if (mappedKeys[g_hiliteMenuItem] == MAX_SPECIAL_KEY )
      mappedKeys[g_hiliteMenuItem] = 0x20;
  else
    mappedKeys[g_hiliteMenuItem]++;
  if (mappedKeys[g_hiliteMenuItem] >= MAX_ASCII_KEY) {
    mappedKeys[g_hiliteMenuItem] = 0xff;
  }
//  set_vk_selected(mappedKeys[g_hiliteMenuItem]);
  displayKey(g_hiliteMenuItem);
}

void decreaseKey(void) {
  
  if (mappedKeys[g_hiliteMenuItem] == 0)
    return;
  
  if (mappedKeys[g_hiliteMenuItem] == 0x20)
    mappedKeys[g_hiliteMenuItem] = MAX_SPECIAL_KEY;
  else {
    mappedKeys[g_hiliteMenuItem]--;
  }
  if (mappedKeys[g_hiliteMenuItem] == 0xfe) {
    mappedKeys[g_hiliteMenuItem] = 0x6a;
  }
  
  displayKey(g_hiliteMenuItem);
}


void selectUserKey(void);

void changeKeys(void) {
  // change the keymapping..
  menuStatus = CHANGE_KEY_MENU;
  clearText();
  leftHandler = decreaseKey;
  rightHandler = increaseKey; 
  AHandler = selectUserKey;
  int key;
  for (key = 0; key < g_optionCount[0]; key++) {
    displayKey(key);
  }
  
  g_hiliteMenuItem = 0;
  g_maxItem = g_optionCount[0];
  
//  gbaprintf("Max items %d",g_optionCount[0]);
  
}

void selectUserKey(void) {
  // draw a keyboard to select the keys...
  if (g_hiliteMenuItem < 10) {
    int tmp_apply_vk_press = apply_vk_press;
    show_vkb();
    set_vk_selected(mappedKeys[g_hiliteMenuItem]);
    // change the keypad handling so now it 
    // takes up/down left/right as moving around
    // the virtual kb.
    wait_for_letgo ();
    REG_IME = 0x00;
    REG_INTERUPT = (u32)VKInterruptHandler;
    REG_IME = 0x01;
    int accepted = 0;
    for (;;) {
      if (KeyPressedNoBounce(KEY_B,5)) {
	break;
      }
      if (KeyPressedNoBounce(KEY_A,5)) {
	accepted = 1;
	break;
      }
    }
    REG_IME = 0x00;
    REG_INTERUPT = (u32)MenuInterruptHandler;
    REG_IME = 0x01;
    if (accepted) {
      // change key...
      int r,c,k;
      get_vk_selected(&r,&c,&k);
      mappedKeys[g_hiliteMenuItem] = k;
    } 

    int tmp = g_hiliteMenuItem;
    changeKeys();
    g_hiliteMenuItem = tmp;
    apply_vk_press = tmp_apply_vk_press;
  } else {
   if (apply_vk_press) {
     // turn off..
     returnIRQ = (u32)irqarm;
     apply_vk_press = 0;
   } else {
     returnIRQ = (u32)VKInterruptHandler;
     apply_vk_press = 1;
   }
   displayKey(g_hiliteMenuItem);
  } 
}

void print_video_option(int opt, int value) {
  clearTextRow(MENU_Y_OFFSET+opt);
  if (opt < 2) {
    printTxt(g_videoOptions[opt].cName, 2,MENU_Y_OFFSET+opt );
  } else {
    char buffer[32];
    siprintf(buffer, "%s %s", g_videoOptions[opt].cName, value?"ON":"OFF");
    printTxt(buffer, 2,MENU_Y_OFFSET+opt );
  }

  
}
#if 0
void print_video_option(int opt, int value) {
  char buffer[32];
  clearTextRow(MENU_Y_OFFSET+opt);
  if (opt != 2) {
    siprintf(buffer, "%s 0x%x", g_videoOptions[opt].cName, value);
  } else {
    siprintf(buffer, "%s %s", g_videoOptions[opt].cName, value?"ON":"OFF");
  }

  printTxt(buffer, 2,MENU_Y_OFFSET+opt );
}

void videoChanger(int increment) {
  if (g_hiliteMenuItem == 0) {
    // alter the X scaling you crazy guy!
    if (rotData[0].pa) {
      //rotData[0].pa = 256;
      setSpriteScreen(rotData[0].pa+increment,0);
      print_video_option(g_hiliteMenuItem, rotData[0].pa);
    }
  } else if (g_hiliteMenuItem == 1) {
    // alter the Y scaling you crazy guy!
    if (rotData[0].pd) {
      setSpriteScreen(0,rotData[0].pd+increment);
      print_video_option(g_hiliteMenuItem, rotData[0].pd);
    } 
  } else if (g_hiliteMenuItem == 2) {
    // Auto adjust X & Y.
    if (!g_autoScale) {
      g_autoScale = 1;
      autoScale();
    } else {
      // set scale back to default...
      g_autoScale = 0;
      setSpriteScreen(128,256);
    }
    videoSettings();
    g_hiliteMenuItem = 2;
  } else if (g_hiliteMenuItem == 3) {
    // X offset adjust
    adjustOffset(increment,0);
    print_video_option(g_hiliteMenuItem, sprites[0].attribute1 & 0x1ff);
  } else if (g_hiliteMenuItem == 4) {
    // Y offset adjust
    adjustOffset(0,increment);
    print_video_option(g_hiliteMenuItem, sprites[0].attribute0&0xff);
  }
}
void videoLeftHandler(void) {
  //decrease the current value
  int increment = KeyPressed(KEY_L)?-5:-1;
  videoChanger(increment);
/*
  if (g_hiliteMenuItem == 0) {
    // alter the X scaling you crazy guy!
    if (rotData[0].pa) {
      //rotData[0].pa = 256;
      setSpriteScreen(rotData[0].pa-increment,0);
      print_video_option(g_hiliteMenuItem, rotData[0].pa);
    }
  } else if (g_hiliteMenuItem == 1) {
    // alter the Y scaling you crazy guy!
    if (rotData[0].pd) {
      setSpriteScreen(0,rotData[0].pd-increment);
      print_video_option(g_hiliteMenuItem, rotData[0].pd);
    } 
  } else if (g_hiliteMenuItem == 2) {
    // Auto adjust X & Y.
    if (!g_autoScale) {
      g_autoScale = 1;
      autoScale();
    } else {
      // set scale back to default...
      g_autoScale = 0;
      setSpriteScreen(128,256);
    }
    print_video_option(g_hiliteMenuItem, g_autoScale);
  } else if (g_hiliteMenuItem == 3) {
    // X offset adjust
    adjustOffset(-increment,0);
    print_video_option(g_hiliteMenuItem, sprites[0].attribute1 & 0x1ff);
  } else if (g_hiliteMenuItem == 4) {
    // Y offset adjust
    adjustOffset(0,-increment);
    print_video_option(g_hiliteMenuItem, sprites[0].attribute0&0xff);
  }
  */
}

void videoRightHandler(void) {
  //increase the current value
  int increment = KeyPressed(KEY_L)?5:1;
  videoChanger(increment);
  /*
  if (g_hiliteMenuItem == 0) {
    // alter the X scaling you crazy guy!
    if (rotData[0].pa < 255) {
      setSpriteScreen(rotData[0].pa+increment,0);
      print_video_option(g_hiliteMenuItem, rotData[0].pa);
    }
  } else if (g_hiliteMenuItem == 1) {
    // alter the Y scaling you crazy guy!
    if (rotData[0].pd < 511) {
      setSpriteScreen(0,rotData[0].pd+increment);
      print_video_option(g_hiliteMenuItem, rotData[0].pd);
    } 
  } else if (g_hiliteMenuItem == 2) {
    // Auto adjust X & Y.
    if (!g_autoScale) {
      g_autoScale = 1;
      autoScale();
    } else {
      // set scale back to default...
      g_autoScale = 0;
      setSpriteScreen(128,256);
    }
    print_video_option(g_hiliteMenuItem, g_autoScale);
  } else if (g_hiliteMenuItem == 3) {
    // X offset adjust
    adjustOffset(increment,0);
    print_video_option(g_hiliteMenuItem, sprites[0].attribute1&0x1ff);
  } else if (g_hiliteMenuItem == 4) {
    // X offset adjust
    adjustOffset(0,increment);
    print_video_option(g_hiliteMenuItem, sprites[0].attribute0&0xff);
  } */
}
#endif

void videoA(void) {
  switch (g_hiliteMenuItem) {
    case 0:
      change_video_scale();
      videoSettings();
      g_hiliteMenuItem = 0;
      break;
    case 1:
      change_video_position();
      videoSettings();
      g_hiliteMenuItem = 1;
      break;
    case 2:
      if ((sprites[0].attribute0 & SIZE_DOUBLE) != 0) {
        autoScale();
      } else {
        // set scale back to default...
        setSpriteScreen(128,256);
      }
      print_video_option(g_hiliteMenuItem,
          (sprites[0].attribute0 & SIZE_DOUBLE) != 0);
      break;
    case 3:
      if (copyColours == 0) {
	copyColours = COPY_COLOURS;
      }
      else {
	copyColours = 0;
      }
      print_video_option(g_hiliteMenuItem, copyColours);
    default: 
      break;
  }
}

// display the video options menu
void videoSettings  (void) {
  // Check that we aren't in Teletext mode.
  // Can't adjust that mode, sadly.
  
  if (((VideoULA_ControlReg&2)==0)) {
    clearText();
    //leftHandler = videoLeftHandler;
    //rightHandler = videoRightHandler;
    //AHandler = emptyHandler;
    leftHandler = emptyHandler;
    rightHandler = emptyHandler;
    AHandler = videoA;
    // each entry is uniqueish, so need to deal with each seperately
    /*
    print_video_option(0, rotData[0].pa);
    print_video_option(1, rotData[0].pd);
    print_video_option(2, g_autoScale);
    print_video_option(3,sprites[0].attribute1&0x1ff );
    print_video_option(4,sprites[0].attribute0&0xff );
    */
    print_video_option(0, 0);
    print_video_option(1, 0);
    print_video_option(2, (sprites[0].attribute0 & SIZE_DOUBLE) != 0 );
    print_video_option(3, copyColours);

    g_hiliteMenuItem = 0;
    g_maxItem = g_optionCount[1];

    menuStatus = VIDEO_MENU;
  }
}

void debugSettings  (void) {
  
//  // steps a frame...  
#if INCLUDE_DEBUG
  needDebug = 1;
  UpdateKeys();
  menuStatus = MENU_EXIT;
  needKeyInterrupt = GO_MENU;
#endif
}

void doLoad(void) {

  if (load()) {
    clearText(); 
    AHandler = emptyHandler;
    menuStatus = LOAD_MENU;
    printTxt("Loaded", 2,MENU_Y_OFFSET);
    g_hiliteMenuItem = 0;
    g_maxItem = 0;
  } else {
    clearTextRow(19);
    printTxt("Data to load not found!", 0,19);
    setRowPalette(19,1);
  }
}

void doSave(void) {
  if (have_save_data() && !KeyPressed(KEY_R)) {
    clearTextRow(19);
    printTxt("Use R+A to overwrite old save", 0,19);
    setRowPalette(19,1);
    return;
  }
  
  clearText(); 
  save();
  AHandler = emptyHandler;
  menuStatus = SAVE_MENU;
  printTxt("Saved", 2,MENU_Y_OFFSET);
  g_hiliteMenuItem = 0;
  g_maxItem = 0;

}

void gmLeft(void) {
  if (selectedGameIndex >= MAX_GAMES_SHOWN) {
    clearText(); 
    g_hiliteMenuItem = 0;
    g_maxItem = MAX_GAMES_SHOWN;
    selectedGameIndex-=MAX_GAMES_SHOWN;
    createGameList(selectedGameIndex,selectedGameType);
  }
}

void gmRight(void) {
  // first find out how many games we have...
  int total = getGameCount(selectedGameType); 
  // now increase to the next page
  if ((selectedGameIndex + MAX_GAMES_SHOWN) < total ) {
    clearText(); 
    g_hiliteMenuItem = 0;
    selectedGameIndex+=MAX_GAMES_SHOWN;
    createGameList(selectedGameIndex,selectedGameType);
    g_maxItem = ((selectedGameIndex + MAX_GAMES_SHOWN) < total)?MAX_GAMES_SHOWN:(total-selectedGameIndex);
  }
}

void selectGame(void) {
  
  // load the current selected game
  //u32 firstGame, firstGameEnd;
  // replace this with a menu parsing thingy..
  // format of the header is:
  unsigned int gameadr, gameadrEnd;
  getGameToLoad(&gameadr, &gameadrEnd, 
      g_hiliteMenuItem+selectedGameIndex,
      selectedGameType);
  
  if (selectedGameType == IS_UEF) { 
    resetBeeb(0);
    LoadUEFState((u8*)gameadr,(u8*)(gameadrEnd));
    menuStatus = MENU_EXIT;
    //needKeyInterrupt = GO_MENU;
  } else {
    // is a SSD
    // turn off interrupts, set the irq handler to use later
    // set SHIFT key pressed (auto loading via BBC OS/WDFS)
    // set frame count to 180 frames (3 seconds, should be enough)
    discs = (unsigned char*)gameadr;
    if (!KeyPressed(KEY_L|KEY_R)) {
      REG_IME = 0;
      // turn the sound to back how it was?
      returnIRQ = (u32)loadingIR;
      resetBeeb(1);
      copyColours = COPY_COLOURS;
      menuStatus = MENU_EXIT;
      shiftDownFrames = 180;
      SysViaKbdState[0][0] = 1;
      KeysDown = 1;
      needKeyInterrupt = 1;
      REG_IME = 1;
    } else {
      // just insert the disc...
      clearTextRow(19);
      printTxt("Disc inserted", 0,19);
      setRowPalette(19,4);
      wait_for_letgo();
    }
  }
}

void generalGameMenu(void) {
  // pogoshell can't load games, just 1 to load
  if (g_isPogoshell) 
    return;
  
  // show uef/ssd files
  int count = getGameCount(selectedGameType);
  if (count) {
    clearText(); 
    menuStatus = LOAD_MENU;
    if (count) {
      AHandler = selectGame;
      leftHandler = gmLeft;
      rightHandler = gmRight;
      g_hiliteMenuItem = 0;
      // fix for "on a blank page" when uef/ssd counts are very different
      if (selectedGameIndex > count)
        selectedGameIndex = 0;

      createGameList(g_hiliteMenuItem+selectedGameIndex,selectedGameType);  
      g_maxItem = ((selectedGameIndex + MAX_GAMES_SHOWN) < count)?MAX_GAMES_SHOWN:(count-selectedGameIndex);
    }
  }
}


void doNewGame (void) {
  // show the menu of games in the rom
  // the roms all have a ~ 36 byte header with
  // size, name, options etc built in
  // read in up to say 8 headers and display them >>
   
  selectedGameType = IS_UEF;
  generalGameMenu();
}

void loadDisc(void) {
  // similar to doNewGame, except this time we are 
  // looking for SSDs instead of UEF save states
  selectedGameType = IS_SSD;
  generalGameMenu();
}

// exit the emulator
void doExit         (void) {
  // copy reset code to EWRAM and execute
  REG_DMA3SAD = (u32) ((u32)reset_gba);
  REG_DMA3DAD = (u32) 0x02000000;
  REG_DMA3CNT = (0xb8) | DMA_32NOW;
  REG_IME = 0x00;
  CallArm((FunctionPtr) 0x02000000);
}

void resetAHandler(void) {
  switch (g_hiliteMenuItem) {
    case 0: // reset 1
      resetBeeb(1);
      menuStatus = MENU_EXIT;
      break;
    case 1:
      resetBeeb(0);
      menuStatus = MENU_EXIT;
      break;
    case 2:
      doExit();
      break;
    default:
      break;
  }
}

void doResetSubmenu(void) {
  clearText();
  leftHandler = emptyHandler;
  rightHandler = emptyHandler;
  AHandler = resetAHandler;
  int i;
  for (i = 0; i < g_optionCount[2]; i++) {
    printTxt(g_resetOptions[i].cName, 2,MENU_Y_OFFSET+i);
  }
  g_maxItem = g_optionCount[2];
  g_hiliteMenuItem = 0;
  menuStatus = RESET_MENU;

}

void sramAHandler(void) {
  switch (g_hiliteMenuItem) {
    case 0: // load
      doLoad();
      break;
    case 1: // save
      doSave();
      break;
    default:
      break;
  }
}


void doSramMenu(void) {
  clearText();
  leftHandler = emptyHandler;
  rightHandler = emptyHandler;
  AHandler = sramAHandler;
  int i;
  for (i = 0; i < g_optionCount[3]; i++) {
    printTxt(g_sramOptions[i].cName, 2,MENU_Y_OFFSET+i);
  }
  g_maxItem = g_optionCount[3];
  g_hiliteMenuItem = 0;
  menuStatus = SRAM_MENU;

}

