#include "6502macros.h"
@sysvia.S

.section .iwram,"ax",%progbits
.code 32
.align 4

.extern UserVIAState
.extern SysVIAState
.extern intStatus
@@  this is what the ifr bits mean:
@@  76543210
@   |||||||+-- 0x01  keyboard interrupt, 'watched' key was pressed
@   ||||||+--- 0x02  CA1 (vblank or hblank)
@   |||||+---- 0x04  keyboard interrupt - user presses any key
@   ||||+----- 0x08  
@   |||+------ 0x10  A to D conversion (joystick)
@   ||+------- 0x20  timer 2
@   |+-------- 0x40  timer 1
@   +--------- 0x80  interrupt occurred



@@ this is SysVIATriggerCA1Int 
.global SysVIAIR
.type SysVIAIR, function
@ input r1 = value
SysVIAIR:
  stmfd   sp!,{r0-r3,lr}
  mov   r3,#2
  ldr   r0,=SysVIAState
  
via0:
  ldrb  r2,[r0,#offset_pcr]
  and   r2,r2,#1
  eors  r2,r2,r1        @ xor value
  bne   9f
	ldrb   r2,[r0,#offset_ifr]
	orr   r2,r2,#2
	strb   r2,[r0,#offset_ifr]
	bl  IFRTopBit
9:
  ldmfd   sp!,{r0-r3,lr}
  mov     pc,lr


.global IFRTopBitUser
.type IFRTopBitUser, function
IFRTopBitUser:
  ldr       r0,=UserVIAState
  mov       r3,#4       @ user = 1<<2
  b         tb0
.global IFRTopBitSys
.type IFRTopBitSys, function
IFRTopBitSys:
  @if (SysVIAState.ifr&(SysVIAState.ier&0x7f))
  @@ make r1 hold this ^ 
  ldr       r0,=SysVIAState
  mov       r3,#2        @ sys = 1<<1
.type IFRTopBit, function
IFRTopBit:
tb0:  
  ldrb      r1,[r0,#offset_ifr]
  ldrb      r2,[r0,#offset_ier]
  and       r2,r2,#0x7f
  ands      r2,r2,r1
  
  @  SysVIAState.ifr|=0x80;
  @else
  @  SysVIAState.ifr&=0x7f;
  @intStatus&=~(1<<sysVia);
  @intStatus|=((SysVIAState.ifr & 128)?(1<<sysVia):0);
  
  orrne     r1,r1,#0x80
  andeq     r1,r1,#0x7f
  strb      r1,[r0,#offset_ifr]
  ldr       r2,=intStatus
  ldr       r1,[r2]
  bic       r1,r1,r3          @&=~1<<1
  orrne     r1,r1,r3
  str       r1,[r2]
  mov     pc,lr

.macro  callIFRReset
  stmfd   sp!,{r1,r3,lr}     @ store regs cos ifr fcuks em up
  bl   IFRTopBitSys        @ this assumes r1 points to the state
  ldmfd  sp!,{r1,r3,lr}  
.endm

.extern SysViaKbdState
.extern KBDCol
.extern KBDRow
.macro slowDBRead
@ return slow databus read
  ldrb     bbc_nz,[r1,#offset_ora]        @ this is the keyboard value just set?
  ldrb     r0,[r1,#offset_ddra]   @ ddra is generally #0x7f
  and      bbc_nz,bbc_nz,#0x7f

@ bbc_nz holds result at this stage, if the key index set by 
@ slow db write is pressed, need to set bit 7 of bbc_nz
@ first find which key we want (can freely use r0,r1,r2) 
@ remember to make SysViaKbdState[16][8]!!


  ldr      r0,=KBDCol
  ldr      r0,[r0]                    @ r0 = KBDCol 
  cmp      r0,#0xa                    @ Last Ninja needs this
  bpl      0f
  
  ldr      r2,=KBDRow
  ldr      r2,[r2]                    @ r2 = KBDRow 
  add      r0,r2,r0,lsl #3            @ r0 = Row + 8*KBDCol

  ldr      r1,=SysViaKbdState
  ldrb     r2,[r1,r0]
  cmp      r2,#0
  orrne    bbc_nz, bbc_nz, #0x80
  orrne    bbc_nz, bbc_nz, #0x80000000   @ seriously. This fixes some odd keyboard reading (ROR bug?)
  @movne    bbc_nz, bbc_nz, lsl #24
  @movne    bbc_nz, bbc_nz, asr #24
0:
@ all done!
.endm

.macro SyncCycles
  tst     bbc_cycles,#1
  subne     bbc_cycles,bbc_cycles,#2
  subeq     bbc_cycles,bbc_cycles,#1
.endm

.global sysVIARead
.type sysVIARead,function
@ register memptr holds the address to read
@ returns the value read in bbc_nz
@ use regs r0 - r2, anything else needs pushing/popping
sysVIARead:
  SyncCycles
  and    r0,memptr,#0xf
  ldr     r1,=SysVIAState
  adr     r2,LsysReadJumps
  ldr     pc,[r2,r0,lsl#2]     @ 4 bytes per symbol
  
case0:  
  @IRB read
  @ Clear bit 4 of IFR from ATOD Conversion
  ldrb     r0,[r1,#offset_ifr]
  and      r0,r0,#0xef      @~16
  strb     r0,[r1,#offset_ifr]
  ldrb     r0,[r1,#offset_orb]
  ldrb     r2,[r1,#offset_ddrb]
  and     bbc_nz, r0,r2
  orr     bbc_nz,bbc_nz,#(16 + 32 + 192)
  mov     pc,lr

case1:  
  ldrb     r0,[r1,#offset_ifr]
  and     r0,r0,#0xfc
  strb     r0,[r1,#offset_ifr]
  callIFRReset
  slowDBRead
  mov     pc,lr

case2:
  @return(SysVIAState.ddrb)
  ldrsb   bbc_nz,[r1,#offset_ddrb]
  mov     pc,lr

case3:
   @ return(SysVIAState.ddra);
   ldrsb   bbc_nz,[r1,#offset_ddra]
   mov     pc,lr

case4:
    @ /* Timer 1 lo counter */
    
     
    @ SysVIAState.ifr &=0xbf; /* clear timer 1 ifr */
     ldrb  r2,[r1,#offset_ifr]
     bic   r2,r2,#0x40
     strb  r2,[r1,#offset_ifr]
     
    
     callIFRReset
    
    @ldrb  r2,[r1,#offset_timer1c]
    @ldrb  r0,[r1,#offset_timer1l]
    ldr   r2,[r1,#offset_timer1c]
    ldrh  r0,[r1,#offset_timer1l]
    @if (SysVIAState.timer1c<=(SysVIAState.timer1l*2))
    subs  r0,r2,r0,lsl #1
    @ return SysVIAState.timer1c / 2;
    movmi     bbc_nz, r2, lsr #1    @ timer1c/2
    @ else return 0xff;
    movpl     bbc_nz,#0xff
    and     bbc_nz,bbc_nz,#0xff
    mov     pc,lr
    

case5:
@  case 5: /* Timer 1 ho counter */
@  tmp=SysVIAState.timer1c /512;
@  if (SysVIAState.timer1c<=(SysVIAState.timer1l*2))
@  return(tmp & 0xff);
@  else
@  return(0xff);
@  ldrsb     bbc_nz,[r1,#offset_timer1l+1]   @ can I do this?
  ldr       r2,[r1,#offset_timer1c]
  ldrh      r0,[r1,#offset_timer1l]
  subs      r0,r2,r0,lsl #1
  movmi     bbc_nz, r2, lsr #9    @ timer1c/512
  movpl     bbc_nz,#0xff
  mov       pc,lr

case6:
@  case 6:return(SysVIAState.timer1l & 0xff);
  ldrsb     bbc_nz,[r1,#offset_timer1l]  
  mov       pc,lr

case7: 
@      /* Timer 1 ho latch */
@      return((SysVIAState.timer1l / 256) & 0xff);
  ldrsb     bbc_nz,[r1,#offset_timer1l+1]  
  mov       pc,lr

case8:
  @ Timer 2 lo counter - like case4.
    
    ldrb  r2,[r1,#offset_ifr]
    bic   r2,r2,#0x20       @ timer 2 ir bit
    strb  r2,[r1,#offset_ifr]
    callIFRReset
    
    ldr   r2,[r1,#offset_timer2c]
    ldrh  r0,[r1,#offset_timer2l]
    subs  r0,r2,r0,lsl #1
    
    movmi     bbc_nz, r2, lsr #1    @ timer1c/2
    movpl     bbc_nz,#0xff
    and     bbc_nz,bbc_nz,#0xff
    mov     pc,lr
   
.type case9,function 
case9:
  @ timer 2 - case5 - looks like there's a bug in beeb em here
  @ no! A bug in my em!
  @ if (SysVIAState.timer1c<=(SysVIAState.timer1l*2))
  @   return((SysVIAState.timer2c / 512) & 0xff);
  @    else
  @  return(0xff);
  ldr      r2,[r1,#offset_timer2c]
  ldrh     r0,[r1,#offset_timer2l]
  subs     r0,r2,r0,lsl #1
  movmi    bbc_nz, r2, asr #9    @ timer2c/512
@  andmi    bbc_nz, bbc_nz, #0xff 
@  orrmi    bbc_nz,bbc_nz,r2,lsl #16
  movpl    bbc_nz,#0xff000000
  movpl    bbc_nz,bbc_nz,asr#24
@  and      r0,bbc_nz,#0xff
  mov      pc,lr

.type cased,function  
cased:
    callIFRReset
    ldrsb   bbc_nz,[r1,#offset_ifr]    @ byte extend for magic mushrooms
    ldrb    r0,[r1,#offset_ifr]     
    mov     pc,lr
casee:
    @return(SysVIAState.ier | 0x80);
    ldrb    bbc_nz,[r1,#offset_ier]
    orr     bbc_nz,bbc_nz,#0x80                @ 0x0000008n
    orr     bbc_nz,bbc_nz,bbc_nz,lsl#24        @ 0x8n00008n
    and     r0,bbc_nz,#0xff                    @ for load...
    mov     pc,lr

casef:
  @ return slow databus read
  @ if KbdOP()  bbc_nz|0x80
  slowDBRead
  mov     pc,lr  
  
casex:
    @ buggy bug bug
    mov     bbc_nz,#0xff000000
    mov     bbc_nz,bbc_nz,asr #24
    mov     r0,#0xff
    mov     pc,lr  

LsysReadJumps:
.word   case0
.word   case1
.word   case2
.word   case3
.word   case4   @timer1 lo
.word   case5   @timer1 hi
.word   case6   @timer1 lo latch
.word   case7   @timer1 hi latch
.word   case8   @timer2 lo
.word   case9   @timer2 hi
.word   casex
.word   casex     @ non existent?
.word   casex     @ non existent?
.word   cased
.word   casee
.word   casef



.global sysVIAWrite
.type sysVIAWrite,function
@ register memptr holds the address to write to
@ use regs r0 holds value to write
@ uses regs r1,r2, r5, r6, r7
sysVIAWrite:
  SyncCycles
  stmfd	sp!, {r5-r7,lr}
  ldr     r1,=SysVIAState


  and     r5,memptr,#0xf
  adr     r2,LsysWriteJumps
  ldr     pc,[r2,r5,lsl#2]     @ 4 bytes per symbol

syswriteCase0:  
                strb    r0,[r1,#offset_orb]
                @ic32 write
                bl IC32Write
                
		ldr     r1,=SysVIAState
                @ this is for joystick irq so not used..
                ldrb    r0,[r1,#offset_ifr]   
                @and     r0,r0,#0xef       @~16
                @strb    r0,[r1,#offset_ifr]
                
                @ some other fruity stuff
                @ that looks a bit hacky...
                tst       r0,#1
                beq       0f @ noIC32WriteKeyIR

                ldrb      r2,[r1,#offset_pcr]
                tst       r2,#2
                bne       0f @noIC32WriteKeyIR
                bic       r0,r0,#1
                strb      r0,[r1,#offset_ifr] 
		mov       memptr,r3
                bl        IFRTopBitSys 
		mov       r3,memptr
noIC32WriteKeyIR:
                
                b     0f
syswriteCase1:  
                strb     r0,[r1,#offset_ora]
                @slow db write of r0
                adr     lr,0f
                ldr     pc,=SlowDataBusWrite
                @bl      SlowDataBusWrite  
                @b     0f
syswriteCase2:  
                strb     r0,[r1,#offset_ddrb]
                b     0f
syswriteCase3:  
                @ddr
                strb     r0,[r1,#offset_ddra]
                b     0f
                
syswriteCase6:  @@@@ fall through  @@@ b     0f
syswriteCase4:  @timer1
                @ SysVIAState.timer1l&=0xff00;
                @ SysVIAState.timer1l|=(Value & 0xff);
                strb     r0,[r1,#offset_timer1l]
                b     0f
.type syswriteCase5, function
.global syswriteCase5
syswriteCase5:  @more timer 1
              @ SysVIAState.timer1l&=0xff;
              @ SysVIAState.timer1l|=(Value & 0xff)<<8;
                ldrb  r2,[r1,#offset_timer1l]
                orr   r2,r2,r0,lsl#8
                strh  r2,[r1,#offset_timer1l]
              @ SysVIAState.timer1c=SysVIAState.timer1l * 2;
                mov   r2,r2,lsl#1
                str  r2,[r1,#offset_timer1c]
              @ SysVIAState.ifr &=0xbf; /* clear timer 1 ifr */
                ldrb  r2,[r1,#offset_ifr]
                bic   r2,r2,#0x40
                strb  r2,[r1,#offset_ifr]
              @ If PB7 toggling enabled, then lower PB7 now 
                ldrb  r2,[r1,#offset_acr]
@                tst   r2,#128
@                @ if (SysVIAState.acr & 128) {
@                @ SysVIAState.orb&=0x7f;
@                @ SysVIAState.irb&=0x7f;
@                @ };
@                ldrbne  r2,[r1,#offset_orb]
@                bicne   r2,r2,#0x80
@                strbne  r2,[r1,#offset_orb]
@                ldrbne  r2,[r1,#offset_irb]
@                bicne   r2,r2,#0x80
@                strbne  r2,[r1,#offset_irb]
              @ UpdateIFRTopBitSys();
                callIFRReset
              @ SysVIAState.timer1hasshot=0;
                mov   r2,#0
                str  r2,[r1,#offset_timer1shot]
                b     0f

syswriteCase7:  @timer1 hi byte
              @ SysVIAState.timer1l&=0xff;
              @ SysVIAState.timer1l|=(Value & 0xff)<<8;
                ldrb  r2,[r1,#offset_timer1l]
                orr   r2,r2,r0,lsl#8
                strh  r2,[r1,#offset_timer1l]
                b     0f
syswriteCase8:  @timer2
              @ SysVIAState.timer2l&=0xff00;
              @ SysVIAState.timer2l|=(Value & 0xff);
                ldr   r2,[r1,#offset_timer2l]
                and   r2,r2,#0xff00
                orr   r0,r2,r0
                strh  r0,[r1,#offset_timer2l]
                b     0f
syswriteCase9:  @timer2
		@SysVIAState.timer2l&=0xff;
		@SysVIAState.timer2l|=(Value & 0xff)<<8;
                ldrb  r2,[r1,#offset_timer2l]
                orr   r2,r2,r0,lsl#8
                strh  r2,[r1,#offset_timer2l]
		@SysVIAState.timer2c=SysVIAState.timer2l * 2;
                mov   r2,r2,lsl#1 
                str r2,[r1,#offset_timer2c]
		@SysVIAState.ifr &=0xdf; /* clear timer 2 ifr */
                ldrb  r2,[r1,#offset_ifr]
@               bic   r2,r2,#0x20
                and r2,r2,#0xdf
                strb  r2,[r1,#offset_ifr]
		@UpdateIFRTopBitSys();
                callIFRReset
		@SysVIAState.timer2hasshot=0;
                mov   r2,#0
                str  r2,[r1,#offset_timer2shot]
                b     0f

syswriteCasea:  @shift reg
                b     0f
                
syswriteCaseb:  strb     r0,[r1,#offset_acr]
                b     0f
syswriteCasec:  strb     r0,[r1,#offset_pcr]
                b     0f
syswriteCased:  ldrb    r2,[r1,#offset_ifr]
                bic     r2,r2,r0
                strb    r2,[r1,#offset_ifr]   
                callIFRReset
                b     0f
syswriteCasee:  @ier
                ldrb  r5,[r1,#offset_ier]
                tst   r0,#0x80
                orrne r5,r5,r0        @ if value & 0x80 NZ
                biceq r5,r5,r0        @ if value & 0x80 == 0
                and   r0,r5,#0x7f
                strb  r0,[r1,#offset_ier]
                callIFRReset
                b     0f

.extern IC32State
.type syswriteCasef,function
syswriteCasef:  
        strb    r0,[r1,#offset_ora]
        bl SlowDataBusWrite
        @ call this ONLY if IC32&8 == 0
        ldr       r2,=IC32State @ LSlowDataBusWrite+8   @ic32state
        ldr       r0,[r2] 
        tst       r0,#8
        bleq      beebKeyIR
        @bl      beebKeyIR 
0:  
  ldmfd	sp!, {r5-r7,lr}
  mov   pc,lr
  

LsysWriteJumps:
.word   syswriteCase0   @ Clear bit 4 of IFR from ATOD Conversion
.word   syswriteCase1
.word   syswriteCase2
.word   syswriteCase3
.word   syswriteCase4   @timer1 set
.word   syswriteCase5   @play about with timer
.word   syswriteCase6   @timer1 set
.word   syswriteCase7   @timer1 hi latch
.word   syswriteCase8   @timer2 lo
.word   syswriteCase9   @clear timer 2 ifr
.word   syswriteCasea   @ shift register
.word   syswriteCaseb   @ acr
.word   syswriteCasec   @ pcr
.word   syswriteCased   @ clear ifr
.word   syswriteCasee   @ ier
.word   syswriteCasef   @ ora


@ now for the user via stuff...
.macro  callIFRResetUser
  stmfd   sp!,{r1,r3,lr}     @ store regs cos ifr fcuks em up
  bl     IFRTopBitUser
  ldmfd   sp!,{r1,r3,lr}  
.endm
  

.global userVIARead
.type userVIARead,function
@ register memptr holds the address to read
@ returns the value read in bbc_nz
@ use regs r0 - r2, anything else needs pushing/popping
userVIARead:
  SyncCycles
  and     r0,memptr,#0xf
  ldr     r1,=UserVIAState
  adr     r2,LuserReadJumps
  ldr     pc,[r2,r0,lsl#2]     @ 4 bytes per symbol
  
userReadCase0:  
  @IRB read..
  @ tmp=(UserVIAState.orb & UserVIAState.ddrb) | (UserVIAState.irb & (~UserVIAState.ddrb));
  ldrb     r0,[r1,#offset_orb]
  ldrb     r2,[r1,#offset_ddrb]
  and      bbc_nz,r0,r2
  ldrb     r0,[r1,#offset_irb]
  eor      r2,r2,#0xff
  and      r0,r0,r2
  orr      bbc_nz,bbc_nz,r0
  mov     pc,lr

userReadCase1:  
  ldrb     r0,[r1,#offset_ifr]
  and     r0,r0,#0xfc
  strb     r0,[r1,#offset_ifr]
  callIFRResetUser
  mov     bbc_nz,#0xff      @ equivalent to slow databus read
  mov     pc,lr

userReadCase4:
    @ /* Timer 1 lo counter */
    @tmp=SysVIAState.timer1c / 2;
    ldr     bbc_nz,[r1,#offset_timer1c]
    @ SysVIAState.ifr &=0xbf; /* clear timer 1 ifr */
    ldrb  r2,[r1,#offset_ifr]
    bic   r2,r2,#0x40
    strb  r2,[r1,#offset_ifr]
     
    @UpdateIFRTopBitSys();
    callIFRResetUser
    @oy, what the hell is all this?
    @if (SysVIAState.timer1c<=(SysVIAState.timer1l*2)) 
    @return(tmp & 0xff);
    @else
    @return(0xff);
    ldr     r0,[r1,#offset_timer1l]
    cmp     bbc_nz,r0,lsl#1     @ cmp = bbc_nz - r0*2 = c - l*2
    movmi     bbc_nz,bbc_nz,asl#23
    movmi     bbc_nz,bbc_nz,asr#24
    movpl     bbc_nz,#0x0
    subpl     bbc_nz,bbc_nz,#1
    and     r0,bbc_nz,#0xff
    mov     pc,lr


userReadCase8:
  @ Timer 2 lo counter - like case4.
    ldrsb     bbc_nz,[r1,#offset_timer2l]
    ldrb  r2,[r1,#offset_ifr]
    bic   r2,r2,#0x20       @ timer 2 ir bit
    strb  r2,[r1,#offset_ifr]
    callIFRResetUser
    mov     pc,lr
  
userReadCased:
    callIFRResetUser
    ldrsb    bbc_nz,[r1,#offset_ifr]    @ byte extend for magic mushrooms
    @ldrb    r0,[r1,#offset_ifr]     
    mov     pc,lr


LuserReadJumps:   @ uses some sys codes too, cos they do the same stuff
.word   userReadCase0
.word   userReadCase1
.word   case2
.word   case3
.word   userReadCase4   @timer1 lo
.word   case5          @timer1 hi
.word   case6   @timer1 lo latch
.word   case7   @timer1 hi latch
.word   userReadCase8   @timer2 lo
.word   case9   @timer2 hi
.word   casex
.word   casex     @ non existent?
.word   casex     @ non existent?
.word   userReadCased
.word   casee
.word   casex  
  



.global userVIAWrite
.type userVIAWrite,function
@ register memptr holds the address to write to
@ use regs r0 holds value to write
@ uses regs r1,r2, r5, r6, r7
userVIAWrite:
  SyncCycles
  stmfd	sp!, {r5-r7,lr}
  ldr     r1,=UserVIAState
.type afteruwd, function
afteruwd:
  and     r5,memptr,#0xf
  adr     r2,LuserWriteJumps
  ldr     pc,[r2,r5,lsl#2]     @ 4 bytes per symbol

userwriteCase0:  
                strb    r0,[r1,#offset_orb]
                ldrb    r0,[r1,#offset_ifr]
                ldrb    r2,[r1,#offset_pcr]
                tst     r0,#1
                beq     0f        
                tst     r2,#2
                bne     0f
                @ is there a nicer way to do this?
                @ something like:
                @ eor r2,r0,r2,lsr#1    @ r2 = ifr ^ (pcr>>1) etc
                
                and     r0,r0,#0xfe
                strb    r0,[r1,#offset_ifr]
                
                @ some other fruity stuff
                @ that looks a bit hacky...
                callIFRResetUser
                b     0f
userwriteCase1:  
                strb     r0,[r1,#offset_ora]
                ldrb    r0,[r1,#offset_ifr]   
                and     r0,r0,#0xfc
                strb    r0,[r1,#offset_ifr]
                callIFRResetUser
                b     0f


userwriteCase5:  @more timer 1
              @ SysVIAState.timer1l&=0xff;
              @ SysVIAState.timer1l|=(Value & 0xff)<<8;
                ldrb  r2,[r1,#offset_timer1l]
                orr   r2,r2,r0,lsl#8
                str   r2,[r1,#offset_timer1l]
              @ SysVIAState.timer1c=SysVIAState.timer1l * 2;
                mov   r2,r2,lsl#1
                str  r2,[r1,#offset_timer1c]
              @ SysVIAState.ifr &=0xbf; /* clear timer 1 ifr */
                ldrb  r2,[r1,#offset_ifr]
                bic   r2,r2,#0x40
                strb  r2,[r1,#offset_ifr]
              @ If PB7 toggling enabled, then lower PB7 now 
@                ldrb  r2,[r1,#offset_acr]
@                tst   r2,#128
@                @ if (SysVIAState.acr & 128) {
@                @ SysVIAState.orb&=0x7f;
@                @ SysVIAState.irb&=0x7f;
@                @ };
@                ldrbne  r2,[r1,#offset_orb]
@                bicne   r2,r2,#0x80
@                strbne  r2,[r1,#offset_orb]
@                ldrbne  r2,[r1,#offset_irb]
@                bicne   r2,r2,#0x80
@                strbne  r2,[r1,#offset_irb]
              @ UpdateIFRTopBitSys();
                mov   r2,#0
                str  r2,[r1,#offset_timer1shot]
                callIFRResetUser
              @ SysVIAState.timer1hasshot=0;
                b     0f

.type userwriteCase9,function
userwriteCase9:  @timer2
@              SysVIAState.timer2l&=0xff;
@              SysVIAState.timer2l|=(Value & 0xff)<<8;
                ldrb  r2,[r1,#offset_timer2l]
                orr   r2,r2,r0,lsl#8
                strh  r2,[r1,#offset_timer2l]
@              SysVIAState.timer2c=SysVIAState.timer2l * 2;
                mov   r2,r2,lsl#1 
                str  r2,[r1,#offset_timer2c]
@              SysVIAState.ifr &=0xdf; /* clear timer 2 ifr */
                ldrb  r2,[r1,#offset_ifr]
                bic   r2,r2,#0x20
                strb  r2,[r1,#offset_ifr]
                mov   r2,#0
                str   r2,[r1,#offset_timer2shot]
@              UpdateIFRTopBitSys();
                callIFRResetUser
@              SysVIAState.timer2hasshot=0;
                b     0f


userwriteCased:  ldrb    r2,[r1,#offset_ifr]
                bic     r2,r2,r0
                strb    r2,[r1,#offset_ifr]   
                callIFRResetUser
                b     0f
userwriteCasee:  @ier
                ldrb  r5,[r1,#offset_ier]
                tst   r0,#0x80
                orrne r5,r5,r0        @ if value & 0x80 NZ
                biceq r5,r5,r0        @ if value & 0x80 == 0
                and   r0,r5,#0x7f
                strb  r0,[r1,#offset_ier]
                callIFRResetUser
                b     0f
userwriteCasef:  strb    r0,[r1,#offset_ora]
                
0:  
  ldmfd	sp!, {r5-r7,lr}
  mov   pc,lr
  
LuserWriteJumps:
.word   userwriteCase0   @ Clear bit 4 of IFR from ATOD Conversion
.word   userwriteCase1
.word   syswriteCase2
.word   syswriteCase3
.word   syswriteCase4   @timer1 set
.word   userwriteCase5   @play about with timer
.word   syswriteCase6   @timer1 set
.word   syswriteCase7   @timer1 hi latch
.word   syswriteCase8   @timer2 lo
.word   userwriteCase9   @clear timer 2 ifr
.word   syswriteCasea   @ shift register
.word   syswriteCaseb   @ acr
.word   syswriteCasec   @ pcr
.word   userwriteCased   @ clear ifr
.word   userwriteCasee   @ ier
.word   userwriteCasef   @ ora



.type timerIR,function
timerIR:
  @ r3 contais 0 for timer 1, 1 for timer 2
  @ r2 = 0 for user, 1 for sys timers
  @ r1 = pointer to state
  
  @ if the timer has shot already, do nothing. timer shot if ifr & 20/40 is not 0
  @ will fire if timershot == 0  - might not be needed if we just check the interrupt state?
  @ ahh, need ot think about "1 shot" interrupts??
  mov      r5,r3,lsl #2               @ r5 is 0 (timer1) or 4 (timer2)
  add      r5,r5,#offset_timer1shot   @ r5 is either timer1shot or timer1shot+4 == timer2shot
  ldr      r4,[r1,r5]                 @ r4 = value of either timer shot 1 or 2...
  cmp      r4,#0
  beq      setTimerIR                 @ not shot yet, should definitely shoot...

@@@ this next part was wrong, the timer does not fire 
@ if (acr & 0x40) == 0 or if we have already "shot our load"...
@ ie. fires when timerXshot == 0 || acr & 0x40   @@@@@ 
@  ldrb     r4,[r1,#offset_ifr]
@  mov      r5,#0x40
@  tst      r4,r5,lsr r3     @ 20 or 40 depending on timer

@@ test acr... nb if timer 2, do not test anything - can only 1 shoot?
  cmp      r3,#0
  movne    pc,lr       @ if != 0, return
@@ OK, so is acr 40 set?
  ldrb     r4,[r1,#offset_acr]
  tst      r4,#0x40  
  moveq      r4,#1
  streq      r4,[r1,r5]  @ timer shot 1 or 2 = #$1
  moveq    pc,lr           @ if acr & 40 == 0, return
.type setTimerIR,function
setTimerIR:
  @@ misses stuff out ... orb^=0x80;
  @@  irb^=0x80;  @ if anyone knows what this does.. "toggle pb7"
  @ first, set timer shot, which is stored in r5
  @ this is so we don't unwittingly repeat the timer
  mov      r4,#1
  str      r4,[r1,r5]  @ timer shot 1 or 2 = #$1
  @ now, what about the mask for the ifr?
  ldrb     r4,[r1,#offset_ifr]
  mov      r5,#0x40
  orr      r4,r4,r5,lsr r3     @ 20 or 40 depending on timer
  strb     r4,[r1,#offset_ifr]
  cmp      r3,#0
  bne      afterFiddleORB
@ for timer 1, need to fiddle orb
  ldrb     r4,[r1,#offset_acr]
  tst      r4,#0x80
  beq      afterFiddleORB
  ldrb    r4,[r1,#offset_orb]
  eor     r4,r4,#0x80
  strb    r4,[r1,#offset_orb]
  ldrb    r4,[r1,#offset_irb]
  eor     r4,r4,#0x80
  strb    r4,[r1,#offset_irb]
  
afterFiddleORB:
  stmfd   sp!,{r0-r3,lr}     
  tst     r2,#1
  adr     lr,endTimerIR
  beq     IFRTopBitUser
  bne     IFRTopBitSys
  
endTimerIR:
  ldmfd   sp!,{r0-r3,lr}     
  mov       pc,lr

@ updates a timer
@ r0 = cycles per scaline
@ r1 = pointer to state
@ r2 = 0 for user, 1 for sys timers
.type updateTimer, function
.global updateTimer
updateTimer:
  @ timer1c -=ncycles
  mov      r8,lr
  ldr     r3,[r1,#offset_timer1c]
  subs     r3,r3,r0 @,lsl#1     @ - cycles
                    @ the lsl hack speeds Snapper up as well
  ldrmih    r4,[r1,#offset_timer1l]
  @ timer1c=(timer1l * 2)-(tCycles-4);    tCycles is the abs value of the cyles left over
  @addmi    r3,r3,r4,lsl#1
  movmi    r3,r4,lsl#1
  str      r3,[r1,#offset_timer1c]
  @ fire the interrupt if needed...
  movmi    r3,#0
  blmi       timerIR
  
  @ timer2c -=ncycles
  ldr     r3,[r1,#offset_timer2c]
  subs     r3,r3,r0 @,lsl#1     @ - cycles
  ldrmih    r4,[r1,#offset_timer2l]
  @ timer1c=(timer1l * 2)-(tCycles-4);    tCycles is the abs value of the cyles left over
  @addmi    r3,r3,r4,lsl#1     @ an add causes Volcano to hang...
  movmi    r3,r4,lsl#1         @ Amazingly, moving r4 to r3 fixes Volcano!
  str     r3,[r1,#offset_timer2c]
  @ fire the interrupt if needed...
  movmi     r3,#1
  blmi      timerIR
  mov   pc,r8

@ this is called every scan line 
@ a scanline is triggered every 128*3*CYCLES cycles...
@ r3 is bbc_cycles - i.e. current cycles until next event (?)
@ not sure why I subtract it here, maybe to get leftover cycles
@ when scanline fires "late"? If so, bbc_cycles is negative

.global doViaPolls
.extern timerDecrement 
.type   doViaPolls,function
doViaPolls:
  @ subtract cyclesPerScanLine from timers, send interrupts
  stmfd   sp!,{r7,r8}      @ could lose these?
  ldr   r0,=timerDecrement
  ldr   r0,[r0]
  sub   r0,r0,r3 
@  mov   r0,r0,lsl #1  @ this hack speeds Snapper up no end
                       @ seems to screw up every other game though!
  ldr   r1,=UserVIAState
  mov   r2,#0
  mov   r7,lr
  bl    updateTimer
  ldr   r1,=SysVIAState
  mov   r2,#1
  bl    updateTimer
  mov   lr,r7            @optimise - b updateTimer, intead of bl, for example
  ldmfd   sp!,{r7,r8}
  mov pc,lr


@typedef struct {
@  unsigned char ora,orb;
@  unsigned char ira,irb;
@  unsigned char ddra,ddrb;
@  unsigned char acr,pcr;
@  unsigned char ifr,ier;
@  int timer1c,timer2c; /* NOTE: Timers descrement at 2MHz and values are */
@  int timer1l,timer2l; /*   fixed up on read/write - latches hold 1MHz values*/
@  int timer1hasshot; /* True if we have already caused an interrupt for one shot mode */
@  int timer2hasshot; /* True if we have already caused an interrupt for one shot mode */
@  int timer2adjust; // Adjustment for 1.5 cycle counts, every other interrupt, it becomes 
@  // 2 cycles instead of one
@} VIAState;

.equ    offset_ora,       0
.equ    offset_orb,       1
.equ    offset_ira,       2
.equ    offset_irb,       3
.equ    offset_ddra,      4
.equ    offset_ddrb,      5
.equ    offset_acr,       6
.equ    offset_pcr,       7
.equ    offset_ifr,       8
.equ    offset_ier,       9

.equ    offset_timer1c,   12
.equ    offset_timer2c,   16

.equ    offset_timer1l,    20
.equ    offset_timer2l,    24

.equ    offset_timer1shot, 28
.equ    offset_timer2shot, 32

.pool

.end
