

#include <stdio.h>
#include <stdlib.h>
#include <iostream>

int padFile(char* bytes, int len, char* fname, int padTo) {
	// write bytes to the file - pad with 0x00 to 32 bit boundaries...
	FILE * fos = fopen(fname,"w");
	if (fos == 0) {
		std::cerr << "Output file: " << fname << " could not be written to" << std::endl;
		return 1;
	}
	//int len = bytes.length;
	int written = fwrite(bytes, 1, len, fos);
	if (written != len) {
		std::cerr << "Output file: " << fname << " couldn't write correctly" << std::endl;
		fclose(fos);
		return 1;
	}
		
	if (padTo > len) {
		int remainder = padTo - len;
		char * padding = (char*)malloc(remainder);
		if (padding == 0) {
			std::cerr << "Error allocating memory for padding" << std::endl;
			fclose(fos);
			return 1;
		}
		for (int i = 0; i < remainder; i++) {
			padding[i] = 0;
		}
		written = fwrite(padding, 1, remainder, fos);
		if (written != remainder) {
			std::cerr << "Output file: " << fname << " couldn't write padding correctly" << std::endl;
			fclose(fos);
			free( padding);
			return 1;
		}
		free( padding);
	}
	fclose(fos);
	return 0;
}

char * getFile(char* theFile, int & length) {
	FILE * file = fopen(theFile,"rb");
	if (!file) {
		return 0;
	}
	fseek(file,0,SEEK_END);
	int nBytes = ftell(file);
	fseek(file,0,SEEK_SET);
	char * sna = (char*)malloc(nBytes);
	if (!sna) {
		std::cerr << "Error allocating memory" << std::endl;
		fclose(file);
		return 0;
	}
	fread(sna,1,nBytes, file);
	fclose(file);
	length = nBytes;
	return sna;
}

int main(int argc, char ** args) 
{
	// open the file, pad to size
	// 0 = in
	// 1 = size
	// 2 = out name
	if (argc < 2) {
		std::cout << "Usage:  padder  infile  size  outfile" << std::endl;
	}
	int length;
	char * inFile = getFile(args[1], length);
	if (inFile == 0) {
		std::cerr << "Input file: " << args[1] << " could not be read" << std::endl;
		return 1;
	}
	int pad = strtol(args[2],0,0);
	if ( padFile(inFile,length, args[3],pad) ) {
		free(inFile);
		return 1;
	}
	free(inFile);
	return 0;
}
