#include "UEFInfo.h"
#include <string>
#include <iostream>
#include "DataParser.h"
#include  <wx/filename.h>
#include  <wx/file.h>
#include  <wx/ffile.h>
#include  <wx/hashmap.h>

int DataParser::default_a = 0x2c; // ,
int DataParser::default_b = 0x20; //space
int DataParser::default_u = (int)(':'); //@
int DataParser::default_d = 0x2f; // /
int DataParser::default_l = 0x5a; // Z
int DataParser::default_r = 0x58; // X
int DataParser::default_1 = 0;
int DataParser::default_2 = 0;
int DataParser::default_t = 0;
int DataParser::default_s = 0xd; //return

/**Default codes*/
int DataParser::defaultKeys [10]= {
  default_a,
  default_b,
  default_s,
  default_t,
  default_r,
  default_l,
  default_u,
  default_d,
  default_1,
  default_2,
};
#if 1
WX_DECLARE_HASH_MAP(
    wxString,      // type of the keys
    UEFInfo*,    // type of the values
    wxStringHash,     // hasher
    wxStringEqual,   // key equality predicate
    UEFInfoHash); // name of the class

UEFInfoHash infoTable;
#endif 
DataParser::DataParser(const wxString& fileName) {
  // open filename, read in list of things
  // if file is non existant, then do nothing
  fileName_ = new wxString(fileName);
  // search classpath
#if 0
  File f;
  f = new File(fileName);
  if (!f.exists()) {
    // try in the executable dir
    wxString sep = System.getProperty("file.separator");
    wxString dir = System.getProperty("java.class.path");
    System.out.println("dir = "+dir);
    if (! (dir.equals("inject.jar") || dir.equals("."))) {
      f = new File(( new File(dir)).getParent());
      f = new File(f.getPath() + sep + fileName);
      System.out.println("Trying "+f.getPath());
      if (f.exists()) {
	fileName_ = f.getPath();
      }
    }
  }
#endif

  parseFile(*fileName_);
}

UEFInfo * DataParser::getInfo(const wxString& fname) {
  // create a UEFInfo entry for this filename, if exists
  // otherwise use defaults.
  // default name is fname - extension
  wxFileName fileName(fname);
  wxString fullName = fileName.GetFullName();
  wxString userName = fileName.GetName();
  UEFInfo * uei =  infoTable[fullName];
  if (uei == NULL) {
    // trim the .uef bit

//    if (userName.Length() > 11)
//      userName = userName.SubString(0,11);
    UEFInfo *newuef = new UEFInfo(fname,userName,defaultKeys);
    setInfo(newuef);
    //uei->init(fname,userName,defaultKeys);
    return newuef;
  }  
 // }
  return uei;
}

void DataParser::setInfo(UEFInfo * ui) {
  // set a UEFInfo entry
  infoTable[(ui->getFileName())] = ui;
}


void DataParser::parseFile(wxString& dataFileName) {
  // the file must be opened and stuff.
  // create HashTable entries - use name as key and UEFInfo as object
  // entries are key = value
#if 1
  
  if (wxFile::Exists(dataFileName)) {
    wxFFile p(dataFileName,"rb");    
    if (p.IsOpened()) {
      size_t len =  p.Length();
      while ((p.Tell() < len) && !p.Eof()) {
	UEFInfo *info = UEFInfo::read(p);
	if (info != NULL) {
	  setInfo(info);
	}
	else
	  break;

      }
    }
  }
#endif

}

void DataParser::closeAndSave() {
  // write the contents to a file...
#if 1
  bool canWrite = false;
  if (wxFile::Access(fileName_->c_str(), wxFile::write)) {
    canWrite = true;
  }
  wxFFile fo;
  if (canWrite)  
    fo.Open(fileName_->c_str(),"wb");

  UEFInfoHash::iterator ii = infoTable.begin();
  for (int i = 0;
      ii != infoTable.end() ;
      i++,ii++) 
  {
    UEFInfo * entry = ii->second;
    if (entry != NULL) {
      if (canWrite)  
	entry->write(fo);
      delete entry;
    }

  }
  infoTable.clear();
#endif
}

int DataParser::trimStateData(unsigned char * StateData, int Length) {
  int index = 10;
  int FLength = Length;
  wxString name(&StateData[0],&StateData[9]); // doesn't count 0 at end
  if (name.Cmp("UEF File!") != 0) {
    //cout << "Name not right = #"<<name<<"#"<<endl;
    return Length;
  }
  unsigned char *trimmed = (unsigned char *)malloc(Length);
  int hv=StateData[index++]&0xff;
  int lv=StateData[index++]&0xff; // version

  memcpy(trimmed, StateData, 12);
  int trimmedIndex = 12;
  int recognised = 0;

  while (index < FLength) {
    int blockPos = index;
    int Block = (StateData[index]&0xff) | ( (StateData[index+1]&0xff) <<8);
    index += 2;

    int Length = (0xff&StateData[index]) | ( (0xff&StateData[index+1]) <<8)
      | ( (0xff&StateData[index+2]) <<16) | ( (0xff&StateData[index+2]) <<24);

    index += 4;
    if (Length > FLength || index >= FLength)
      break;
    int CPos = index;
    if (Block==0x0460 ||
	Block==0x0462 ||
	Block==0x0467 ||
	Block==0x0468)
    {
      // recognised block, copy Length + header bytes to output
      memcpy(trimmed+trimmedIndex, StateData+blockPos, Length+6);
      trimmedIndex += (Length+6);
      recognised++;
    } 
      
    // Skip unrecognised blocks (and over any gaps)
    index = CPos+Length;
  }
  memcpy(StateData, trimmed, trimmedIndex);
  free(trimmed);
  if (!recognised) {
    return 0;
  }
  return trimmedIndex;
}


