//#include <iostream.h>
#include "wx/filename.h"
#include "UEFInfo.h"

UEFInfo::UEFInfo():
isnull(1), m_userName(0)
{

  //cout << "Construct null Info " << endl;
}

UEFInfo::UEFInfo(
    const wxString & fileName, 
    const wxString & userName, 
    int * keys) :
isnull(0), m_userName(0)
{
  // create info from these inputs

  m_fileName = new wxFileName(fileName);

  setUserName(userName);

  for (int i = 0; i < 10; i++) {
    m_keys[i] = keys[i];
  }
  xScale = 128;
  yScale = 256;
  xOffset = 0;
  yOffset = 0;

}

void UEFInfo::init(
    const wxString & fileName, 
    const wxString & userName, 
    int * keys) {
  if (isnull) {
    isnull = 0;
    m_fileName = new wxFileName(fileName);
    setUserName(m_fileName->GetName());

    for (int i = 0; i < 10; i++) {
      m_keys[i] = keys[i];
    }
  }
}
UEFInfo::~UEFInfo() {
//  cout << "Destruct Info " << *m_userName << endl;
  delete m_userName;
  delete m_fileName;
}

unsigned char * UEFInfo::getInfoAsBytes(int realUEFSize) {
  // the header for the GBA menu has this form:
  /*
   * Size: 4 byte size
userName : 12 bytes for the name, username[11] == 0
keys[]  : 12 bytes 10 for the key maps, 2 spare
x       : 2 bytes for x scale value
y       : 2 bytes for y scale value
dx      : 2 bytes for x offset
dy      : 2 bytes for y offset

total size 36 bytes  
   */

  unsigned char *  hdr = (unsigned char*)malloc(HEADER_SIZE);
  int index = 0;
  // the size
  hdr[index++] = (unsigned char)(realUEFSize&0xff);
  hdr[index++] = (unsigned char)((realUEFSize>>8)&0xff);
  hdr[index++] = (unsigned char)((realUEFSize>>16)&0xff);
  hdr[index++] = (unsigned char)((realUEFSize>>24)&0xff);

  // the name
  const wxChar * str = m_userName->c_str();
  for (unsigned int i = 0; i < 12; i++) {
    if (i < m_userName->Length())
      hdr[index++] = (unsigned char)str[i];
    else
      hdr[index++] = (unsigned char)0;
  }

  // the keys
  for (int i = 0; i < KEYS_LENGTH; i++) {
    hdr[index++] = (unsigned char)m_keys[i];
  }
  // key array padding
  hdr[index++] = 0;
  hdr[index++] = 0;

  // x       : 2 bytes for x scale value
  hdr[index++] = (unsigned char)(xScale&0xff);
  hdr[index++] = (unsigned char)((xScale>>8)&0xff);
  // y       : 2 bytes for y scale value
  hdr[index++] = (unsigned char)(yScale&0xff);
  hdr[index++] = (unsigned char)((yScale>>8)&0xff);
  // dx      : 2 bytes for x offset
  hdr[index++] = (unsigned char)(xOffset&0xff);
  hdr[index++] = (unsigned char)((xOffset>>8)&0xff);
  // dy      : 2 bytes for y offset
  hdr[index++] = (unsigned char)(yOffset&0xff);
  hdr[index++] = (unsigned char)((yOffset>>8)&0xff);

  return hdr;
}


int UEFInfo::getKey(int keyIndex) {

  int value = m_keys[keyIndex];
  if (value < 0xf && value >= 0) 
    return value;
  else if (value >= 0x20 && value < (0x6a) ) {
    return value-0x20+0xf;
  }
  return 0;
}

void UEFInfo::setKey(int keyIndex, int value) {
  if (value < 0xf && value >= 0) {
      m_keys[keyIndex] = value;
  } else if (value >= 0xf && value < (0x6a-0x20+0xf) ) {
    // ascii:
    m_keys[keyIndex] = value+0x20-0xf;
  }
}

#define WriteShort(sh)  \
  tmp[0] = sh&0xff; \
  tmp[1] = (sh>>8)&0xff;\
  out.Write(tmp,2);

#define ReadShort(sh)  \
  in.Read(tmp,2); \
  sh = tmp[0]&0xff; \
  sh |= (tmp[1]&0xff)<<8;

#define WriteUTF(str)  \
  tc = (char*)str.c_str(); \
  out.Write(tc,str.Length());
  
//#define ReadUTF(str)  \
  in.Read(tmp,12); \
  str = tmp;


void UEFInfo::write(wxFFile & out) 
{
  
  // write this UEFInfo to the given file
  unsigned char tmp[2]; 
  char *tc;
  for (int i = 0; i < KEYS_LENGTH; i++) {
    WriteShort(m_keys[i]);
  }
  WriteShort(xScale);
  WriteShort(yScale );
  WriteShort(xOffset);
  WriteShort(yOffset);
    
  WriteUTF((*m_userName));
  tmp[0] = 0;
  out.Write(tmp,1);
  wxString fileName = m_fileName->GetFullName();
  WriteUTF(fileName);
  tmp[0] = 0;
  out.Write(tmp,1);
}

UEFInfo * UEFInfo::read(wxFFile & in) {
  // read info from the file, return new UEFInfo or null if not valid data.
  unsigned char tmp[12]; 
  int keys[10];
  for (int i = 0; i < KEYS_LENGTH; i++) {
    ReadShort(keys[i]);
  }
  int xScale,yScale,xOffset,yOffset;

  ReadShort(xScale);
  ReadShort(yScale);
  ReadShort(xOffset);
  ReadShort(yOffset);
  wxString userName;
  wxString fileName;
  
  in.Read(tmp,1);
  while (tmp[0] != 0) {
    userName += tmp[0];
    in.Read(tmp,1);
  }
  in.Read(tmp,1);
  while (tmp[0] != 0) {
    fileName += tmp[0];
    in.Read(tmp,1);
  }
  return new UEFInfo(fileName, userName, keys);
}



