//standard headers and library calls
#include"gba.h"
#include"screenmode.h"
#include"dma.h"
#include "gamemenu.h"
#include "keypad.h"
#include "screenmode.h"
#include "gfx.h"

u8 *const ndata = (u8*)0x0200c000;


void load_pcx(u8 * gfx) {
  SetMode(MODE_3 |BG2_ENABLE); //set mode 3
  REG_BG2CNT = 0x80;
  int i, i8, j8;
  // 128 byte PCXFILEHEADER
  u8 pf[128];
  
  REG_DMA3SAD = (u32)gfx; 
  REG_DMA3DAD = (u32)pf;
  REG_DMA3CNT = 64 | DMA_16NOW; // 16 bit transfers from gfx array via DMA  
  int gfx_index = 128;

  
  // interpret the header
  // ZSoft .PCX FILE HEADER FORMAT
  // 0         Manufacturer 1      Constant Flag, 10 = ZSoft .pcx
  //int nmanf = pf[0]&0xFF;
  
  // 1         Version      1      Version information
  //0 = Version 2.5 of PC Paintbrush
  //2 = Version 2.8 w/palette information
  //3 = Version 2.8 w/o palette information
  //4 = PC Paintbrush for Windows(Plus for
  //   Windows uses Ver 5)
  //5 = Version 3.0 and > of PC Paintbrush
  //   and PC Paintbrush +, includes
  //   Publisher's Paintbrush . Includes
  //   24-bit .PCX files
  //int nvers = pf[1]&0xFF;
  
  // 2         Encoding      1     1 = .PCX run length encoding
  //int nenc = pf[2]&0xFF;
  
  // 3         BitsPerPixel  1     Number of bits to represent a pixel
  //    (per Plane) - 1, 2, 4, or 8
  //int nbpp = pf[3]&0xFF;               
    
  // 4         Window        8     Image Dimensions: Xmin,Ymin,Xmax,Ymax
  int ndim[4];
  ndim[0] = ((pf[5]&0xFF)<<8)|(pf[4]&0xFF);
  ndim[1] = ((pf[7]&0xFF)<<8)|(pf[6]&0xFF);
  ndim[2] = ((pf[9]&0xFF)<<8)|(pf[8]&0xFF);
  ndim[3] = ((pf[11]&0xFF)<<8)|(pf[10]&0xFF);
  
  // 12         HDpi          2     Horizontal Resolution of image in DPI*
  //int nhdpi =  ((pf[13]<<8)|(pf[12]&0xFF))&0xFFFF;
  // 14         VDpi          2     Vertical Resolution of image in DPI*
  //int nvdpi =  ((pf[15]<<8)|(pf[14]&0xFF))&0xFFFF;
  // 16         Colormap     48     Color palette setting, see text
  //int ColorMap[48];
  // 64         Reserved      1     Should be set to 0.
  // 65         NPlanes       1     Number of color planes
  int nplanes =  pf[65]&0xFF;
  
  // 66         BytesPerLine  2     Number of bytes to allocate for a scanline
  //                                plane.  MUST be an EVEN number.  Do NOT
  //                                calculate from Xmax-Xmin.
  int nbpl =  ((pf[67]<<8)|(pf[66]&0xFF))&0xFFFF;
  
  // 68         PaletteInfo   2     How to interpret palette- 1 = Color/BW,
  //                                 2 = Grayscale (ignored in PB IV/ IV +)
  //int npalinfo =  ((pf[69]<<8)|(pf[68]&0xFF))&0xFFFF;
  // 70         HscreenSize   2     Horizontal screen size in pixels. New field
  //                                 found only in PB IV/IV Plus
  //int nhss =  ((pf[71]<<8)|(pf[70]&0xFF))&0xFFFF;
  // 72         VscreenSize   2     Vertical screen size in pixels. New field
  //                                 found only in PB IV/IV Plus
  //int nvss =  ((pf[73]<<8)|(pf[72]&0xFF))&0xFFFF;
  
  //74         Filler       54     Blank to fill out 128 byte header.  Set all
  //                                 bytes to 0
  
  
  // Decoding .PCX Files
  // First, find the pixel dimensions of the image by calculating
  // [XSIZE = Xmax - Xmin + 1] and [YSIZE = Ymax - Ymin + 1].  
  // XMin = ndim[0]
  // YMin = ndim[1]
  // XMax = ndim[2]
  // YMax = ndim[3]
  int nxsize = ndim[2] - ndim[0] + 1;
  int nysize = ndim[3] - ndim[1] + 1;
  
  
  // Then calculate how many bytes are required to hold 
  // one complete uncompressed scan line:
  // TotalBytes = NPlanes * BytesPerLine
  int ntbpl = nplanes * nbpl;
  
  if (nysize > 160 || nxsize > 240) {
    BGPaletteMem[0] = RGB16(22,11,11);
    SetMode(MODE_3); //set mode 0
    return;
  }
  
  //  Note that since there are always an even number of bytes per scan line,
  //  there will probably be unused data at the end of each scan line.  TotalBytes
  //  shows how much storage must be available to decode each scan line, including
  //  any blank area on the right side of the image.  You can now begin decoding
  //  the first scan line - read the first byte of data from the file.
  int line = 0;
  while (line < nysize) {
    int pilen=1;
    int nslindex = 0;
    
    u8 pi;
    while (nslindex < ntbpl) {
      pi = gfx[gfx_index++];
      
      //  If the top two bits are set, the remaining six bits in the byte 
      //  show how many times to duplicate the next byte in the file.    
      
      if ((pi & 0xC0)==0xC0) {
        // top 2 set
        pilen= pi & 0x3F; 
        pi = gfx[gfx_index++];
        
        for (i = 0; i < pilen; i++) {
          ndata[(line*ntbpl)+nslindex++] = pi;
        }
      } else {
        // If the top two bits are not set,
        // the first byte is the data itself, with a count of one.
        ndata[(line*ntbpl)+nslindex++] = pi;
      }
    }
    line++;
  }
  
  // here ndata contains the references into the colour palette
  if (nplanes == 1) {
    // can only do 1 plane pcxs :( 3 planes are true 24 bit coloured
    u8 pi;
    pi = gfx[gfx_index++];
    
    if ((pi & 0xFF) == 0x0C) {
      
      // pcx palette len: 256 colour, 3 bytes represent each colour
      u8 r, g, b;
      int palette[256]; // the palette
      
      for (i = 0; i < 256; i++) {
        r = gfx[gfx_index++]>>3;
        g = gfx[gfx_index++]>>3;
        b = gfx[gfx_index++]>>3;
        
        palette[i] = RGB16(r, g, b);
        BGPaletteMem[i] = palette[i];
      }
      
      BGPaletteMem[0] = 0;
      // convert the scanline data from earlier to the
      // 15 bit image data here...
      int nindex8 = 0;
      
      for (j8 = 0; j8 < nysize; j8++) {
        for (i8 = 0; i8 < nxsize; i8++) {
          VideoBuffer[240*j8+i8] = palette[ndata[nindex8++]];
        }
      }
    }
  }
//  free(ndata);
}

void showSplash(void) {
  unsigned char * splashData = (unsigned char *) endOfGBA;
  unsigned long splash = * (unsigned long*) (endOfGBA-8);
  if (splash) {
    fade_down();
    load_pcx(splashData);
    fade_up();
    int i;
    for (i = 0; i < 180; i++) {
      UpdateKeys();
      wait_vsync_int();
      if (KeyPressedNoBounce(KEY_A,5))
	break;
    }
    fade_down();
    clear_vram();
  }
  else {
    fade_down();
  }
}

