#include "savestate.h"
#include "saving.h"
#include "6502core.h"
#include "bbcmemory.h"
#include "bbcsound.h"
#include "sysvia.h"
#include "uservia.h"
#include "fastvideo.h"
#include "gba.h"
#include "visoly.h"
#include <string.h>
#include "gfx.h"

void Load6502UEF(const unsigned char * const SUEF) {
    T6502State *state;
    state = (T6502State*)&Asm6502State;
    int i = 0;
    state->programCounter = (SUEF[i]|(SUEF[i+1]<<8))&0xffff;
    i+=2;
    state->accumulator=(SUEF[i++]&0xff);
    state->XReg=(SUEF[i++]&0xff);
    state->YReg=(SUEF[i++]&0xff);
    stack=(SUEF[i++]&0xff)|0x100;
    state->PSR=(SUEF[i++]&0xff);
    
    i+=4;
    intStatus=(SUEF[i++]&0xff);//fgetc(SUEF);
    
  
}
void LoadUEFState(const unsigned char * const StateData, const unsigned char * const end) {
    
    unsigned long FLength,CPos;
    unsigned long Block,Length;
    int hv,lv;

    FLength= ((int)end) - ((int)StateData); //ftell(UEFState);
    unsigned long index = 10;
    if (strcmp(StateData,"UEF File!")!=0) {
      return;
    }
    hv=StateData[index++];
    lv=StateData[index++]; // version

    while (index < FLength) {
      Block = StateData[index] | ( StateData[index+1] <<8);
      index += 2;
      
      Length = StateData[index] | ( StateData[index+1] <<8) 
               | ( StateData[index+2] <<16) | ( StateData[index+2] <<24); 

      index += 4;
      if (Length > FLength || index >= FLength)
        break;
      
      CPos = index;
      
      if (Block==0x0460) { 
        Load6502UEF(&StateData[index]); 
      }

      if (Block==0x0462) { 
        LoadMainMemUEF(&StateData[index]);
      }
      if (Block==0x0467) { 
        LoadViaUEF(&StateData[index]);
      }
      if (Block==0x0468) { 
        LoadVideoUEF(&StateData[index]);
      }
      
      // Skip unrecognised blocks (and over any gaps)
      index = CPos+Length;
      
    }
}

const char * save_key = "BBC";

#define SAVE_OFFSET 3
#define Save8(b)  SaveByte(index++, b&0xff);
#define Load8(b)  b = LoadByte(index++);

#define Save16(hw) \
  SaveByte(index++,hw&0xff); \
  SaveByte(index++,(hw>>8)&0xff);

#define Load16(hw) \
  hw = LoadByte(index++); \
  hw |= LoadByte(index++)<<8;

#define Save32(w) \
  SaveByte(index++, w&0xff); \
  SaveByte(index++, (w>>8)&0xff); \
  SaveByte(index++, (w>>16)&0xff); \
  SaveByte(index++, (w>>24)&0xff);

#define Load32(w) \
  w = LoadByte(index++); \
  w |= LoadByte(index++)<<8; \
  w |= LoadByte(index++)<<16; \
  w |= LoadByte(index++)<<24;

void SaveString(char* str, int * index) {
  int i = 0;
  while (str[i] != 0) {
    SaveByte(*index,str[i++]);
    *index = (*index)+1;
  }

}

typedef void (*LineRoutinePtr)(void);
extern LineRoutinePtr scanlineFunction;
extern unsigned long cyclesPerScanline;
int have_save_data(void) {
  // check we have valid data...
  char key[4];
  int i;
  for (i = 0; i < 3; i++)
    key[i] =  LoadByte(i);
 
  key[3] = 0;
  return (strcmp(key,save_key) == 0);
}

int load(void) {
  // load from sram - call from asm?
  if (have_save_data()) {
    resetBeeb(0);
    LoadUEFState((char*)(SRAM+SAVE_OFFSET),(char*)(SRAM+SAVE_OFFSET+0x8000+140));
    return 1;
  } else {
    return 0;
  }
}

void save(void) {
  // save to sram - called from asm
  // Asm6502State
  T6502State *state;
  state = (T6502State*)&Asm6502State;
  u32 i;
  int asm_pc = state->programCounter;
  int asm_a = state->accumulator;
  int asm_x = state->XReg;

  int asm_y = state->YReg;
  int asm_stack = stack;
  int asm_psr = state->PSR;
  
  // save the key so we know it's our data
  for (i = 0; i < 3; i++)
    SaveByte(i,save_key[i]);
  
  // save 6502 core:
  int index = SAVE_OFFSET;

  SaveString("UEF File!",&index);
  Save8(0);
  Save8(8); Save8(0);
  //12 bytes

  Save16(0x0460);
  Save32(16);
  Save16(asm_pc);
  Save8(asm_a);
  Save8(asm_x);
  Save8(asm_y);
  Save8(asm_stack);
  Save8(asm_psr);
  Save32(0);
  Save8(intStatus);
  Save32(0);
  // 22 bytes  
  // total = 13+22 = 35
  
  VIAState *VIAPtr=&SysVIAState;
  // save sys via
  Save16(0x0467);
  Save32(22);
  Save8(0); // 0 for sys via, 1 for user
  Save8(VIAPtr->orb);
  Save8(VIAPtr->irb);
  Save8(VIAPtr->ora);
  Save8(VIAPtr->ira);
  Save8(VIAPtr->ddrb);
  Save8(VIAPtr->ddra);
  Save16(VIAPtr->timer1c);
  Save16(VIAPtr->timer1l);
  Save16(VIAPtr->timer2c);
  Save16(VIAPtr->timer2l);
  Save8(VIAPtr->acr);
  Save8(VIAPtr->pcr);
  Save8(VIAPtr->ifr);
  Save8(VIAPtr->ier);
  Save8(VIAPtr->timer1hasshot);
  Save8(VIAPtr->timer2hasshot);
  Save8(IC32State);
  // 28
  // total = 35 + 28 = 63
  // save video uef
  
  Save16(0x0468);
  Save32(47);
  Save8(CRTCControl.CRTC_HorizontalTotal);
  Save8(CRTCControl.CRTC_HorizontalDisplayed);
  Save8(CRTCControl.CRTC_HorizontalSyncPos);
  Save8(CRTCControl.CRTC_SyncWidth);
  Save8(CRTCControl.CRTC_VerticalTotal);
  Save8(CRTCControl.CRTC_VerticalTotalAdjust);
  Save8(CRTCControl.CRTC_VerticalDisplayed);
  Save8(CRTCControl.CRTC_VerticalSyncPos);
  Save8(CRTCControl.CRTC_InterlaceAndDelay);
  Save8(CRTCControl.CRTC_ScanLinesPerChar);
  Save8(CRTCControl.CRTC_CursorStart);
  Save8(CRTCControl.CRTC_CursorEnd);
  Save8(CRTCControl.CRTC_ScreenStartHigh);
  Save8(CRTCControl.CRTC_ScreenStartLow);
  Save8(CRTCControl.CRTC_CursorPosHigh);
  Save8(CRTCControl.CRTC_CursorPosLow);
  Save8(CRTCControl.CRTC_LightPenHigh);
  Save8(CRTCControl.CRTC_LightPenLow);
  Save8(VideoULA_ControlReg);
  // 25
  // total = 63+25 = 88
  
  int col;
  unsigned char p[16];
  get_pal_indices(p);

  for (col = 0; col < 16; ++col)
  	Save8(p[col]); /* Use real ULA values */
  	//Save8((VideoULA_Palette[col] ^ 7)); /* Use real ULA values */
  // 16 bytes
  // total = 88+16 = 104
  // should really be based on OBJPaletteMem?
  
  Save16(/*ActualScreenWidth*/0);
  Save32(/*ScreenAdjust*/0);
  Save8(CRTCControl.CRTCControlReg);
  Save8(/*TeletextStyle*/0);
  Save32(0); // Pad out
  // 12 bytes
  // total = 104+12 = 116

  // save main mem
  Save16(0x0462); // Main Memory
  Save32(32768);
  for (i = 0; i < 0x8000; i++)
    SaveByte(index++,WholeRam[i]);


}


