
#include "regs.h"

.section .iwram,"ax",%progbits
.align 4
.code 32


.macro  getSoundReg  regOut, channel
  mov     r2,#REG_BASE          
  mov     \regOut,#(REG_SG11 & 0xFFFF)
  orr     \regOut,\regOut,r2,lsl#20     @0x40 -> 0x4000000
  add     \regOut,\regOut,\channel,lsl#3
.endm

.macro  createStereoMask  channel
  adr   r4,stereoMaskTbl
  ldr   r4,[r4,\channel,lsl#2]     @ word table
.endm
  
  
sound3voltbl:
@ 000 - Mute
@ 001 - 100%
@ 100 - 75%
@ 010 - 50%
@ 011 - 25%
.byte  0    @ 0
.byte  3*2  @ 1  25
.byte  2*2  @ 2  50
.byte  2*2  @ 3  50
.byte  4*2  @ 4  75
.byte  4*2  @ 5  75
.byte  1*2  @ 6  100
.byte  1*2  @ 7  100

/**
* Called when a write is made to the 76489
*
* 3 channels and a noise generator
*
* Can only write a byte at a time, natch, but 2 bytes can be written 
* in sequence to effect upper and lower parts of a channel's freqeuncy.
*
* If the byte bit 7 is set, then it is a write to a register
* Otherwise a top half frequency change is made to the last
* channel adjusted.
*/

.align 4
.global writeSound
.type writeSound, function
writeSound:
  stmfd	  sp!, {r3,r4,r5,r6}
  tst     r0,#0x80
  bne     soundRegWrite
  
  @ top bit not set - frequency write for byte 2
  @ offset_lastTone is the last channel written to
  
  adr     r1,AsmSoundState
  ldrb    r2,[r1,#offset_lastTone]
  @ r2 = last channel changed
  @ adjust top half of this channel's frequency 
  @ new frequency = (current frequency & 0xf) | ((value&0x3f)<<4)
  
  ldr     r3,[r1,r2,lsl #2]      @ r3 = current freq
  and     r3,r3,#0xf             @ keep low nibble
  and     r0,r0,#0x3f
  orr     r3,r3,r0,lsl#4        @ r3 = (top&#3f)<<4 | (bottom&#f)
  str     r3,[r1,r2,lsl #2]    @ store new frequency
  b       0f

soundRegWrite:
  @ top nibble of r0 (&7) holds the sound register to write to:
  @ 0 = ToneFreq[0]
  @ 1 = ToneVolume[0]
  @ 2 = ToneFreq[1]
  @ 3 = ToneVolume[1]
  @ 4 = ToneFreq[2]
  @ 5 = ToneVolume[2]
  @ 6 = ToneFreq[3]     @noise
  @ 7 = ToneVolume[3]     @noise
  
  adr   r1,AsmSoundState
  and   r3,r0,#0x70
  mov   r2,r3,lsr#5     @ shift 0x70 down 5 = 7/2 = 0 to 3
  strb   r2,[r1,#offset_lastTone]    @ last tone changed 
  cmp   r3,#0x60
  beq   doNoise     @ noise frequency is slightly different
  @ set volume/freq of standard chnls
  mov   r3,r3,lsr#3       @ r3 = 0x70>>3 = 0-E
  tst   r3,#0x2           @ the bottom bit is in bit 1
  and   r2,r0,#15         @ r2 = new frequency or volume
  @ if bit 0 set, then it is a volume set, else freq set
  @ the channel to write to is r3 (well, r3 div 4)
  @ it is possible to switch off the channel:
  @ if value to write == 15 then the channel is off
  
  @ if Z is not set (ne) then a volume change (i.e. bit 0 was set)
  eorne     r2,r2,#0xf    @ volume = 15 - (value&15)...
  bic       r3,r3,#2      @ bit clear, so r3 = channel*4
  @mov       r3,r3,lsl#1   @ 4 bytes per entry channel * 4
  addne     r3,r3,#offset_tonevolume
  @ if Z is set (eq) then a frequency change
  ldreqh    r0,[r1,r3]      @get current freq as a half word
  andeq     r0,r0,#0x3f0    @ mask off low nibble
  orreq     r2,r0,r2        @ the new low nibble
  @ store the new value to frequency/volume
  strh       r2,[r1,r3]
  b     0f
@.type doNoise,function
doNoise:
  @ 6 = ToneFreq[3]     @noise
  @ noise.freq = value & 3
  @ noise.fb = (value & 4) >> 2 @ freq bit check can do this better
  and        r2,r0,#7
  strh       r2,[r1,#12] @ ToneFrequency[noise] (3*4 = 12 bytes into the array)
  
0:
  @ set sound REGs based on the value offset_lastTone
  adr   r1,AsmSoundState
  ldrb   r0,[r1,#offset_lastTone]
  @ switch on r0 - the last tone set. 
  @ sound keeps playing on a channel until switched off again by this routine

@ update the sound channel...
.type soundChannel,function
@ input r1 = AsmSoundState
@ input r0 = last tone played
soundChannel:
  
  @ write the frequency only if vol > 0
  @ otherwise switch off sound
  @ get volume
  mov     r3,#offset_tonevolume
  add     r3,r3,r0,lsl#2
  ldrb     r6,[r1,r3]
  @ r3 contains the volume of this channel

  @ if r3 == 0, need to clear bits 8 and 12 of REG_SGCNT0_L 
  @ (set channel 1 neither left nor right - else set these bits to enable stereo sound)
  @ r2 = 0x4000060
  mov   r4, #4
  cmp     r6,#0  @ is vol == 0?
  getSoundReg r3, r4      @ get channel r4 into r3 (channel 4 ==> REG_SGCNT0_L)
                          @REG_SGCNT0_L   control 0
  sub     r2,r3,#4      @ cos control 0 is 4 less than we calcd
  ldrh    r5,[r2,#0]
  @ create a bit mask to enable/disable this channel
  
  createStereoMask  r0
  
  biceq   r5,r5,r4   @ disable if vol == 0
  orrne   r5,r5,r4   @ enable  if vol != 0
  strh    r5,[r2,#0]
  
  @beq     endWriteSound  @ don't do this, as the volume doesn't take effect 
                          @ until we write the reset bit to the freq reg
  
  @ has volume, so set frequency...
  @ need to set the frequency here - 11 bits of frequency, 0-0x7FF
  @ but beeb only has 10 bits of frequency in its channels, what to do?
  @ just put the emu freq as the real freq and cross fingers!
  @ lowest 11 bits of SGn1 are the frequency (unless noise)
.type setFreqReal, function
setFreqReal:
  getSoundReg r3, r0             @ r3 = SGn1 for channel r0
  
  ldr    r1,[r1,r0,lsl#2]       @ now r1 = frequency of this channel 
                                @ (half word wanted, but store only half below)
  
  cmp     r0,#3         @ is it the noise channel?

  andeq   r1,r1,#3      @ simplify for noise: channel 3 = freq & 3
  beq     storeFreq
  
  @ not the noise channel
  @ Real F(hz)=4194304/(32*(2048- 11 bit reg value))
  @ emu frequency  F(hz) = 2000000 / 32 * (10 bit value)
  @ SG_Freq = 2048 - 2 * BBC_Freq
  @cmp     r0,#2
  ldr     r4,eorMask        @ real frequency is 0x800 - val*2
  @subne   r1,r4,r1,lsl#1    @(0x7ff - val*2)
  sub    r1,r4,r1,lsl#1    @(0x7ff - val*2)
  @ this next line shouldn't be needed, but without it everything sounds wrong
  @ perhaps the square wave in sound bank 3 has an incorrect frequency ?
  @subeq   r1,r4,r1,lsr#1    @(0x7ff - val/2) for GBA channel 3, need to halve the freq (?)
  
.global storeFreq
.type storeFreq,function
storeFreq:
	@ mix in volume from r6
  tsts       r0,#1
  ldreqh     r5, [r3,#-2]     @ load the current envelope value for 1,3 (0,2) this is 2 less than SGn1
  ldrneh     r5, [r3,#-4]     @ load the current envelope value for 2,4 (1,3) this is 4 less than SGn1

  bic        r5,r5,#0xf000    @ clear current top nibble (current volume)

  @ sound 3 volume is odd. Deal with that now...
  cmp        r0,#2            @ is it sound channel 3?
  ldreq      r2, =sound3voltbl
  ldreqb     r6,[r2,r6,lsr#1] @ map bbc volume (r6) to nearly-equivalent sound 3 volume

  orr        r5,r5,r6,lsl#12  @ start envelope is the top nibble, orr in bbc volume

  tsts       r0,#1            @ retest which channel we are changing
  streqh     r5, [r3,#-2]     @ update envelope 1 + 3
  strneh     r5, [r3,#-4]     @ update envelope 2 + 4

  @ finally, store the frequency in SGn1, which also activates volume changes
  orr     r1,r1,#0x8000          @ initialize and continue until we tell you to stop
  strh    r1,[r3]                @ store frequency in SGn1
  @ (note to self: sound length?)

endWriteSound:

@ debug start
#if 0
  @ dump sound reg and frequency value
  stmfd   sp!,{r0-r12,lr}
  adr     lr,endDumpSound
  mov     r2,r6
  ldr     r7,LsoundDump
  bx  r7
endDumpSound:
  ldmfd   sp!,{r0-r12,lr}
#endif
@ debug end

  ldmfd   sp!, {r3,r4,r5,r6}
  mov     pc,lr
  
LsoundDump:
.word dumpSoundReg

eorMask: .word 0x800

.global AsmSoundState
AsmSoundState:
.space (4*4+4*4+4)

@ unsigned int ToneFreq[4];
@ unsigned int ToneVolume[4]; 
@ int LastToneFreqSet; /* the tone generator last set - for writing the 2nd byte */

.equ offset_tonefreq, 0
.equ offset_tonevolume, 4*4
.equ offset_lastTone, 4*4+4*4


/*
void dumpSoundReg(int ch, int freq, int dunno, int SGn1) {
  gbaprintf("Sound ch %02x F %04x REG %08x\n", ch, freq, SGn1);
} */

stereoMaskTbl:
.word 0x1100
.word 0x2200
.word 0x4400
.word 0x8800
@.byte 0x00,0x11
@.byte 0x00,0x22
@.byte 0x00,0x44
@.byte 0x00,0x88

.pool
.end
