#include "6502macros.h"

.extern FastTable
.extern FastTableDWidth
.extern nextEvent
.extern endScanline
.extern doInterrupt
.extern SysVIAIR
.extern VideoState
.section .iwram,"ax",%progbits
.align 4
.code 32


@ checks at the start of each frame and adjust vramOffset accordingly
.macro checkDataPtr
  @ check that r1 is in bounds, wrap around if need be
  sub       r3,r1,bbc_mem
  cmp       r3,#0x8000
  ldrcs     r3,vramOffset
  subcs     r1,r1,r3
.endm


.global lowLevelDoScanLineNarrow
.type lowLevelDoScanLineNarrow, function
lowLevelDoScanLineNarrow:

  
  @ need PixmapLine&~7 * 64 + PixmapLine&7
  ldr   r2,[r5,#OFFSET_CharLine]
  ldr   r1,[r5,#OFFSET_InCharLineUp]
  orr   r1,r1,r2,lsl#8        @ r1 = PixmapLine&~7 * 32 + PixmapLine&7

  @ find out bytes needed to copy:
  ldr   r0,CRTCVals
  ldr   r4,[r0,#OFFSET_HorizontalDisplayed]
  movs   r4,r4,lsr#2
  moveq  pc,lr                 @ if HD is 0, then return (nothing to display)
  
  @ this is for sprites
  mov   r2,#0x60
  mov   r0,#0x10000
  orr   r0,r0,r2,lsl#20
  add   r0,r0,r1,lsl#2      @ r0 = r0 + (r1*2)    (in bytes remember, u16 is tiledata)
  
  ldr   r1,[r5,#OFFSET_DataPtr]       @ r1 = VideoState.DataPtr
  ldr   r2,[r5,#OFFSET_InCharLineUp]       @ r2 = VideoState.InCharLineUp
  add   r1,r1,r2            @ r1 = gba ewram address of BBC video ram.
  @ InCharLineUp is the line in this character (i.e. from 0-7)
  
  ldr   memptr,=FastTable     @Asmbg_level      @ memptr is now fast table of exploded 8 bits -> pixels
  @ r0 = gba vram address to write to vidPtr
  @ r1 = address of bbc memory of gfx table CurrentPtr
  @ memptr = FastTable
  
commonFast:  
  @ check that the whole row will be in range
  @ if it is, no need to check the individual values
  @ i.e. if current pointer + 32*r4 > 0x8000 then the line will overflow at some point
  add   r3, r1,r4,lsl#5
  sub   r4,r4,#1          @ loop--
  sub   r3,r3,bbc_mem     @r3 = index + 32*r4
  subs  r3,r3,#0x8000
  bcc   fastLoopRow       @ the start doesn't overflow
  checkDataPtr
  bcs   fastLoopRow       @ adjust the start, if there is no overflow then fastline.
.type loopRow,function
loopRow:
  checkDataPtr
  ldrb   r3,[r1],#8                     @r3 = *CurrentPtr
  ldrb    r2,[memptr,r3]@,lsl#2]              @ r2 = FastTable[r3], u16 table (2 bytes per entry)

  checkDataPtr
  ldrb   r3,[r1],#8                  @r3 = *(CurrentPtr+8)
  ldrb    r3,[memptr,r3]@,lsl#2]        @ r2 = FastTable[r3], u16 table (2 bytes per entry)
  orr    r2,r2,r3,lsl#8
  strh   r2,[r0],#2
  
  checkDataPtr
  ldrb   r3,[r1],#8                 @r3 = *(CurrentPtr+16)
  ldrb    r2,[memptr,r3]@,lsl#2]        @ r2 = FastTable[r3], u16 table (2 bytes per entry)    
  
  checkDataPtr
  ldrb   r3,[r1],#8                  @r3 = *(CurrentPtr+24)
  ldrb    r3,[memptr,r3]@,lsl#2]        @ r2 = FastTable[r3], u16 table (2 bytes per entry)
  orr    r2,r2,r3,lsl#8
  strh   r2,[r0],#30
  
  subs   r4,r4,#1          @ loop--
  bpl   loopRow         @ fixes bug where HDisplay is 0
  mov   r15,r14

fastLoopRow:
  ldrb   r3,[r1],#8                     @r3 = *CurrentPtr
  ldrb   r2,[memptr,r3]@,lsl#2]              @ r2 = FastTable[r3], u16 table (2 bytes per entry)
  
  ldrb   r3,[r1],#8                  @r3 = *(CurrentPtr+8)
  ldrb    r3,[memptr,r3]@,lsl#2]        @ r2 = FastTable[r3], u16 table (2 bytes per entry)
  orr    r2,r2,r3,lsl#8
  strh   r2,[r0],#2
  
  ldrb   r3,[r1],#8                 @r3 = *(CurrentPtr+16)
  ldrb    r2,[memptr,r3]@,lsl#2]        @ r2 = FastTable[r3], u16 table (2 bytes per entry)    
  
  ldrb   r3,[r1],#8                 @r3 = *(CurrentPtr+24)
  ldrb    r3,[memptr,r3]@,lsl#2]        @ r2 = FastTable[r3], u16 table (2 bytes per entry)
  orr    r2,r2,r3,lsl#8
  strh   r2,[r0],#30
  
  subs  r4,r4,#1          @ loop--
  bpl   fastLoopRow
  mov   r15,r14

.global lowLevelDoScanLineWide
.type lowLevelDoScanLineWide, function
lowLevelDoScanLineWide:

  @ now find out ptr for where to copy to
  
  
  @ need PixmapLine&~7 * 64 + PixmapLine&7
  ldr   r2,[r5,#OFFSET_CharLine]
  ldr   r1,[r5,#OFFSET_InCharLineUp]
  orr   r1,r1,r2,lsl#8        @ r1 = PixmapLine&~7 * 32 + PixmapLine&7
  
  @ find out bytes needed to copy:
  ldr   r0,CRTCVals
  ldr   r4,[r0,#OFFSET_HorizontalDisplayed]
  movs   r4,r4,lsr#2
  moveq  pc,lr

  @ this is for sprites
  mov   r2,#0x60
  mov   r0,#0x10000
  orr   r0,r0,r2,lsl#20
  add   r0,r0,r1,lsl#2      @ r0 = r0 + (r1*2)    (in bytes remember, u16 is tiledata)
  
  ldr   r1,[r5,#OFFSET_DataPtr]       @ r1 = VideoState.DataPtr
  ldr   r2,[r5,#OFFSET_InCharLineUp]       @ r2 = VideoState.InCharLineUp
  add   r1,r1,r2            @ r1 = gba ewram address of BBC video ram.
  
  ldr   memptr,=FastTableDWidth    @Asmbg_level+4      @ memptr is now fast table of exploded 8 bits -> pixels
  @ r0 = gba vram address to write to vidPtr
  @ r1 = address of bbc memory of gfx table CurrentPtr
  @ memptr = FastTable
  
  @ check that the whole row will be in range
  @ if it is, no need to check the individual values
  @ i.e. if current pointer + 32*r4 > 0x8000 then the line will overflow at some point
  add   r3, r1,r4,lsl#5
  sub   r4,r4,#1          @ loop--
  sub   r3,r3,bbc_mem     @r3 = index + 32*r4
  subs  r3,r3,#0x8000
  bcc   fastLoopRowWide
  checkDataPtr
  bcs   fastLoopRowWide
.type loopRowWide,function
loopRowWide:
  checkDataPtr
  ldrb   r3,[r1],#8                     @r3 = *CurrentPtr
  ldr    r2,[memptr,r3,lsl#2]              @ r2 = FastTable[r3], u16 table (2 bytes per entry)

  checkDataPtr
  ldrb   r3,[r1],#8                  @r3 = *(CurrentPtr+8)
  ldr    r3,[memptr,r3,lsl#2]        @ r2 = FastTable[r3], u16 table (2 bytes per entry)
  orr    r2,r2,r3,lsl#16
  str    r2,[r0],#32
  
  checkDataPtr
  ldrb   r3,[r1],#8                 @r3 = *(CurrentPtr+16)
  ldr    r2,[memptr,r3,lsl#2]        @ r2 = FastTable[r3], u16 table (2 bytes per entry)    
  
  checkDataPtr
  ldrb   r3,[r1],#8                  @r3 = *(CurrentPtr+24)
  ldr    r3,[memptr,r3,lsl#2]        @ r2 = FastTable[r3], u16 table (2 bytes per entry)
  orr    r2,r2,r3,lsl#16
  str    r2,[r0],#32
  
  subs   r4,r4,#1          @ loop--
  bpl   loopRowWide
  mov   r15,r14

fastLoopRowWide:
  ldrb   r3,[r1],#8                     @r3 = *CurrentPtr
  ldr   r2,[memptr,r3,lsl#2]              @ r2 = FastTable[r3], u16 table (2 bytes per entry)
  
  ldrb   r3,[r1],#8                  @r3 = *(CurrentPtr+8)
  ldr    r3,[memptr,r3,lsl#2]        @ r2 = FastTable[r3], u16 table (2 bytes per entry)
  orr    r2,r2,r3,lsl#16
  str   r2,[r0],#32
  
  ldrb   r3,[r1],#8                 @r3 = *(CurrentPtr+16)
  ldr    r2,[memptr,r3,lsl#2]        @ r2 = FastTable[r3], u16 table (2 bytes per entry)    
  
  ldrb   r3,[r1],#8                 @r3 = *(CurrentPtr+24)
  ldr    r3,[memptr,r3,lsl#2]        @ r2 = FastTable[r3], u16 table (2 bytes per entry)
  orr    r2,r2,r3,lsl#16
  str   r2,[r0],#32
  
  subs  r4,r4,#1          @ loop--
  bpl   fastLoopRowWide
  mov   r15,r14


.global lowLevelDoScanLine1bpp
.type lowLevelDoScanLine1bpp, function
lowLevelDoScanLine1bpp:

  ldr   r2,[r5,#OFFSET_CharLine]
  ldr   r1,[r5,#OFFSET_InCharLineUp]
  orr   r1,r1,r2,lsl#8        @ r1 = PixmapLine&~7 * 32 + PixmapLine&7

  ldr   r4,[r0,#OFFSET_HorizontalDisplayed]          @CRTC_HorizontalDisplayed (64?)
  movs   r4,r4,lsr#2
  moveq  pc,lr                 @ if HD is 0, then return (nothing to display)
  
  mov   r2,#0x60
  mov   r0,#0x10000
  orr   r0,r0,r2,lsl#20
  add   r0,r0,r1,lsl#2      @ r0 = r0 + (r1*2)    (in bytes remember, u16 is tiledata)
  
  ldr   r1,[r5,#OFFSET_DataPtr]       @ r1 = VideoState.DataPtr
  ldr   r2,[r5,#OFFSET_InCharLineUp]       @ r2 = VideoState.InCharLineUp
  add   r1,r1,r2            @ r1 = gba ewram address of BBC video ram.
  
  ldr   memptr,=FastTable1bpp
  @ check that the whole row will be in range
  @ if it is, no need to check the individual values
  @ i.e. if current pointer + 32*r4 > 0x8000 then the line will overflow at some point
  add   r3, r1,r4,lsl#5
  sub   r4,r4,#1          @ loop--
  sub   r3,r3,bbc_mem     @r3 = index + 32*r4
  subs  r3,r3,#0x8000
  bcc   fastLoopRow1bpp       @ the start doesn't overflow
  checkDataPtr
  bcs   fastLoopRow1bpp       @ adjust the start, if there is no overflow then fastline.
.type loopRow1bpp,function
loopRow1bpp:
  checkDataPtr
  ldrb    r3,[r1],#8                     @r3 = *CurrentPtr
  ldr     r2,[memptr,r3,lsl#2]           @ r2 = FastTable[r3], u32 table
  str     r2,[r0],#32

  checkDataPtr
  ldrb    r3,[r1],#8                     @r3 = *CurrentPtr
  ldr     r2,[memptr,r3,lsl#2]           @ r2 = FastTable[r3], u32 table
  str     r2,[r0],#32

  checkDataPtr
  ldrb    r3,[r1],#8                     @r3 = *CurrentPtr
  ldr     r2,[memptr,r3,lsl#2]           @ r2 = FastTable[r3], u32 table
  str     r2,[r0],#32

  checkDataPtr
  ldrb    r3,[r1],#8                     @r3 = *CurrentPtr
  ldr     r2,[memptr,r3,lsl#2]           @ r2 = FastTable[r3], u32 table
  str     r2,[r0],#32
  
  subs   r4,r4,#1          @ loop--
  bpl   loopRow1bpp         @ fixes bug where HDisplay is 0
  mov   r15,r14

fastLoopRow1bpp:
  ldrb    r3,[r1],#8                     @r3 = *CurrentPtr
  ldr     r2,[memptr,r3,lsl#2]           @ r2 = FastTable[r3], u32 table
  str     r2,[r0],#32
  ldrb    r3,[r1],#8                     @r3 = *CurrentPtr
  ldr     r2,[memptr,r3,lsl#2]           @ r2 = FastTable[r3], u32 table
  str     r2,[r0],#32
  ldrb    r3,[r1],#8                     @r3 = *CurrentPtr
  ldr     r2,[memptr,r3,lsl#2]           @ r2 = FastTable[r3], u32 table
  str     r2,[r0],#32
  ldrb    r3,[r1],#8                     @r3 = *CurrentPtr
  ldr     r2,[memptr,r3,lsl#2]           @ r2 = FastTable[r3], u32 table
  str     r2,[r0],#32
  
  subs  r4,r4,#1          @ loop--
  bpl   fastLoopRow1bpp
  mov   r15,r14

.align 4
.global cyclesPerScanline
cyclesPerScanline: .space 4
vramOffset: .space 4
interlace: .space 4
.global timerDecrement
timerDecrement: .space 4

.macro updateTimers
@ update the timers
  adr   lr,9f
  ldr   pc,=doViaPolls @ VIAPolls
9:
.endm

#if USE_1770_DISC
.macro updateDisc
  ldr   r1,=disc_trigger
  ldr   r2,[r1]
  cmp   r2,#0
  beq   8f

  subs  r2,r2,#128
  movmi r2,#0
  str   r2,[r1]
  bpl   8f

  ldr   r2,=poll1770
  adr   lr,8f
  bx    r2
8:
.endm

#else
.macro  updateDisc  
  ldr   r1,=disc_trigger
  ldr   r2,[r1]
  cmp   r2,#0
  beq   8f
  subs  r2,r2,#128
  movmi r2,#0
  str   r2,[r1]
  bpl   8f
  @ if -ve, need to call function
  ldr    r1,=current_command
  ldr    r1,[r1]
  ldr    r2,[r1,#16]  @ the intfunc
  cmp    r2,#0
  adrne   lr,8f
  bxne    r2
8:
.endm
#endif
  
#if defined(ELITE_HACK)
.type fixEliteSplit,function
fixEliteSplit:
  ldr    r1,=VideoULA_ControlReg
  ldrb   r2,[r1]
  @  need to reset the scanline function..
  @  ldr    r1,=scanlineFunction  @LscanlineF
  cmp    r2,#16
  bgt    endEliteHack
  ldr    r1,=LscanlineFns
  tst    r2,#16
  ldrne  r12,[r1,#8]
  ldreq  r12,[r1,#0]
  beq    1f
  tst  r2,#8
  ldrne  r12,[r1,#4]
  ldreq  r12,[r1,#0]
1:
  str    r12,scanlineFunction
  b endEliteHack

LscanlineFns:
.word lowLevelDoScanLineWide
.word lowLevelDoScanLine1bpp
.word lowLevelDoScanLineNarrow


#endif

.global videoDoScanLine
.extern doViaPolls
.extern intStatus
.type videoDoScanLine,function
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
videoDoScanLine:
  @ This is called every scanline. 
  @ right, check where we are on screen:
  @ can use r0-r2 freely, push anything else...
  @CRTC_VerticalDisplayed
  stmfd   sp!,{r3-r12}

  updateTimers
#if !defined(USE_1770_DISC)
  updateDisc    @ unfotunately, I need to poll the disc and timer
                @ frequently. This is a major slowdown.
                @ Both are expensive even if nothing is set.
#endif
  
  ldr     r0,CRTCVals  
  @ check teletext fudge:
  
  ldr     r5,=VideoState
  ldr     r1,[r5,#OFFSET_IsTeletext]
  cmp     r1,#0
  bne     skipScanLine     @ could get rid of this check by making teletext
                           @ use its own cut-down "scanline" routine.
                           @ but I'd only save 3 instructions per scan...

@  ldr     r1,interlace
@  cmp     r1,#0
@  bne     skipScanLine
  
  ldr     r2,[r0,#OFFSET_VerticalDisplayed]    @CRTC_VerticalDisplayed
  @VideoState.CharLine
  ldr     r1,[r5,#OFFSET_CharLine]
  cmp     r1,r2           @ charline - CRTC_VerticalDisplayed
  bcs     skipScanLine    @ branch if charline >= CRTC_VerticalDisplayed

#if defined(ELITE_HACK)
  cmp     r1,#0x18
  beq     fixEliteSplit   
endEliteHack:
#endif

  @ so we are in a frame...
  @ if InCharLine == ScanLinesPerChar, we need to get video mem ptr
  ldr     r1,[r5,#OFFSET_InCharLine]
  cmp     r1,#7
  beq     addressFetch
  adr     lr,endScanLineFunction      @ down there
  ldr     pc,scanlineFunction

addressFetch:
   @ if we are at start of character, get beeb video mem address
   @ misses out a few important facts, but this is basically it...
    ldr    r1,[r5,#OFFSET_Addr]
    add    r2,bbc_mem,r1,lsl#3          @ r2 = WholeRam + VideoState.Addr*8
    str    r2,[r5,#OFFSET_DataPtr]      @ videostate.DataPtr = r2
   @ also VideoState.Addr+=CRTC_HorizontalDisplayed;
    ldr   r2,[r0,#OFFSET_HorizontalDisplayed]    @r2 = CRTC_HorizontalDisplayed
    add   r1,r1,r2
    str   r1,[r5,#OFFSET_Addr]
@skipAddressFetch:
  @ now check if we need to draw a scaline... assume we always do for now
  @ should jump to wide OR narrow depending on bit 4 of the video control reg
  @ if bit 4 set == narrow pixels 
  @ bit 4 not set == wide pixels
  adr      lr,endScanLineFunction
  ldr      pc,scanlineFunction
.global scanlineFunction @ space for the scanline function!
scanlineFunction: .space 4
  @bl       lowLevelDoScanLineNarrow
  @ VideoState.PixmapLine+=1;
endScanLineFunction:
  ldr     r0,CRTCVals
.type skipScanLine,function
skipScanLine:
@  ldr      r1,[r5,#OFFSET_PixmapLine]
@  add      r1,r1,#1
@  str      r1,[r5,#OFFSET_PixmapLine]
  @Move onto next physical scanline as far as timing is concerned
  add   r2,r5,#OFFSET_CharLine
  ldmia r2,{r3-r6}             @ load with temp increment, r2 unaffected
  @ r3 = VideoState.CharLine
  @ r4 = VideoState.InCharLine
  @ r5 = VideoState.InCharLineUp
  @ r6 = VideoState.VSyncState
  
  add      r5,r5,#1       @InCharLineUp += 1
  subs     r6,r6,#1           @ vsync-- if != 0
@if (!(--VideoState.VSyncState)) ...
  movmi      r6,#0
  movpl      r1,#0      @ r1 = 0 (input to routine)
  adrpl      lr,5f
  ldrpl      pc,=SysVIAIR
5:
  subs      r4,r4,#1     @InCharLine--     (at start of frame this gets set to 7)
  bpl      storeVidState        @ not reached the end of a character...
  @ if InCharLine < 0, then 
  add      r3,r3,#1       @CharLine++      (reached end of a character)
  mov      r4,#7          @InCharLine = 7
  mov      r5,#0          @InCharLineUp = 0
  @ compare vsync
  cmp      r6,#0          @vsync == 0?
  bne      storeVidState
  ldr     r1,[r0,#OFFSET_VerticalSyncPos]
  cmp     r3,r1
  bne     storeVidState

.global vsyncTrigger
.type vsyncTrigger, function
vsyncTrigger:
  @ if vsync == 0 AND charline == CRTC_VerticalSyncPos
  @ then PixmapLine = 0 and VSyncState=3
  @ also, call sys interrupt trigger 
  @str     r6,[r2,#OFFSET_PixmapLine]    @ pixmapline = 0?
  mov     r6,#3
  mov     r1,#1      @ r1 = 1 - input to SysVIAIR routine
  adr     lr,storeVidState
  ldr      pc,=SysVIAIR
  
  @ now store the registers...
.type storeVidState,function
storeVidState:
  @ stores charline to vsyncstate in VideoState
  stmia    r2,{r3-r6}
  
  
  @ return
.global doEnding
.type doEnding, function
doEnding:
@ pop main emulation regs
  ldmfd   sp!,{r3-r12}
  ldr     r2,[r0,#OFFSET_VerticalTotal]      @CRTC_VerticalTotal
  ldr     r0,=VideoState @AsmVideoState
  ldr     r1,[r0,#OFFSET_CharLine]
  cmp     r1,r2         @ r1 - r2
  adrmi   r2,videoDoScanLine
  adrpl   r2,videoStartOfFrame      @ next event
  ldr     r1,cyclesPerScanline              @ based on h total etc
  addmi   bbc_cycles,bbc_cycles,r1
@  ldr bbc_cycles,cyclesPerScanline
  bmi     scanLine
@ else is a frame
  ldr     r0,interlace
  eors    r0,r0,#1
  movne   r1,r1,lsl#2
  moveq   r1,r1,lsl#3
  add     bbc_cycles,bbc_cycles,r1
  str     r0,interlace 
scanLine:
#if defined(ELITE_HACK)
  sub     r1,bbc_cycles,#23  @ hack to take SysVIAs into account!
  str     r1,timerDecrement
#else
  str     bbc_cycles,timerDecrement
#endif
  ldr     r1, =intStatus
  ldr     r0, [r1]            @ r0 = int status
  cmp     r0, #0
  ldreq   r0, =endScanline    @ intstatus = 0, end
  beq   1f                    @ no interrupt
  @ else test I val
@.type endingSetI,function
@endingSetI:
@  ldr     r1,bbc_di
@  tst     r1,#I                   @ is I set?
@  ldreq   r0,=doInterrupt         @ Equals = Zero = bit not set
@  ldrne   r0,=endScanline         @ Equals = Zero = bit not set
@1:
@  ldr     r1,=nextEvent
@  str     r2,[r1]
@  mov     pc,r0 @=endScanline or doInterrupt
@ test I val
  ldr     r1,bbc_di
  tst     r1,#I                   @disable IRQ
  ldreq   r0,=doInterrupt         @ Equals = Zero = bit not set
  bne     1f                      @ no interrupt
  ldr     r1,=nextEvent
  str     r2,[r1]
  mov     pc,r0
1:
  @ the interrupt bit was 0, or I not set
  @ check NMI ir
  ldr     r1,=nextEvent
  str     r2,[r1]
  ldr     r1, =NMIintStatus
  ldr     r2, [r1]            @ get disk interrupt status
  tst     r2,#8
  ldreq   r0,=endScanline 
  moveq   pc,r0
  @ else disk interrupt is set
  tst     r2,#16              @ are we already NMIing?
  orr     r2,r2,#16      
  ldrne   r0,=endScanline     @ yes, don't do so again
  ldreq   r0,=NMIinterrupt    @ no, cause NMI interrupt right now
  streq   r2, [r1]            @ store disk interrupt status
  mov   pc,r0


.global doTeletextScreen
.extern teletextScreen
.type doTeletextScreen, function
doTeletextScreen:
  @ Teletext screen - or Mode 7 as it is also known...
  @ A basic text mode (shocker), but with a 
  @ few tricks thrown in too,  "double mode" and flashing for example.
  @ OK... lets try a basic teletext mode using 2 gba backgrounds.. 
  @ can't use rotate bg/sprites for a couple of reasons
  @ 1) Palettes - without them it makes text a pain to emulate
  @ 2) Size - standard 8*8 text glyphs would be pretty unreadable shrunk to fit
  @           The size of a teletext screen is 272*256 - too big in x and y
  @ So that leaves me with this: bg 1 and 2 in a very literal "text mode".
  @ forground colours are bg 1, bg colours use bg 2


@ TODO: need to sort out interrupts. they generally work as is...
@ when we finish the screen, fire a CA1 = 1 event
@ when we  start the screen, fire a CA1 = 0 event
  ldr   r1,=disc_trigger  @ this hack means games load a bit quicker. 
  ldr   r0,[r1]
  cmp   r0,#0
  bleq    asm_teletext
.type endOfTT,function
endOfTT:
@ End of teletext
  movpl      r1,#0      @ r1 = 0 (input to routine)
  adrpl      lr,0f
  ldrpl      pc,=SysVIAIR
0:
  @ check here if disc_trigger is non-zero
  @ if it is, call the disc interrupt
  @ then if NMIintStatus is non-zero, go there immediately
  ldr   r1,=disc_trigger
  ldr   r0,[r1]
  cmp   r0,#0
 
  ldr     bbc_cycles,cyclesPerScanline   @128 cycles per scanline?
  moveq   bbc_cycles,bbc_cycles,lsl#3   @ this fixes MM and Swoop!
@  movne   bbc_cycles,bbc_cycles,lsl#3   @ this fixes MM and Swoop!
  movne    bbc_cycles, #0
  str     bbc_cycles,timerDecrement
  ldr   r1, =intStatus
  ldr   r0, [r1]            @ get interrupt status - prolly should check every scanline?
  cmp   r0,#0

  adr   r2,videoDoScanLine @ next event
  
  beq   1f
  @ test I val
  ldr     r1,bbc_di
  tst     r1,#I                   @disable IRQ
  ldreq   r0,=doInterrupt         @ Equals = Zero = bit not set
  bne     1f
  ldr     r1,=nextEvent
  str     r2,[r1]
  mov     pc,r0
1:
  @ the interrupt bit was 0, or I not set
  @ check NMI ir
  ldr   r1,=nextEvent
  str   r2,[r1]
  ldr   r1, =NMIintStatus
  ldr   r2, [r1]            @ get disk interrupt status
  tst   r2,#8
  ldreq   r0,=endScanline 
  moveq   pc,r0
  
  tst   r2,#16
  orr   r2,r2,#16
  ldrne   r0,=endScanline
  ldreq   r0,=NMIinterrupt
  streq   r2, [r1]            @ get disk interrupt status
  mov   pc,r0


.type skipScreenDraw, function
skipScreenDraw:
  ldr     r0,CRTCVals
  ldr   r2,[r0,#OFFSET_VerticalTotal]      @CRTC_VerticalTotal
  ldr   r0,=VideoState @AsmVideoState
  str   r2,[r0,#OFFSET_CharLine]
  mov      r1,#0      @ r1 = 0 (input to routine)
  adr      lr,0f
  ldr      pc,=SysVIAIR
0:
  mov    bbc_cycles, #0
  str     bbc_cycles,timerDecrement
  ldr   r1, =intStatus
  ldr   r0, [r1]            @ get interrupt status - prolly should check every scanline?
  cmp   r0,#0


  adr   r2,videoDoScanLine @ next event
  beq   1f
  @ test I val
  ldr     r1,bbc_di
  tst     r1,#I                   @disable IRQ
  ldreq   r0,=doInterrupt         @ Equals = Zero = bit not set
  bne     1f
  ldr     r1,=nextEvent
  str     r2,[r1]
  mov     pc,r0
1:
  @ the interrupt bit was 0, or I not set
  @ check NMI ir
  ldr   r1,=nextEvent
  str   r2,[r1]
  ldr   r1, =NMIintStatus
  ldr   r2, [r1]            @ get disk interrupt status
  tst   r2,#8
  ldreq   r0,=endScanline 
  moveq   pc,r0
  
  tst   r2,#16
  orr   r2,r2,#16
  ldrne   r0,=endScanline
  ldreq   r0,=NMIinterrupt
  streq   r2, [r1]            @ get disk interrupt status
  mov   pc,r0



.global videoStartOfFrame
.extern checkKeys
.extern IC32State
.type videoStartOfFrame, function
@ every CRTC_VerticalTotal lines, call this routine
@ Sets start of video ram reference, etc
videoStartOfFrame:
  @ can use just regs 0-2 so as not to wreck cpu state...
  @ VideoState.CharLine=0;
  @ VideoState.InCharLine=CRTC_ScanLinesPerChar;
  @ VideoState.InCharLineUp=0;

@ each frame check keys...
  adr       lr,endKeyCheck
  ldr       pc,=checkKeys
  
endKeyCheck:
  stmfd sp!,{r0-r11}   @ push just those used by doViaPoll.
  updateTimers
#if defined(USE_1770_DISC)
  updateDisc
#endif
  ldmfd sp!,{r0-r11}
  
  @bl    beebKeyIR
  ldr   r1,=VideoState @AsmVideoState
  @ check teletext mode isn't on...
  ldr     r2,[r1,#OFFSET_IsTeletext]
  cmp     r2,#0
  bne     doTeletextScreen  @ does the whole thing at once
  @
#if defined(USE_1770_DISC)
  ldr   r2,=disc_trigger
  ldr   r2,[r2]
  cmp   r2,#0
  bne   skipScreenDraw
#endif

#if defined(ELITE_HACK)
@ In elite, the screen has a mode split at a certain line
@ this is the most reliable way to time the change
@ as my "timers" are not accurate enough and split it about
@ 20 scanlines too high, then halfway through the second split.
  ldr    r1,=VideoULA_ControlReg
  ldrb   r3,[r1]
  @  need to reset the scanline function..
  cmp    r3,#16
  bcs    1f

  ldr    r1,=LscanlineFns
  tst    r3,#16
  ldrne  r2,[r1,#8]
  ldreq  r2,[r1,#0]
  strne  r2,scanlineFunction
  bne    1f
  tst    r3,#8
  ldrne  r2,[r1,#4]
  ldreq  r2,[r1,#0]
  str    r2,scanlineFunction
1:
  ldr   r1,=VideoState @AsmVideoState
#endif

@  mov     r2,#128
@  str     r2,discDecrement

  @ set the vramOffset for the screen wraparound...
  ldr   r2,=IC32State
  ldr   r2,[r2]
  and   r2,r2,#0x30
  ldr   r0,=LvramOffsetTable
  ldr   r0,[r0,r2,lsr#2]           @ r0 = value in table ic32&0x30 >> 4
  str   r0,vramOffset
  
  mov   r2,#0x7                     @CRTC_ScanLinesPerChar = 7
  str   r2,[r1,#OFFSET_InCharLine]    
  mov   r2,#0
  str   r2,[r1,#OFFSET_CharLine]
  str   r2,[r1,#OFFSET_InCharLineUp]
@  str   r2,[r1,#OFFSET_PixmapLine]
  @VideoState.Addr=CRTC_ScreenStartLow+(CRTC_ScreenStartHigh<<8);
  ldr     r0,CRTCVals
  ldrb  r2,[r0,#OFFSET_ScreenStartLow]
  ldrb  r3,[r0,#OFFSET_ScreenStartHigh]         @bbc_cycles is recycled below
  orr   r2,r2,r3,lsl#8      @ Addr = low|hi<<8
  str   r2,[r1,#OFFSET_Addr]
@ set cycles for the next event, next event is a scanline
endVideoStartFrame:
  ldr   bbc_cycles,cyclesPerScanline   @128 cycles per scanline?
  str   bbc_cycles,timerDecrement
  @ check irq 
  ldr   r1, =intStatus
  ldr   r0, [r1]            @ get interrupt status - prolly 
                            @ should check every scanline?
  cmp   r0,#0
  adr   r2,videoDoScanLine      @ next event
  @ldr   r0,=endScanline         @ go here now
  
  ldreq   r0,=endScanline         @ go here now  
  beq   noInterrupt
  @ test I val
.type setInterrupt,function
setInterrupt:
  ldr     r1,bbc_di
  tst     r1,#I                   @disable IRQ
  ldreq   r0,=doInterrupt         @ Equals = Zero = bit not set
@  ldrne   r0,=endScanline         @ Not Equals = Not Zero = bit is set
  bne noInterrupt
  ldr     r1,=nextEvent
  str     r2,[r1]
  mov     pc,r0
noInterrupt:
@  ldr   r1,=nextEvent
@  str   r2,[r1]
@  mov   pc,r0
@ the interrupt bit was 0, or I not set
  @ check NMI ir
  ldr   r1,=nextEvent
  str   r2,[r1]
  ldr   r1, =NMIintStatus
  ldr   r2, [r1]            @ get disk interrupt status
  tst   r2,#8
  ldreq   r0,=endScanline 
  moveq   pc,r0
  
  tst   r2,#16
  orr   r2,r2,#16
  ldrne   r0,=endScanline
  ldreq   r0,=NMIinterrupt
  streq   r2, [r1]            @ get disk interrupt status
  mov   pc,r0
  


.align 4
@LcheckKeys:
@.word checkKeys
@VIAPolls: .word doViaPolls
@LIC32State: .word IC32State

@@ when vram address is > 8000, these values should be subtracted
@@ index is (ic32state&0x30)>>4
LvramOffsetTable:
.word 0x4000
.word 0x2000
.word 0x5000
.word 0x2800
 
  
  

@ copy the vide state in here
@AsmVideoState:  
@  .word VideoState    @<- pointer to videostate
@ the offsets to the various parts of the ULA video state
.equ OFFSET_Addr,                     0   @Address of start of next visible character line in beeb memory  
@.equ OFFSET_PixmapLine,               4   @Current line in the pixmap 
.equ OFFSET_PreviousFinalPixmapLine,  8   @The last pixmap line on the previous frame 
.equ OFFSET_IsTeletext,               12  @This frame is a teletext frame - do things differently 
.equ OFFSET_DataPtr,                  16  @Pointer into host memory of video data
.equ OFFSET_CharLine,                 20  @6845 counts in characters vertically - 0 at the top , incs by 1 
                                          @ -1 means we are in the bit before the actual display starts
.equ OFFSET_InCharLine,               24  @Scanline within a character line - counts down
.equ OFFSET_InCharLineUp,          28 @   32  @Scanline within a character line - counts up
.equ OFFSET_VSyncState,            32 @   36  @When >0 VSync is high

@VideoStateT

.align 4
@.global CRTCVals
CRTCVals:
.word CRTCControl

.equ  OFFSET_ControlReg           , (0*4)
.equ  OFFSET_HorizontalTotal      , (1*4)
.equ  OFFSET_HorizontalDisplayed  , (2*4)
.equ  OFFSET_HorizontalSyncPos    , (3*4)
.equ  OFFSET_SyncWidth            , (4*4)
.equ  OFFSET_VerticalTotal        , (5*4)
.equ  OFFSET_VerticalTotalAdjust  , (6*4)
.equ  OFFSET_VerticalDisplayed    , (7*4)
.equ  OFFSET_VerticalSyncPos      , (8*4)
.equ  OFFSET_InterlaceAndDelay    , (9*4)
.equ  OFFSET_ScanLinesPerChar     , (10*4)
.equ  OFFSET_CursorStart          , (11*4)
.equ  OFFSET_CursorEnd            , (12*4)
.equ  OFFSET_ScreenStartHigh      , (13*4)
.equ  OFFSET_ScreenStartLow       , (14*4)
.equ  OFFSET_CursorPosHigh        , (15*4)
.equ  OFFSET_CursorPosLow         , (16*4)
.equ  OFFSET_LightPenHigh         , (17*4)
.equ  OFFSET_LightPenLow          , (18*4)

             
.pool        
