
#include "backgrounds.h"
#include "inputcodes.h"
#include "screenmode.h"
#include "dma.h"
#include "gfx.h"

/*

F0,F1,F2,F3,F4,F5,F6,F7,F8,F9,BREAK
ESC,1,2,3,4,5,6,7,8,9,0,-,^,\,LFT,RGT
TAB,Q,W,E,R,T,Y,U,I,O,P,@,[,_,up,down
CAPS,CTRL,A,S,D,F,G,H,J,K,L,;,:,],RETURN
SFTLOCK,SHIFT,Z,X,C,V,B,N,M,,,.,/,SHIFT,DEL,COPY
        SPACE
*/
#define VK_TILE_INDEX  128    // initial tile index
const int rowCount[] = {
  11,
  16,
  16,
  15,
  15,
  1,
};
const int rowOffset[] = {
  2,
  1,
  1,
  0,
  0,
  5,
};

// key maps map to indices in the input code tables
const u8 virtualRow0[] = {
  0/*F0*/,1/*F1*/,2/*F2*/,3/*F3*/,4/*F4*/,5/*F5*/,
  6/*F6*/,7/*F7*/,8/*F8*/,9/*F9*/,10|0x80/*BREAK*/
};
const u8 keymapRow0[] = {
  0x61/*F0*/,0x62/*F1*/,0x63/*F2*/,0x64/*F3*/,0x65/*F4*/,0x66/*F5*/,
  0x67/*F6*/,0x68/*F7*/,0x69/*F8*/,0x6a/*F9*/,0x6b/*BREAK*/
};

const u8 virtualRow1[] = {
  11|0x80/*ESC*/,12/*1*/,13/*2*/,14/*3*/,15/*4*/,16/*5*/,17/*6*/,
  18/*7*/,19/*8*/,20/*9*/,21/*0*/,22/*-*/,23/*^*/,24/*\*/,25/*LFT*/,26/*RGT*/
};
const u8 keymapRow1[] = {
  0x05/*ESC*/,0x31/*1*/,0x32/*2*/,0x33/*3*/,0x34/*4*/,0x35/*5*/,0x36/*6*/,
  0x37/*7*/,0x38/*8*/,0x39/*9*/,0x30/*0*/,0x2d/*-*/,0x5e/*^*/,0x5c/*\*/,0x09/*LFT*/,0x0a/*RGT*/
};

const u8 virtualRow2[] = {
  27|0x80/*TAB*/,28/*Q*/,29/*W*/,30/*E*/,31/*R*/,32/*T*/,33/*Y*/,34/*U*/,
  35/*I*/,36/*O*/,37/*P*/,38/*@*/,39/*[*/,40/*_*/,41/*up*/,42/*down*/,
};
const u8 keymapRow2[] = {
  0x02/*TAB*/,'Q'/*Q*/,'W'/*W*/,'E'/*E*/,'R'/*R*/,'T'/*T*/,'Y'/*Y*/,'U'/*U*/,
  'I'/*I*/,'O'/*O*/,'P'/*P*/,'@'/*@*/,'['/*[*/,'_'/*_*/,0x07/*up*/,0x08/*down*/,
};

const u8 virtualRow3[] = {
  43|0x80/*CAPS*/,44|0x80/*CTRL*/,45/*A*/,46/*S*/,47/*D*/,48/*F*/,49/*G*/,
  50/*H*/,51/*J*/,52/*K*/,53/*L*/,54/*;*/,55/*:*/,56/*]*/,57|0x80/*RETURN*/
};
const u8 keymapRow3[] = {
  0x04/*CAPS*/,0x03/*CTRL*/,'A'/*A*/,'S'/*S*/,'D'/*D*/,'F'/*F*/,'G'/*G*/,
  'H'/*H*/,'J'/*J*/,'K'/*K*/,'L'/*L*/,';'/*;*/,':'/*:*/,']'/*]*/,0xd/*RETURN*/
};

const u8 virtualRow4[] = {
  58|0x80/*SFTLOCK*/,59|0x80/*SHIFT*/,60/*Z*/,61/*X*/,62/*C*/,63/*V*/,64/*B*/,
  65/*N*/,66/*M*/,67/*,*/,68/*.*/,69/* / */,59|0x80/*SHIFT*/,70/*DEL*/,71/*COPY*/
};
const u8 keymapRow4[] = {
  0x0e/*SFTLOCK*/,0xc/*SHIFT*/,'Z'/*Z*/,'X'/*X*/,'C'/*C*/,'V'/*V*/,'B'/*B*/,
  'N'/*N*/,'M'/*M*/,','/*,*/,'.'/*.*/,0x2f/* / */,0xc/*SHIFT*/,0xb/*DEL*/,0x6/*COPY*/
};
const u8 virtualRow5[] = {
  73|0x80 // space
};
const u8 keymapRow5[] = {
  0x20 // space
};

const u8 * const rows[] = {
  virtualRow0,
  virtualRow1,
  virtualRow2,
  virtualRow3,
  virtualRow4,
  virtualRow5,
};
const u8 * const rowKeyMaps[] = {
  keymapRow0,
  keymapRow1,
  keymapRow2,
  keymapRow3,
  keymapRow4,
  keymapRow5
};

#define  font_data  _binary_font_raw_start 
#define  font_pal  _binary_font_pal_start 
#define  font_map  _binary_font_map_start 
extern unsigned char font_data[];
extern unsigned char font_pal[];
extern unsigned short font_map[];
int apply_vk_press = 0;


void load_vk_tiles(void) {
  // load the tiles to bg tile bank 0, some way in
  int i;
  for (i = 0 ; i < 128; i++) {
    REG_DMA3SAD = (u32)(&font_data[font_map[i]*32]);
    REG_DMA3DAD = (u32)(0x6001000+i*32);
    REG_DMA3CNT = (32) | DMA_16NOW; // 16 bit transfers from gfx array via DMA
  }
  
  REG_DMA3SAD = (u32)font_pal;
  REG_DMA3DAD = (u32)(&BGPaletteMem[16*14]);
  REG_DMA3CNT = 16 | DMA_16NOW; // 16 bit transfers from gfx array via DMA

  volatile u16 * selectedPal = &BGPaletteMem[16*15];
  volatile u16 * regularPal = &BGPaletteMem[16*14];
  for (i = 1; i < 16; i++) {
    int c = regularPal[i];
    int r = c&0x1f;
    int g = (c>>5)&0x1f;
    int b = (c>>10)&0x1f;
    selectedPal[i] = RGB16((0x1f-r),(0x1f-g),(0x1f-b));
    
  }
}

// now show the keyboard...
void show_vkb(void) {
  // show it on bg - use palette 14-15
  int i,j,k,x;
  j = VK_TILE_INDEX; // j = tile index
  const u8 * vr;
  for (i = 0; i < 5; i++) {
    vr = rows[i];
    x = 0;
    for (k = 0; k < rowCount[i]; k++) {
      // show this row of tiles
      bg_menu.mapData[i*32 + x + rowOffset[i]] = j|(14<<12);
      j++;
      x++;
      if (vr[k] & 0x80) {
        bg_menu.mapData[i*32 + x + rowOffset[i]] = j|(14<<12);
        j++;
        x++;
      }
    }
  }
  x = 0;
  for (k = 0; k < 5; k++) {
    bg_menu.mapData[i*32 + x + rowOffset[i]] = j|(14<<12);
    j++;
    x++;
  }
}
typedef struct {
  u16 row;
  u16 col;
} vk_selection_t;

vk_selection_t vk_selection = {};

void get_vk_selected(int * row, int * col, int * key) {
  *key = rowKeyMaps[vk_selection.row][vk_selection.col];
  *row = GET_TABLE_VAL(*key,KEY_ROW);
  *col = GET_TABLE_VAL(*key,KEY_COL);
}



void update_vk_selection(void) {
  show_vkb();
  if (vk_selection.row != 5) {
    int j,x;
    x = 0;
    int i = vk_selection.row;
    for (j = 0; j < vk_selection.col; j++) {
      x++;
      if (rows[i][j] &0x80)
        x++;
    }
   // x--;
    int k = bg_menu.mapData[i*32 + x + rowOffset[i] ] & 255;
  
    bg_menu.mapData[i*32 + x + rowOffset[i] ] = k|(15<<12);
    if (rows[i][j] &0x80)
      bg_menu.mapData[i*32 + x + 1+ rowOffset[i] ] = (k+1)|(15<<12);
  } else {
    int j,x = 0;
    int k = bg_menu.mapData[5*32 + 0+ rowOffset[5] ] & 255;
    for (j = 0; j < 5; j++) {
      bg_menu.mapData[5*32 + x+ rowOffset[5] ] = k|(15<<12);
      k++;
      x++;
    }
  }
}

void set_vk_selected(int key) {
  // check over rows/cols
  int i,j;
  for (i = 0; i < 6; i++) {
    for (j = 0; j < rowCount[i]; j++) {
      if (key == rowKeyMaps[i][j]) {
        vk_selection.row = i;
        vk_selection.col = j;
        update_vk_selection();
        return;
      }
    }
  }
}

void vk_left(void) {
  if (vk_selection.col)
    vk_selection.col--;
  else
    vk_selection.col = rowCount[vk_selection.row]-1;
  update_vk_selection();
}
void vk_right(void) {
  if (vk_selection.col < (rowCount[vk_selection.row]-1))
    vk_selection.col++;
  else
    vk_selection.col = 0;
  update_vk_selection();
}

void vk_up(void) {
  if (vk_selection.row)
    vk_selection.row--;

  if (vk_selection.col >= rowCount[vk_selection.row]) {
    vk_selection.col = rowCount[vk_selection.row]-1;
  }
  if (vk_selection.row == 4)
    vk_selection.col = 2;  
  update_vk_selection();
}
void vk_down(void) {
  if (vk_selection.row < 5)
    vk_selection.row++;

  if (vk_selection.col >= rowCount[vk_selection.row]) {
    vk_selection.col = rowCount[vk_selection.row]-1;
  }
  update_vk_selection();
}

const u8 vk_positions[6][2] = {
  {0,0},
  {(255-96),0},
  {0,(255-56)},
  {(255-96),(255-56)},
  {0,(255-112)},
  {(255-96),(255-112)},

};
void move_vk(int direction) {
  int curr_pos = 0;
  int i;
  for (i = 0; i < 6; i++) {
    if (vk_positions[i][0] == bg_menu.x_scroll &&
	vk_positions[i][1] == bg_menu.y_scroll)
    {
      curr_pos = i;
      break;
    }
  }
  curr_pos += direction;
  if (curr_pos < 0) curr_pos = 5;
  if (curr_pos > 5) curr_pos = 0;

  bg_menu.x_scroll = vk_positions[curr_pos][0];
  bg_menu.y_scroll = vk_positions[curr_pos][1];
  UpdateBackground(&bg_menu);
}

